@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -41,9 +41,8 @@ using namespace ipelua;
 // --------------------------------------------------------------------
 
 const char * const AppUiBase::selectorNames[] =
-  { "stroke", "fill", "pen", "textsize", "markshape",
-    "symbolsize", "gridsize", "anglesize", "view", "page",
-    "viewmarked", "pagemarked" };
+  { "stroke", "fill", "pen", "dashstyle", "textsize", "markshape", "symbolsize",
+    "opacity", "gridsize", "anglesize", "view", "page", "viewmarked", "pagemarked" };
 
 AppUiBase::AppUiBase(lua_State *L0, int model)
 {
@@ -68,30 +67,36 @@ AppUiBase::AppUiBase(lua_State *L0, int model)
   lua_pop(L, 1); // width_notes_bookmarks
 
   iUiScale = 100;
-  lua_getfield(L, -1, "ui_scale");
+  lua_getfield(L, -1, "ui_scaling");
   if (lua_isnumber(L, -1))
     iUiScale = lua_tointegerx(L, -1, nullptr);
-  lua_pop(L, 1); // ui_scale
+  lua_pop(L, 1); // ui_scaling
 
   iToolbarScale = 100;
-  lua_getfield(L, -1, "toolbar_scale");
+  lua_getfield(L, -1, "toolbar_scaling");
   if (lua_isnumber(L, -1))
     iToolbarScale = lua_tointegerx(L, -1, nullptr);
-  lua_pop(L, 1); // toolbar_scale
+  lua_pop(L, 1); // toolbar_scaling
 
+  // win_ui_gap separates the input elements vertically, so the
+  // can be touched with a finger
   iUiGap = 0;
-  lua_getfield(L, -1, "ui_gap");
+  lua_getfield(L, -1, "win_ui_gap");
   if (lua_isnumber(L, -1))
     iUiGap = lua_tointegerx(L, -1, nullptr);
-  lua_pop(L, 1); // ui_gap
+  lua_pop(L, 1); // win_ui_gap
 
-  lua_getfield(L, -1, "mini_edit");
+  // win_mini_edit leaves only the most important buttons on the Edit
+  // toolbar
+  lua_getfield(L, -1, "win_mini_edit");
   isMiniEdit = lua_toboolean(L, -1);
-  lua_pop(L, 1); // mini_edit
+  lua_pop(L, 1); // win_mini_edit
 
-  lua_getfield(L, -1, "left_dock_floats");
+  // win_left_panels_float makes the panels on the left float on top
+  // of the canvas
+  lua_getfield(L, -1, "win_left_panels_float");
   iLeftDockFloats = lua_toboolean(L, -1);
-  lua_pop(L, 1); // left_dock_floats
+  lua_pop(L, 1); // win_left_panels_float
 
   iScalings.push_back(1);
   lua_getfield(L, -1, "scale_factors");
@@ -115,6 +120,11 @@ AppUiBase::~AppUiBase()
   lua_gc(L, LUA_GCCOLLECT, 0);
 }
 
+int AppUiBase::dpi() const
+{
+  return 96;
+}
+
 // --------------------------------------------------------------------
 
 void AppUiBase::buildMenus()
@@ -154,6 +164,7 @@ void AppUiBase::buildMenus()
   addItem(EEditMenu, "Cut", "cut");
   addItem(EEditMenu, "Copy", "copy");
   addItem(EEditMenu, "Paste", "paste");
+  addItem(EEditMenu, "Paste with layer", "paste_with_layer");
   addItem(EEditMenu, "Paste at cursor", "paste_at_cursor");
   addItem(EEditMenu, "Delete", "delete");
   addItem(EEditMenu);
@@ -222,6 +233,12 @@ void AppUiBase::buildMenus()
   addSubItem("No", "transformabletext|false");
   endSubMenu();
 
+  startSubMenu(EPropertiesMenu, "Spline type");
+  addSubItem("bspline", "splinetype|bspline");
+  addSubItem("cardinal", "splinetype|cardinal");
+  addSubItem("spiro", "splinetype|spiro");
+  endSubMenu();
+
   addItem(EModeMenu, "Select objects (with Shift: non-destructive)",
 	  "mode_select");
   addItem(EModeMenu, "Translate objects (with Shift: horizontal/vertical)",
@@ -233,6 +250,7 @@ void AppUiBase::buildMenus()
   addItem(EModeMenu, "Move graph nodes", "mode_graph");
   addItem(EModeMenu, "Pan the canvas", "mode_pan");
   addItem(EModeMenu, "Shred objects", "mode_shredder");
+  addItem(EModeMenu, "Laser pointer", "mode_laser");
   addItem(EModeMenu);
   addItem(EModeMenu, "Text labels", "mode_label");
   addItem(EModeMenu, "Mathematical symbols", "mode_math");
@@ -270,6 +288,7 @@ void AppUiBase::buildMenus()
   addItem(ESnapMenu, "Snap to boundary", "@*snapbd");
   addItem(ESnapMenu, "Snap to intersection", "@*snapint");
   addItem(ESnapMenu, "Snap to grid", "@*snapgrid");
+  addItem(ESnapMenu, "Snap to custom grid", "@*snapcustom");
   addItem(ESnapMenu, "Angular snap", "@*snapangle");
   addItem(ESnapMenu, "Automatic snap", "@*snapauto");
   addItem(ESnapMenu);
@@ -343,8 +362,12 @@ void AppUiBase::buildMenus()
   addItem(EViewMenu, "New view", "new_view");
   addItem(EViewMenu, "Delete view", "delete_view");
   addItem(EViewMenu);
+  addItem(EViewMenu, "Mark views from this view", "mark_from_view");
+  addItem(EViewMenu, "Unmark views from this view", "unmark_from_view");
+  addItem(EViewMenu);
   addItem(EViewMenu, "Jump to view", "jump_view");
   addItem(EViewMenu, "Edit view", "edit_view");
+  addItem(EViewMenu, "View sorter", "view_sorter");
 
   addItem(EPageMenu, "Next page", "next_page");
   addItem(EPageMenu, "Previous page", "previous_page");
@@ -644,6 +667,11 @@ void AppUiBase::showInCombo(const Cascade *sheet, Kind kind,
     addCombo(sel, deflt);
     iComboContents[sel].push_back(deflt);
   }
+  if (kind != EGridSize && kind != EAngleSize && kind != EDashStyle
+      && kind != EOpacity) {
+    addCombo(sel, IPEABSOLUTE);
+    iComboContents[sel].push_back(IPEABSOLUTE);
+  }
   for (const auto & att : seq) {
     String s = att.string();
     addCombo(sel, s);
@@ -677,6 +705,8 @@ void AppUiBase::setupSymbolicNames(const Cascade *sheet)
   showInCombo(sheet, EPen, EUiPen);
   showInCombo(sheet, ETextSize, EUiTextSize);
   showInCombo(sheet, ESymbolSize, EUiSymbolSize);
+  showInCombo(sheet, EDashStyle, EUiDashStyle);
+  showInCombo(sheet, EOpacity, EUiOpacity);
   showMarksInCombo(sheet);
   showInCombo(sheet, EGridSize, EUiGridSize, "16pt");
   showInCombo(sheet, EAngleSize, EUiAngleSize, "45 deg");
@@ -710,7 +740,7 @@ void AppUiBase::setGridAngleSize(Attribute abs_grid, Attribute abs_angle)
 
 void AppUiBase::setAttribute(int sel, Attribute a)
 {
-  String s = a.string();
+  String s = a.isSymbolic() ? a.string() : IPEABSOLUTE;
   for (int i = 0; i < int(iComboContents[sel].size()); ++i) {
     if (iComboContents[sel][i] == s) {
       setComboCurrent(sel, i);
@@ -735,6 +765,8 @@ void AppUiBase::setAttributes(const AllAttributes &all, Cascade *sheet)
   setAttribute(EUiPen, iAll.iPen);
   setAttribute(EUiTextSize, iAll.iTextSize);
   setAttribute(EUiSymbolSize, iAll.iSymbolSize);
+  setAttribute(EUiDashStyle, iAll.iDashStyle);
+  setAttribute(EUiOpacity, iAll.iOpacity);
 
   String s = stripMark(iAll.iMarkShape);
   for (int i = 0; i < int(iComboContents[EUiMarkShape].size()); ++i) {
@@ -746,6 +778,7 @@ void AppUiBase::setAttributes(const AllAttributes &all, Cascade *sheet)
 
   setCheckMark("horizontalalignment", Attribute(iAll.iHorizontalAlignment));
   setCheckMark("verticalalignment", Attribute(iAll.iVerticalAlignment));
+  setCheckMark("splinetype", Attribute(iAll.iSplineType));
   setCheckMark("pinned", Attribute(iAll.iPinned));
   setCheckMark("transformabletext",
 	       Attribute::Boolean(iAll.iTransformableText));
