@@ -3,7 +3,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -44,28 +44,31 @@ using namespace ipelua;
 // in ipeui_cocoa:
 extern NSImage *colorIcon(double red, double green, double blue, int pixels);
 
+// in ipebitmap_unix:
+extern bool cgImageDecode(CGImageRef bitmap, Buffer pixelData);
+
 // --------------------------------------------------------------------
 
 static const char * const snapbutton_action[] = {
   "snapvtx", "snapctl", "snapbd", "snapint",
-  "snapgrid", "snapangle", "snapauto",
+  "snapgrid", "snapangle", "snapcustom", "snapauto",
   "grid_visible" };
 
 static const char * const touchbar_action[] = {
   "escape", "set_origin", "set_direction", "set_line",
   "show_axes", "reset_direction", "set_tangent_direction",
   "snapvtx", "snapctl", "snapbd", "snapint",
-  "snapgrid", "snapangle", "snapauto" };
+  "snapgrid", "snapangle", "snapcustom", "snapauto" };
 
 #if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
 static const char * const touchbar_titles[] = {
   "esc", "org", "dir", "line", "axes", "reset", "tangent",
-  nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
+  nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
 
 static const char * const touchbar_customization_label[] = {
   "Escape", "Set origin", "Set direction", "Set to edge", "Show axes", "Reset direction", "Set to tangent direction",
   "Snap to vertex", "Snap to control point", "Snap to boundary", "Snap to intersection", "Snap to grid",
-  "Angular snap", "Automatic snap" };
+  "Angular snap", "Snap to custom grid", "Automatic snap" };
 #endif
 
 @interface IpeAction : NSObject
@@ -140,7 +143,7 @@ static const char * const touchbar_customization_label[] = {
 
 // --------------------------------------------------------------------
 
-static NSImage *colorIcon(Color color, int pixels = 24)
+static NSImage *colorIcon(Color color, int pixels = 20)
 {
   double red = color.iRed.toDouble();
   double green = color.iGreen.toDouble();
@@ -342,7 +345,7 @@ static NSImage *loadIcon(String action)
 	     @"fit_width", @"keyboard",
 	     @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
@@ -359,7 +362,7 @@ static NSImage *loadIcon(String action)
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
@@ -374,7 +377,7 @@ static NSImage *loadIcon(String action)
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles2", @"mode_rectangles3",
@@ -397,7 +400,7 @@ static NSImage *loadIcon(String action)
   bar.customizationAllowedItemIdentifiers =
     @[ @"show_axes", @"set_origin", @"set_direction", @"reset_direction",
 	@"set_line", @"set_tangent_direction",
-	@"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid", @"snapangle", @"snapauto" ];
+	@"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid", @"snapangle", @"snapcustom", @"snapauto" ];
   bar.escapeKeyReplacementItemIdentifier = @"escape";
   return bar;
 }
@@ -700,9 +703,17 @@ AppUi::AppUi(lua_State *L0, int model)
   layout(iMouseIndicator, iZoomIndicator, "r=l");
   layout(iZoomIndicator, iContent, "r=r");
 
+  lua_getglobal(L, "prefs");
+  lua_getfield(L, -1, "osx_properties_width");
+  if (lua_isnumber(L, -1)) {
+    double width = lua_tonumber(L, -1);
+    layout(iLayerBox, nil, "w>0", width);
+  }
+  lua_pop(L, 2);
+
   iViewToSnapBar = layout(iView, iSnapBar, "t=b", 0.0, 1.0, NO);
   iViewToTop = layout(iView, iContent, "t=t", 0.0, 1.0, NO);
-  activateConstraint(iViewToSnapBar, YES);
+  iViewToSnapBar.active = YES;
 
   [iStatus
     setContentHuggingPriority:NSLayoutPriorityDefaultLow
@@ -735,14 +746,16 @@ void AppUi::makePropertiesTool()
   iPropertiesBox = [[NSBox alloc] initWithFrame:NSMakeRect(0.,0.,50.,300.)];
   [iPropertiesBox setTitle:@"Properties"];
 
-  for (int i = 0; i <= EUiSymbolSize; ++i) {
-    iButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0., 0., 20., 20.)];
-    [iButton[i] setButtonType:NSMomentaryPushInButton];
-    [iButton[i] setImagePosition:NSImageOnly];
-    [iButton[i] setBezelStyle:NSRegularSquareBezelStyle];
-    iButton[i].action = @selector(ipeAbsoluteButton:);
-    iButton[i].tag = i;
-    addToLayout(iPropertiesBox, iButton[i]);
+  for (int i = 0; i <= EUiOpacity; ++i) {
+    if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
+      iButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0., 0., 12., 12.)];
+      [iButton[i] setButtonType:NSMomentaryPushInButton];
+      [iButton[i] setImagePosition:NSImageOnly];
+      [iButton[i] setBezelStyle:NSRegularSquareBezelStyle];
+      iButton[i].action = @selector(ipeAbsoluteButton:);
+      iButton[i].tag = i;
+      addToLayout(iPropertiesBox, iButton[i]);
+    }
 
     iSelector[i] = [[NSPopUpButton alloc]
 		     initWithFrame:NSMakeRect(0., 0., 50., 20.)
@@ -752,10 +765,7 @@ void AppUi::makePropertiesTool()
     iSelector[i].tag = i;
     addToLayout(iPropertiesBox, iSelector[i]);
   }
-  iButton[EUiMarkShape].enabled = NO;
-  // [iButton[EUiMarkShape] setImageDimsWhenDisabled:NO];
-
-  iModeIndicator = [[NSImageView alloc] initWithFrame:NSMakeRect(0, 0, 20, 20)];
+  iModeIndicator = [[NSImageView alloc] initWithFrame:NSMakeRect(0., 0., 12., 12.)];
   iModeIndicator.editable = NO;
   iModeIndicator.image = loadIcon("mode_select");
   addToLayout(iPropertiesBox, iModeIndicator);
@@ -764,7 +774,6 @@ void AppUi::makePropertiesTool()
   [iButton[EUiFill] setImage:colorIcon(Color(1000, 1000, 0))];
   [iButton[EUiPen] setImage:loadIcon("pen")];
   [iButton[EUiTextSize] setImage:loadIcon("mode_label")];
-  [iButton[EUiMarkShape] setImage:loadIcon("mode_marks")];
   [iButton[EUiSymbolSize] setImage:loadIcon("mode_marks")];
 
   [iButton[EUiStroke] setToolTip:@"Absolute stroke color"];
@@ -779,6 +788,8 @@ void AppUi::makePropertiesTool()
   [iSelector[EUiTextSize] setToolTip:@"Symbolic text size"];
   [iSelector[EUiMarkShape] setToolTip:@"Mark shape"];
   [iSelector[EUiSymbolSize] setToolTip:@"Symbolic symbol size"];
+  [iSelector[EUiDashStyle] setToolTip:@"Dash style"];
+  [iSelector[EUiOpacity] setToolTip:@"Opacity"];
 
   iViewNumber = [[NSButton alloc] initWithFrame:NSMakeRect(0, 0, 10, 20)];
   [iViewNumber setButtonType:NSMomentaryPushInButton];
@@ -826,28 +837,50 @@ void AppUi::makePropertiesTool()
 
   NSView *inside = [iPropertiesBox contentView];
   id guide = layoutGuide(iPropertiesBox);
-  layout(iButton[EUiStroke], inside, "t=t", 2.0);
-  for (int i = 0; i <= EUiSymbolSize; ++i) {
-    layout(iButton[i], iButton[i], "w=h");
-    layout(iButton[i], inside, "l=l");
-    layout(iButton[i], guide, "r=l");
-    layout(guide, iSelector[i], "r=l");
-    layout(iSelector[i], inside, "r=r");
-    layout(iSelector[i], iButton[i], "y=y");
-    if (i != EUiStroke && i != EUiTextSize)
-      layout(iButton[i], iButton[i-1], "t=b", 2.0);
-  }
   layout(guide, inside, "t=t");
   layout(guide, inside, "b=b");
-  layout(iPathView, iButton[EUiPen], "t=b", 2.0);
-  layout(iButton[EUiTextSize], iPathView, "t=b", 2.0);
-  layout(iPathView, guide, "l=r");
-  layout(iPathView, inside, "r=r");
-  layout(iModeIndicator, iPathView, "t=t");
-  layout(iModeIndicator, iPathView, "b=b");
+
+  // left-right layout
+  for (int i = 0; i <= EUiOpacity; ++i) {
+    if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
+      layout(iButton[i], inside, "l=l");
+      layout(iButton[i], guide, "r=l");
+      layout(iButton[i], nil, "h>0", 28.);
+      layout(iButton[i], nil, "w=0", 30.);
+    }
+    layout(guide, iSelector[i], "r=l");
+    layout(iSelector[i], inside, "r=r");
+  }
   layout(iModeIndicator, inside, "l=l");
   layout(iModeIndicator, guide, "r=l");
-  layout(iViewNumber, iButton[EUiSymbolSize], "t=b", 2.0);
+  layout(iPathView, guide, "l=r");
+  layout(iPathView, inside, "r=r");
+
+  // top-down layout
+  layout(iButton[EUiStroke], inside, "t=t", 2.0);
+  layout(iSelector[EUiStroke], iButton[EUiStroke], "t=t");
+  layout(iButton[EUiFill], iButton[EUiStroke], "t=b", 2.0);
+  layout(iSelector[EUiFill], iButton[EUiFill], "y=y");
+  layout(iButton[EUiPen], iButton[EUiFill], "t=b", 2.0);
+  layout(iSelector[EUiPen], iButton[EUiPen], "t=t");
+  layout(iSelector[EUiDashStyle], iSelector[EUiPen], "t=b", 2.0);
+  layout(iSelector[EUiDashStyle], iButton[EUiPen], "b=b");
+
+  layout(iPathView, iButton[EUiPen], "t=b", 6.0);
+  layout(iModeIndicator, iPathView, "t=t");
+  layout(iModeIndicator, iPathView, "b=b");
+
+  layout(iButton[EUiTextSize], iPathView, "t=b", 2.0);
+  layout(iSelector[EUiTextSize], iButton[EUiTextSize], "y=y");
+
+  layout(iButton[EUiSymbolSize], iButton[EUiTextSize], "t=b", 2.0);
+  layout(iSelector[EUiMarkShape], iButton[EUiSymbolSize], "t=t");
+  layout(iSelector[EUiSymbolSize], iSelector[EUiMarkShape], "t=b", 2.0);
+  layout(iSelector[EUiSymbolSize], iButton[EUiSymbolSize], "b=b");
+
+  layout(iSelector[EUiOpacity], iButton[EUiSymbolSize], "t=b", 2.0);
+
+  layout(iViewNumber, iSelector[EUiOpacity], "t=b", 2.0);
   layout(iViewNumber, inside, "l=l");
   layout(iPageNumber, iViewNumber, "l=r", 2.0);
   layout(iPageNumber, iViewNumber, "t=t");
@@ -860,9 +893,9 @@ void AppUi::makePropertiesTool()
   layout(inside, iViewMarked, "b=b", 5.0);
 
   layout(guide, nil, "w=0", 5.0);
-  layout(iPathView, iPathView, "h=w", 0.0, 0.3);
+  layout(iPathView, nil, "h>0", 40.0);
+  //layout(iPathView, iPathView, "h=w", 0.0, 0.3);
   layout(iViewNumber, iPageNumber, "w=w");
-  layout(iModeIndicator, iModeIndicator, "w=h");
 }
 
 void AppUi::makeSnapBar()
@@ -872,7 +905,7 @@ void AppUi::makeSnapBar()
 
   iSnapBar = [[NSView alloc] initWithFrame:NSMakeRect(0., 0., 600., 32.)];
 
-  for (int i = 0; i < 8; ++i) {
+  for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
     iSnapButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0, 0, 32, 32)];
     iSnapButton[i].buttonType = NSPushOnPushOffButton;
     iSnapButton[i].imagePosition = NSImageOnly;
@@ -904,7 +937,8 @@ void AppUi::makeSnapBar()
   iSelector[EUiAngleSize].tag = EUiAngleSize;
   addToLayout(iSnapBar, iSelector[EUiAngleSize]);
 
-  for (int i = 0; i < 8; ++i) {
+  // Order: "vtx", "ctl", "bd", "int", "grid", "angle", "custom", "auto", "visible"
+  for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
     layout(iSnapButton[i], iSnapBar, "t=t", MARGIN);
     layout(iSnapBar, iSnapButton[i], "b=b", MARGIN);
     if (0 < i && i < 5)
@@ -922,8 +956,9 @@ void AppUi::makeSnapBar()
   layout(iSnapButton[5], iSelector[EUiGridSize], "l=r", PAD);
   layout(iSelector[EUiAngleSize], iSnapButton[5], "l=r", PAD);
   layout(iSnapButton[6], iSelector[EUiAngleSize], "l=r", PAD);
-  layout(iSnapButton[7], iSnapButton[6], "l>r", PAD);
-  layout(iSnapBar, iSnapButton[7], "r=r", MARGIN);
+  layout(iSnapButton[7], iSnapButton[6], "l=r", PAD);
+  layout(iSnapButton[8], iSnapButton[7], "l>r", PAD);
+  layout(iSnapBar, iSnapButton[8], "r=r", MARGIN);
 }
 
 // --------------------------------------------------------------------
@@ -939,6 +974,10 @@ void AppUi::resetCombos()
 void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 {
   iInUiUpdate = true;
+  [iSelector[EUiStroke] addItemWithTitle:C2N(IPEABSOLUTE)];
+  [iSelector[EUiFill] addItemWithTitle:C2N(IPEABSOLUTE)];
+  iComboContents[EUiStroke].push_back(IPEABSOLUTE);
+  iComboContents[EUiFill].push_back(IPEABSOLUTE);
   for (uint i = 0; i < sym.size(); ++i) {
     Color color = abs[i].color();
     NSImage *im = colorIcon(color, 12);
@@ -1093,11 +1132,11 @@ void AppUi::toggleSnapbarShown()
 {
   iSnapBar.hidden = ![iSnapBar isHidden];
   if ([iSnapBar isHidden]) {
-    activateConstraint(iViewToSnapBar, NO);
-    activateConstraint(iViewToTop, YES);
+    iViewToSnapBar.active = NO;
+    iViewToTop.active = YES;
   } else {
-    activateConstraint(iViewToTop, NO);
-    activateConstraint(iViewToSnapBar, YES);
+    iViewToTop.active = NO;
+    iViewToSnapBar.active = YES;
   }
   iCanvas->update();
 }
@@ -1240,7 +1279,7 @@ void AppUi::setActionState(const char *name, bool value)
     }
   } else if ([a hasPrefix:@"snap"] || [a isEqualToString:@"grid_visible"] ||
 	     [a isEqualToString:@"show_axes"]) {
-    for (int i = 0; i < 8; ++i) {
+    for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
       if ([a isEqualToString:C2N(snapbutton_action[i])])
 	iSnapButton[i].state = value ? NSOnState : NSOffState;
     }
@@ -1248,12 +1287,15 @@ void AppUi::setActionState(const char *name, bool value)
 #if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
 #pragma clang diagnostic push
 #pragma clang diagnostic ignored "-Wdeprecated-declarations"
-    if (iDelegate.touchBar != nil) {
-      auto bar = (NSTouchBar *) iDelegate.touchBar;
-      auto item = [bar itemForIdentifier:a];
-      if (item != nil) {
-	auto b = (NSButton *) item.view;
-	b.state =  value ? NSOnState : NSOffState;
+    if ([iDelegate respondsToSelector:@selector(touchBar)]) {
+      id tbar = [iDelegate performSelector:@selector(touchBar)];
+      if (tbar != nil) {
+	auto bar = (NSTouchBar *) tbar;
+	auto item = [bar itemForIdentifier:a];
+	if (item != nil) {
+	  auto b = (NSButton *) item.view;
+	  b.state =  value ? NSOnState : NSOffState;
+	}
       }
     }
 #pragma clang diagnostic push
@@ -1267,7 +1309,7 @@ void AppUi::setActionState(const char *name, bool value)
 
 void AppUi::action(String name)
 {
-  ipeDebug("action %s", name.z());
+  // ipeDebug("action %s", name.z());
   if (name == "escape") {
     if (iCanvas && iCanvas->tool())
       iCanvas->tool()->key("\033", 0);
@@ -1379,10 +1421,25 @@ void AppUi::closeWindow()
 
 int AppUi::clipboard(lua_State *L)
 {
-  // bool allow_bitmap = lua_toboolean(L, 2);
-  // NSImage can be read in the same way, but I don't
-  // know how to convert it to a bitmap...
   NSPasteboard *pasteBoard = [NSPasteboard generalPasteboard];
+  bool allow_bitmap = lua_toboolean(L, 2);
+  if (allow_bitmap) {
+    NSArray *arr = [pasteBoard readObjectsForClasses:@[[NSImage class]]
+					     options:@{ } ];
+    if (arr && [arr count] > 0) {
+      CGImageRef bitmap = [arr[0] CGImageForProposedRect:nil context:nil hints:nil];
+      int w = CGImageGetWidth(bitmap);
+      int h = CGImageGetHeight(bitmap);
+      Buffer data(w * h * sizeof(uint32_t));
+      if (cgImageDecode(bitmap, data)) {
+	Bitmap bitmap(w, h, Bitmap::ENative, data);
+	ipe::Rect r(Vector::ZERO, Vector(w, h));
+	Image *im = new Image(r, bitmap);
+	push_object(L, im);
+	return 1;
+      }
+    }
+  }
   NSArray *arr = [pasteBoard readObjectsForClasses:@[[NSString class]]
 					   options:@{ } ];
   if (arr && [arr count] > 0) {
