@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -68,7 +68,7 @@ public:
 
   virtual void setBookmarks(int no, const String *s) override;
   virtual void setToolVisible(int m, bool vis) override;
-  virtual int pageSorter(lua_State *L, Document *doc,
+  virtual int pageSorter(lua_State *L, Document *doc, int pno,
 			 int width, int height, int thumbWidth) override;
 
   virtual int clipboard(lua_State *L) override;
@@ -112,6 +112,8 @@ private:
   virtual void setButtonColor(int sel, Color color) override;
 
 private:
+  static const int NUMSNAPBUTTONS = 9;
+
   NSMutableDictionary <NSString *, IpeAction *> *iActions;
   bool iActionsEnabled;
   bool iInUiUpdate;
@@ -130,7 +132,7 @@ private:
   IpeLayerView *iLayerView;
   std::vector<String> iLayerNames;
 
-  NSButton *iButton[EUiGridSize];
+  NSButton *iButton[EUiOpacity]; // EUiDashStyle and EUiMarkShape are not used
   NSPopUpButton *iSelector[EUiView];
 
   NSImageView *iModeIndicator;
@@ -140,7 +142,7 @@ private:
   NSButton *iPageMarked;
 
   NSView *iSnapBar;
-  NSButton *iSnapButton[8];
+  NSButton *iSnapButton[NUMSNAPBUTTONS];
   NSLayoutConstraint *iViewToTop;
   NSLayoutConstraint *iViewToSnapBar;
 
