@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -320,6 +320,7 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   iSnapTools->addWidget(iSelector[EUiGridSize]);
   addSnap("snapangle");
   iSnapTools->addWidget(iSelector[EUiAngleSize]);
+  addSnap("snapcustom");
   addSnap("snapauto");
 
   addEdit("copy");
@@ -389,18 +390,42 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   QButtonGroup *bg = new QButtonGroup(wg);
   bg->setExclusive(false);
   connect(bg, SIGNAL(buttonClicked(int)), this, SLOT(absoluteButton(int)));
-  iButton[EUiMarkShape] = nullptr;   // no such button
-  for (int i = 0; i <= EUiSymbolSize; ++i) {
-    if (i != EUiMarkShape) {
+  iButton[EUiMarkShape] = nullptr;   // no such buttons
+  iButton[EUiDashStyle] = nullptr;
+  /* enum { EUiStroke, EUiFill, EUiPen, EUiDashStyle,
+     EUiTextSize, EUiMarkShape, EUiSymbolSize, EUiOpacity ... }
+     layout:
+     0: stroke
+     1: fill
+     2: pen       \   pen button
+     3: dashstyle /
+     4: pathview
+     5: textsize
+     6: markshape   \ symbolsize button
+     7: symbolsize  /
+     8: opacity
+   */
+  for (int i = 0; i <= EUiOpacity; ++i) {
+    if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
       iButton[i] = new QToolButton();
       iButton[i]->setFocusPolicy(Qt::NoFocus);
+      iButton[i]->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding,
+					    QSizePolicy::ToolButton));
       bg->addButton(iButton[i], i);
-      lo->addWidget(iButton[i], i > 2 ? i+1 : i, 0);
+      if (i == EUiPen)
+	lo->addWidget(iButton[i], i, 0, 2, 1);
+      else if (i == EUiSymbolSize)
+	lo->addWidget(iButton[i], i, 0, 2, 1);
+      else
+	lo->addWidget(iButton[i], i >= EUiTextSize ? i+1 : i, 0);
       iButton[i]->setIconSize(adapt_size(iButton[i]->iconSize(), iUiScale));
     }
     iSelector[i] = new QComboBox();
     iSelector[i]->setFocusPolicy(Qt::NoFocus);
-    lo->addWidget(iSelector[i], i > 2 ? i+1 : i, 1);
+    if (i == EUiOpacity)
+      lo->addWidget(iSelector[i], i+1, 0, 1, 2);
+    else
+      lo->addWidget(iSelector[i], i >= EUiTextSize ? i+1 : i, 1);
     connect(iSelector[i], SIGNAL(activated(int)), comboMap, SLOT(map()));
     comboMap->setMapping(iSelector[i], i);
   }
@@ -422,27 +447,24 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   iSelector[EUiTextSize]->setToolTip("Symbolic text size");
   iSelector[EUiMarkShape]->setToolTip("Mark shape");
   iSelector[EUiSymbolSize]->setToolTip("Symbolic symbol size");
+  iSelector[EUiDashStyle]->setToolTip("Dash style");
+  iSelector[EUiOpacity]->setToolTip("Opacity");
 
   iSelector[EUiGridSize]->setToolTip("Grid size");
   iSelector[EUiAngleSize]->setToolTip("Angle for angular snap");
 
   connect(comboMap, SIGNAL(mapped(int)), this, SLOT(comboSelector(int)));
 
-  QLabel *msl = new QLabel();
-  msl->setPixmap(prefsPixmap("mode_marks"));
-  lo->addWidget(msl, EUiMarkShape + 1, 0);
-
-  iPathView = new PathView();
+  iPathView = new PathView(iUiScale);
   connect(iPathView, SIGNAL(activated(String)), SLOT(action(String)));
   connect(iPathView, SIGNAL(showPathStylePopup(Vector)),
 	  SLOT(showPathStylePopup(Vector)));
-  lo->addWidget(iPathView, 3, 1, 1, 1);
-  // lo->setRowStretch(3, 1);
+  lo->addWidget(iPathView, 4, 1);
   iPropertiesTools->setWidget(wg);
 
   iModeIndicator = new QLabel();
   iModeIndicator->setPixmap(prefsPixmap("mode_select"));
-  lo->addWidget(iModeIndicator, 3, 0);
+  lo->addWidget(iModeIndicator, 4, 0);
 
   QHBoxLayout *hol = new QHBoxLayout();
   iViewNumber = new QToolButton();
@@ -467,7 +489,7 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   hol->addStretch(1);
   hol->addWidget(iPageMarked);
   hol->addWidget(iPageNumber);
-  lo->addLayout(hol, EUiSymbolSize + 2, 0, 1, -1);
+  lo->addLayout(hol, EUiOpacity + 2, 0, 1, -1);
 
   iPageNotes = new QTextEdit();
   iPageNotes->setAcceptRichText(false);
@@ -651,12 +673,16 @@ void AppUi::abortDrawing()
 
 void AppUi::resetCombos()
 {
-  for (int i = 0; i < EUiView; ++i)
+  for (int i = 0; i < EUiOpacity; ++i)
     iSelector[i]->clear();
 }
 
 void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 {
+  iSelector[EUiStroke]->addItem(IPEABSOLUTE);
+  iSelector[EUiFill]->addItem(IPEABSOLUTE);
+  iComboContents[EUiStroke].push_back(IPEABSOLUTE);
+  iComboContents[EUiFill].push_back(IPEABSOLUTE);
   for (uint i = 0; i < sym.size(); ++i) {
     Color color = abs[i].color();
     String s = sym[i].string();
@@ -874,8 +900,10 @@ static const char * const aboutText =
   "<a href=\"http://www.facebook.com/drawing.editor.Ipe7\">Facebook</a>.</p>"
   "<p>If you are an Ipe fan and want to show others, have a look at the "
   "<a href=\"https://www.shirtee.com/en/store/ipe\">Ipe T-shirts</a>.</p>"
-  "<h3>Platinum sponsors</h3>"
+  "<h3>Platinum and gold sponsors</h3>"
   "<ul><li>Hee-Kap Ahn</li>"
+  "<li>Günter Rote</li>"
+  "<li>SCALGO</li>"
   "<li>Martin Ziegler</li></ul>"
   "<p>If you enjoy Ipe, you can become a member of the exclusive community of "
   "<a href=\"http://patreon.com/otfried\">Ipe patrons</a>. "
@@ -921,14 +949,17 @@ void AppUi::action(String name)
 
 // --------------------------------------------------------------------
 
-int AppUi::pageSorter(lua_State *L, Document *doc,
+int AppUi::pageSorter(lua_State *L, Document *doc, int pno,
 		      int width, int height, int thumbWidth)
 {
   QDialog *d = new QDialog();
-  d->setWindowTitle("Ipe Page Sorter");
+  if (pno >= 0)
+    d->setWindowTitle("Ipe View Sorter");
+  else
+    d->setWindowTitle("Ipe Page Sorter");
 
   QLayout *lo = new QVBoxLayout;
-  PageSorter *p = new PageSorter(doc, thumbWidth);
+  PageSorter *p = new PageSorter(doc, pno, thumbWidth);
   QDialogButtonBox *buttonBox =
     new QDialogButtonBox(QDialogButtonBox::Ok|QDialogButtonBox::Cancel);
   lo->addWidget(p);
@@ -950,8 +981,14 @@ int AppUi::pageSorter(lua_State *L, Document *doc,
     lua_pushinteger(L, p->pageAt(i-1) + 1);
     lua_rawseti(L, -2, i);
   }
+  int m = p->iMarks.size();
+  lua_createtable(L, m, 0);
+  for (int i = 1; i <= m; ++i) {
+    lua_pushboolean(L, p->iMarks[i-1]);
+    lua_rawseti(L, -2, i);
+  }
   delete d;
-  return 1;
+  return 2;
 }
 
 // --------------------------------------------------------------------
