@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -86,7 +86,7 @@ public:
   virtual void showWindow(int width, int height) override;
   virtual void setBookmarks(int no, const String *s) override;
   virtual void setToolVisible(int m, bool vis) override;
-  virtual int pageSorter(lua_State *L, Document *doc,
+  virtual int pageSorter(lua_State *L, Document *doc, int pno,
 			 int width, int height, int thumbWidth) override;
   virtual int clipboard(lua_State *L) override;
   virtual int setClipboard(lua_State *L) override;
@@ -155,7 +155,7 @@ private:
 
   QMenu *iMenu[ENumMenu];
 
-  QToolButton *iButton[EUiGridSize];
+  QToolButton *iButton[EUiOpacity]; // EUiDashStyle and EUiMarkShape are not used
   QComboBox *iSelector[EUiView];
 
   QToolButton *iViewNumber;
