@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -60,10 +60,17 @@
 using namespace ipe;
 using namespace ipelua;
 
+#define WM_DPICHANGED       0x02E0
+
 // --------------------------------------------------------------------
 
 const int TBICONSIZE = 24;
 
+const int GRIDSIZE_WIDTH = 160;
+const int ANGLESIZE_WIDTH = 100;
+const int COLORBUTTON_SIZE = 22;
+const int COLORICON_SIZE = 18;
+
 static HINSTANCE win_hInstance = nullptr;
 static int win_nCmdShow = 0;
 
@@ -163,7 +170,7 @@ static HBITMAP loadIcon(String action, int w, int h, int r0, int g0, int b0)
 
 static int loadIcon(String action, HIMAGELIST il, int scale)
 {
-  int size = scale * TBICONSIZE / 100;
+  int size = scale * TBICONSIZE / 9600;
   HBITMAP bm = loadIcon(action, size, size, -1, -1, -1);
   if (bm == nullptr)
     return -1;
@@ -174,7 +181,7 @@ static int loadIcon(String action, HIMAGELIST il, int scale)
 
 static HBITMAP loadButtonIcon(String action, int scale)
 {
-  int size = scale * 24 / 100;
+  int size = scale * 24 / 9600;
   DWORD rgb = GetSysColor(COLOR_BTNFACE);
   int r0 = GetRValue(rgb);
   int g0 = GetGValue(rgb);
@@ -182,20 +189,18 @@ static HBITMAP loadButtonIcon(String action, int scale)
   return loadIcon(action, -size, -1, r0, g0, b0);
 }
 
-static HBITMAP colorIcon(Color color, int scale)
+static HBITMAP colorIcon(Color color, int size)
 {
   int r = int(color.iRed.internal() * 255 / 1000);
   int g = int(color.iGreen.internal() * 255 / 1000);
   int b = int(color.iBlue.internal() * 255 / 1000);
   COLORREF rgb = RGB(r, g, b);
-  int cx = scale * 22 / 100;
-  int cy = scale * 22 / 100;
   HDC hdc = GetDC(nullptr);
   HDC memDC = CreateCompatibleDC(hdc);
-  HBITMAP bm = CreateCompatibleBitmap(hdc, cx, cy);
+  HBITMAP bm = CreateCompatibleBitmap(hdc, size, size);
   SelectObject(memDC, bm);
-  for (int y = 0; y < cy; ++y) {
-    for (int x = 0; x < cx; ++x) {
+  for (int y = 0; y < size; ++y) {
+    for (int x = 0; x < size; ++x) {
       SetPixel(memDC, x, y, rgb);
     }
   }
@@ -217,8 +222,8 @@ void AppUi::addRootMenu(int id, const char *name)
 void AppUi::createAction(String name, String tooltip, bool canWhileDrawing)
 {
   SAction s;
-  s.name = String(name);
-  s.icon = loadIcon(name, hIcons, iToolbarScale);
+  s.name = name;
+  s.icon = loadIcon(name, hIcons, iDpi * iToolbarScale);
   s.tooltip = tooltip;
   s.alwaysOn = canWhileDrawing;
   iActions.push_back(s);
@@ -316,6 +321,7 @@ void AppUi::addTButton(HWND tb, const char *name, int flags)
     WString tt(iActions[i].tooltip);
     tbb.iString = INT_PTR(tt.data());
     SendMessageW(tb, TB_ADDBUTTONS, 1, (LPARAM) &tbb);
+    ++iToolButtonCount;
   }
 }
 
@@ -384,6 +390,8 @@ void AppUi::toggleFullscreen()
 AppUi::AppUi(lua_State *L0, int model)
   : AppUiBase(L0, model)
 {
+  iDpi = 96;  // default
+
   // windows procedure may receive WM_SIZE message before canvas is created
   iCanvas = nullptr;
   HWND hwnd = CreateWindowExW(WS_EX_CLIENTEDGE, className, L"Ipe",
@@ -413,15 +421,12 @@ static void insert_tb(HWND hRebar, REBARBANDINFO &rbBand, HWND tb, int size)
 void AppUi::initUi()
 {
   // update scale if dpi information is available
-  int dpi = Canvas::getDpiForWindow(hwnd);
-  if (iUiScale == 100)
-    iUiScale = dpi;
-  if (iToolbarScale == 100)
-    iToolbarScale = dpi;
+  iDpi = Canvas::getDpiForWindow(hwnd);
+  ipeDebug("DPI and scaling: %d %d %d", iDpi, iUiScale, iToolbarScale);
 
-  int tbsize = iToolbarScale * TBICONSIZE / 100;
+  int tbsize = iDpi * iToolbarScale * TBICONSIZE / 9600;
   hIcons = ImageList_Create(tbsize, tbsize, ILC_COLOR32, 20, 4);
-  hColorIcons = ImageList_Create(uiscale(22), uiscale(22), ILC_COLOR32, 20, 4);
+  hColorIcons = ImageList_Create(uiscale(COLORICON_SIZE), uiscale(COLORICON_SIZE), ILC_COLOR32, 20, 4);
   HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hwnd, GWLP_HINSTANCE);
 
   hMenuBar = CreateMenu();
@@ -443,6 +448,7 @@ void AppUi::initUi()
 
   // ------------------------------------------------------------
 
+  iToolButtonCount = 0;
   hSnapTools = createToolBar(hInst);
 
   addTButton(hSnapTools, "snapvtx", BTNS_CHECK);
@@ -452,17 +458,17 @@ void AppUi::initUi()
   addTButton(hSnapTools, "snapgrid", BTNS_CHECK);
 
   addTButton(hSnapTools, "snapangle", BTNS_CHECK);
+  addTButton(hSnapTools, "snapcustom", BTNS_CHECK);
   addTButton(hSnapTools, "snapauto", BTNS_CHECK);
 
-#define SNAP_BUTTONS 7
+  iSnapButtons = iToolButtonCount;
 
+  iToolButtonCount = 0;
   hEditTools = createToolBar(hInst);
-  int EDIT_BUTTONS = 0;
 
   if (iLeftDockFloats) {
     createAction("dock_visible", "Show properties and layers", true);
     addTButton(hEditTools, "dock_visible", BTNS_CHECK);
-    EDIT_BUTTONS += 1;
   }
 
   if (!isMiniEdit) {
@@ -470,11 +476,9 @@ void AppUi::initUi()
     addTButton(hEditTools, "cut");
     addTButton(hEditTools, "paste");
     addTButton(hEditTools, "delete");
-    EDIT_BUTTONS += 4;
   }
   addTButton(hEditTools, "undo");
   addTButton(hEditTools, "redo");
-  EDIT_BUTTONS += 2;
   if (!isMiniEdit) {
     addTButton(hEditTools, "zoom_in");
     addTButton(hEditTools, "zoom_out");
@@ -482,15 +486,16 @@ void AppUi::initUi()
     addTButton(hEditTools, "fit_page");
     addTButton(hEditTools, "fit_width");
     addTButton(hEditTools, "keyboard");
-    EDIT_BUTTONS += 6;
   }
   createAction("shift_key", "Press the Shift key", true);
   addTButton(hEditTools, "shift_key", BTNS_CHECK);
   addTButton(hEditTools, "grid_visible", BTNS_CHECK);
   createAction("stop", "Abort object being drawn [Esc]", true);
   addTButton(hEditTools, "stop");
-  EDIT_BUTTONS += 3;
 
+  iEditButtons = iToolButtonCount;
+
+  iToolButtonCount = 0;
   hObjectTools = createToolBar(hInst);
 
   addTButton(hObjectTools, "mode_select", BTNS_CHECKGROUP);
@@ -501,6 +506,7 @@ void AppUi::initUi()
   addTButton(hObjectTools, "mode_graph", BTNS_CHECKGROUP);
   addTButton(hObjectTools, "mode_pan", BTNS_CHECKGROUP);
   addTButton(hObjectTools, "mode_shredder", BTNS_CHECKGROUP);
+  addTButton(hObjectTools, "mode_laser", BTNS_CHECKGROUP);
   addTButton(hObjectTools, nullptr, BTNS_CHECKGROUP);
   addTButton(hObjectTools, "mode_label", BTNS_CHECKGROUP);
   addTButton(hObjectTools, "mode_math", BTNS_CHECKGROUP);
@@ -522,7 +528,7 @@ void AppUi::initUi()
   addTButton(hObjectTools, "mode_circle3", BTNS_CHECKGROUP);
   addTButton(hObjectTools, "mode_ink", BTNS_CHECKGROUP);
 
-#define MODE_BUTTONS 24
+  iObjectButtons = iToolButtonCount;
 
   // size is used for dropdown list!  Don't set to zero.
   for (int i = EUiGridSize; i <= EUiAngleSize; ++i)
@@ -557,7 +563,9 @@ void AppUi::initUi()
   int bw = LOWORD(dwBtnSize);
 
   lua_getglobal(L, "prefs");
-  lua_getfield(L, -1, "toolbar_order");
+  lua_getfield(L, -1, "win_toolbar_order");
+  bool showGridSizeCombo = false;
+  bool showAngleSizeCombo = false;
   if (lua_istable(L, -1)) {
     int n = lua_rawlen(L, -1);
     for (int i = 1; i <= n; ++i) {
@@ -565,20 +573,26 @@ void AppUi::initUi()
       if (lua_isstring(L, -1)) {
 	String s = lua_tolstring(L, -1, nullptr);
 	if (s == "edit")
-	  insert_tb(hRebar, rbBand, hEditTools, EDIT_BUTTONS * bw);
-	else if (s == "grid")
-	  insert_tb(hRebar, rbBand, hSelector[EUiGridSize], 160);
-	else if (s == "angle")
-	  insert_tb(hRebar, rbBand, hSelector[EUiAngleSize], 100);
-	else if (s == "snap")
-	  insert_tb(hRebar, rbBand, hSnapTools, SNAP_BUTTONS * bw);
+	  insert_tb(hRebar, rbBand, hEditTools, iEditButtons * bw);
+	else if (s == "grid") {
+	  insert_tb(hRebar, rbBand, hSelector[EUiGridSize], uiscale(GRIDSIZE_WIDTH));
+	  showGridSizeCombo = true;
+	} else if (s == "angle") {
+	  insert_tb(hRebar, rbBand, hSelector[EUiAngleSize], uiscale(ANGLESIZE_WIDTH));
+	  showAngleSizeCombo = true;
+	} else if (s == "snap")
+	  insert_tb(hRebar, rbBand, hSnapTools, iSnapButtons * bw);
 	else if (s == "mode")
-	  insert_tb(hRebar, rbBand, hObjectTools, MODE_BUTTONS * bw);
+	  insert_tb(hRebar, rbBand, hObjectTools, iObjectButtons * bw);
       }
       lua_pop(L, 1);
     }
   }
-  lua_pop(L, 2); // toolbar_order, prefs
+  lua_pop(L, 2); // win_toolbar_order, prefs
+  if (!showGridSizeCombo)
+    ShowWindow(hSelector[EUiGridSize], SW_HIDE);
+  if (!showAngleSizeCombo)
+    ShowWindow(hSelector[EUiAngleSize], SW_HIDE);
 
   // ------------------------------------------------------------
 
@@ -587,32 +601,19 @@ void AppUi::initUi()
 				0, 100, 200, 280, hwnd, nullptr, hInst, nullptr);
 
   for (int i = 0; i <= EUiSymbolSize; ++i) {
-    if (i != EUiMarkShape)
+    if (i != EUiDashStyle && i != EUiMarkShape)
       hButton[i] = createButton(hInst, ID_ABSOLUTE_BASE + i);
+    else
+      hButton[i] = nullptr; // unused
   }
+
   // size is used for dropdown list!  Don't set to zero.
   for (int i = 0; i < EUiGridSize; ++i)
     hSelector[i] = CreateWindowExW(0, WC_COMBOBOXEXW, nullptr,
 				   WS_CHILD | WS_VISIBLE | CBS_DROPDOWNLIST,
 				   0, 0, uiscale(100), uiscale(300), hwnd,
 				   (HMENU) (INT_PTR) (ID_SELECTOR_BASE + i), hInst, nullptr);
-  SendMessage(hButton[EUiPen], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) loadButtonIcon("pen", iUiScale));
-  SendMessage(hButton[EUiTextSize], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) loadButtonIcon("mode_label", iUiScale));
-  SendMessage(hButton[EUiSymbolSize], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) loadButtonIcon("mode_marks", iUiScale));
-  SendMessage(hButton[EUiStroke], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) colorIcon(Color(1000, 0, 0), iUiScale));
-  SendMessage(hButton[EUiFill], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) colorIcon(Color(1000, 1000, 0), iUiScale));
-
-  hButton[EUiMarkShape] =
-    CreateWindowExW(0, L"static", nullptr,
-		    WS_CHILD | WS_VISIBLE | SS_BITMAP | SS_CENTERIMAGE,
-		    0, 0, 0, 0, hwnd, nullptr, hInst, nullptr);
-  SendMessage(hButton[EUiMarkShape], STM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) loadButtonIcon("mode_marks", iUiScale));
+  setButtonIcons();
 
   iPathView = new PathView(hwnd, ID_PATHVIEW);
 
@@ -676,6 +677,8 @@ void AppUi::initUi()
   setTooltip(hSelector[EUiTextSize], "Symbolic text size", true);
   setTooltip(hSelector[EUiMarkShape], "Mark shape", true);
   setTooltip(hSelector[EUiSymbolSize], "Symbolic symbol size", true);
+  setTooltip(hSelector[EUiDashStyle], "Dash style", true);
+  setTooltip(hSelector[EUiOpacity], "Opacity", true);
 
   setTooltip(hSelector[EUiGridSize], "Grid size", true);
   setTooltip(hSelector[EUiAngleSize], "Angle for angular snap", true);
@@ -692,6 +695,42 @@ void AppUi::initUi()
 
   SetFocus(hwndCanvas);
 
+  createFont();
+
+  // default is on, Lua code will turn off if necessary
+  CheckMenuItem(hMenuBar, actionId("toggle_notes"), MF_CHECKED);
+  CheckMenuItem(hMenuBar, actionId("toggle_bookmarks"), MF_CHECKED);
+
+  if (iLeftDockFloats)
+    setLeftDockVisibility(false);
+}
+
+AppUi::~AppUi()
+{
+  // canvas is deleted by Windows
+  ImageList_Destroy(hIcons);
+  ImageList_Destroy(hColorIcons);
+  KillTimer(hwnd, ID_STATUS_TIMER);
+  DeleteObject(hFont);
+}
+
+void AppUi::setButtonIcons()
+{
+  // TODO: should we destroy the old icons first when there is a dpi change?
+  SendMessage(hButton[EUiPen], BM_SETIMAGE, IMAGE_BITMAP,
+	      (LPARAM) loadButtonIcon("pen", iDpi * iUiScale));
+  SendMessage(hButton[EUiTextSize], BM_SETIMAGE, IMAGE_BITMAP,
+	      (LPARAM) loadButtonIcon("mode_label", iDpi * iUiScale));
+  SendMessage(hButton[EUiSymbolSize], BM_SETIMAGE, IMAGE_BITMAP,
+	      (LPARAM) loadButtonIcon("mode_marks", iDpi * iUiScale));
+  SendMessage(hButton[EUiStroke], BM_SETIMAGE, IMAGE_BITMAP,
+	      (LPARAM) colorIcon(Color(1000, 0, 0), uiscale(COLORBUTTON_SIZE)));
+  SendMessage(hButton[EUiFill], BM_SETIMAGE, IMAGE_BITMAP,
+	      (LPARAM) colorIcon(Color(1000, 1000, 0), uiscale(COLORBUTTON_SIZE)));
+}
+
+void AppUi::createFont()
+{
   hFont = CreateFontW(uiscale(18), 0, 0, 0, FW_DONTCARE,
 		      FALSE, FALSE, FALSE, ANSI_CHARSET,
 		      OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
@@ -708,23 +747,18 @@ void AppUi::initUi()
     SendMessage(hViewNumber, WM_SETFONT, WPARAM (hFont), TRUE);
     SendMessage(hPageNumber, WM_SETFONT, WPARAM (hFont), TRUE);
   }
-  // default is on, Lua code will turn off if necessary
-  CheckMenuItem(hMenuBar, actionId("toggle_notes"), MF_CHECKED);
-  CheckMenuItem(hMenuBar, actionId("toggle_bookmarks"), MF_CHECKED);
-
-  if (iLeftDockFloats)
-    // SendMessage(hEditTools, TB_SETSTATE, actionId("dock_visible"),
-    // (LPARAM) (TBSTATE_ENABLED | TBSTATE_CHECKED));
-    setLeftDockVisibility(false);
 }
 
-AppUi::~AppUi()
+void AppUi::createColorIcons()
 {
-  // canvas is deleted by Windows
-  ImageList_Destroy(hIcons);
-  ImageList_Destroy(hColorIcons);
-  KillTimer(hwnd, ID_STATUS_TIMER);
-  DeleteObject(hFont);
+  ImageList_Remove(hColorIcons, -1);
+  for (const auto &color : iColorIcons) {
+    HBITMAP bm = colorIcon(color, uiscale(COLORICON_SIZE));
+    ImageList_Add(hColorIcons, bm, nullptr);
+    DeleteObject(bm);
+  }
+  SendMessage(hSelector[EUiStroke], CBEM_SETIMAGELIST, 0, (LPARAM) hColorIcons);
+  SendMessage(hSelector[EUiFill], CBEM_SETIMAGELIST, 0, (LPARAM) hColorIcons);
 }
 
 void AppUi::layoutChildren(bool resizeRebar)
@@ -750,18 +784,27 @@ void AppUi::layoutChildren(bool resizeRebar)
     InvalidateRect(hwnd, &rect, TRUE);
 
     int dy = 28 + iUiGap;
-    int pbh = uiscale(dy * 6 + 28 + 70);
+    int pbh = uiscale(dy * 8 + 28 + 70);
     MoveWindow(hProperties, uiscale(8), tbh + uiscale(2),
 	       uiscale(180), pbh, TRUE);
-    for (int i = 0; i <= EUiSymbolSize; ++i) {
-      int y = uiscale(24 + dy * i + (i > 2 ? 42 : 0));
-      MoveWindow(hButton[i], uiscale(16), tbh + y,
-		 uiscale(26), uiscale(26), TRUE);
+    for (int i = 0; i <= EUiOpacity; ++i) {
+      int y0 = uiscale(24 + dy * (i-1) + 42);
+      int y = uiscale(24 + dy * i + (i >= EUiTextSize ? 42 : 0));
+      if (i == EUiPen) {
+	MoveWindow(hButton[i], uiscale(16), tbh + y,
+		   uiscale(28), uiscale(22 + dy), TRUE);
+      } else if (i == EUiSymbolSize) {
+	MoveWindow(hButton[i], uiscale(16), tbh + y0,
+		   uiscale(28), uiscale(22 + dy), TRUE);
+      } else if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
+	MoveWindow(hButton[i], uiscale(16), tbh + y,
+		   uiscale(28), uiscale(26), TRUE);
+      }
       MoveWindow(hSelector[i], uiscale(50), tbh + y,
 		 uiscale(132), uiscale(26), TRUE);
     }
 
-    int y = tbh + uiscale(24 + dy * 6 + 42);
+    int y = tbh + uiscale(24 + dy * 8 + 42);
     MoveWindow(hViewNumber, uiscale(16), y, uiscale(68), uiscale(26), TRUE);
     MoveWindow(hPageNumber, uiscale(116), y, uiscale(68), uiscale(26), TRUE);
     MoveWindow(hViewMarked, uiscale(85), y + uiscale(6),
@@ -770,7 +813,7 @@ void AppUi::layoutChildren(bool resizeRebar)
 	       uiscale(13), uiscale(13), TRUE);
 
     MoveWindow(iPathView->windowId(), uiscale(16),
-	       tbh + uiscale(24 + dy * 2 + 28 + iUiGap / 2),
+	       tbh + uiscale(24 + dy * 3 + 28 + iUiGap / 2),
 	       uiscale(168), uiscale(40), TRUE);
 
     MoveWindow(hLayerGroup, uiscale(8), tbh + uiscale(6) + pbh,
@@ -817,8 +860,93 @@ void AppUi::layoutChildren(bool resizeRebar)
   BringWindowToTop(hProperties);
 }
 
+void AppUi::handleDpiChange(HWND hwnd, WPARAM wParam, LPARAM lParam)
+{
+  int oldDpi = iDpi;
+  iDpi = HIWORD(wParam);
+  ipeDebug("DPI change: %d -> %d", oldDpi, iDpi);
+
+  setButtonIcons();
+  DeleteObject(hFont);
+  createFont();
+
+  // need to reload toolbar buttons at right size
+  int n = ImageList_GetImageCount(hIcons);
+  std::vector<String> iconNames(n);
+  for (const auto &a : iActions) {
+    if (a.icon >= 0)
+      iconNames[a.icon] = a.name;
+  }
+  ImageList_Destroy(hIcons);
+  int tbsize = iDpi * iToolbarScale * TBICONSIZE / 9600;
+  hIcons = ImageList_Create(tbsize, tbsize, ILC_COLOR32, 20, 4);
+  for (int i = 0; i < n; ++i)
+    loadIcon(iconNames[i], hIcons, iDpi * iToolbarScale);
+
+  // update toolbars with new image list
+  SendMessage(hObjectTools, TB_SETIMAGELIST, 0, (LPARAM) hIcons);
+  SendMessage(hEditTools, TB_SETIMAGELIST, 0, (LPARAM) hIcons);
+  SendMessage(hSnapTools, TB_SETIMAGELIST, 0, (LPARAM) hIcons);
+
+  int bands = SendMessage(hRebar, RB_GETBANDCOUNT, 0, 0);
+  // Get new height of the toolbars
+  DWORD dwBtnSize = (DWORD) SendMessage(hEditTools, TB_GETBUTTONSIZE, 0, 0);
+
+  for (int band = 0; band < bands; ++band) {
+    REBARBANDINFO rbBand = { sizeof(REBARBANDINFO) };
+    rbBand.fMask = RBBIM_CHILD; // get hwndChild
+    SendMessage(hRebar, RB_GETBANDINFO, (WPARAM) band, (LPARAM) &rbBand);
+    HWND child = rbBand.hwndChild;
+
+    rbBand.fMask  = RBBIM_CHILDSIZE   // child size members are valid.
+      | RBBIM_SIZE;       // cx is valid.
+    rbBand.cyMinChild = HIWORD(dwBtnSize) + 4;
+    rbBand.cyChild = HIWORD(dwBtnSize) + 4;
+    int bw = LOWORD(dwBtnSize);
+    int size = 0;
+    if (child == hSnapTools)
+      size = iSnapButtons * bw;
+    else if (child == hEditTools)
+      size = iEditButtons * bw;
+    else if (child == hObjectTools)
+      size = iObjectButtons * bw;
+    else if (child == hSelector[EUiGridSize])
+      size = uiscale(GRIDSIZE_WIDTH);
+    else if (child == hSelector[EUiAngleSize])
+      size = uiscale(ANGLESIZE_WIDTH);
+    rbBand.cxMinChild = size;
+    rbBand.cx = size;
+    SendMessage(hRebar, RB_SETBANDINFO, (WPARAM) band, (LPARAM) &rbBand);
+  }
+
+  // list of color icons
+  ImageList_Destroy(hColorIcons);
+  hColorIcons = ImageList_Create(uiscale(COLORICON_SIZE), uiscale(COLORICON_SIZE), ILC_COLOR32, 20, 4);
+  createColorIcons();
+
+  // Resize the window
+  auto lprcNewScale = reinterpret_cast<RECT*>(lParam);
+
+  SetWindowPos(hwnd, nullptr, lprcNewScale->left, lprcNewScale->top,
+	       lprcNewScale->right - lprcNewScale->left, lprcNewScale->bottom - lprcNewScale->top,
+	       SWP_NOZORDER | SWP_NOACTIVATE);
+
+  // finally, notify Lua code
+  lua_rawgeti(L, LUA_REGISTRYINDEX, iModel);
+  lua_getfield(L, -1, "dpiChange");
+  lua_insert(L, -2); // before model
+  lua_pushinteger(L, oldDpi);
+  lua_pushinteger(L, iDpi);
+  luacall(L, 3, 0);
+}
+
 // --------------------------------------------------------------------
 
+int AppUi::dpi() const
+{
+  return iDpi;
+}
+
 // add item to ComboBoxEx
 static void addComboEx(HWND h, String s)
 {
@@ -838,16 +966,22 @@ void AppUi::resetCombos()
 
 void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 {
-  // clear all icons
-  ImageList_Remove(hColorIcons, -1);
+  iColorIcons.clear();
   COMBOBOXEXITEM item;
-  item.mask = CBEIF_TEXT | CBEIF_IMAGE;
+  item.mask = CBEIF_TEXT;
   item.iItem = (INT_PTR) -1;
+  // absolute colors
+  WString wabs(IPEABSOLUTE);
+  item.pszText = wabs.data();
+  iComboContents[EUiStroke].push_back(IPEABSOLUTE);
+  iComboContents[EUiFill].push_back(IPEABSOLUTE);
+  SendMessage(hSelector[EUiStroke], CBEM_INSERTITEM, 0, (LPARAM) &item);
+  SendMessage(hSelector[EUiFill], CBEM_INSERTITEM, 0, (LPARAM) &item);
+
+  item.mask = CBEIF_TEXT | CBEIF_IMAGE;
   for (int i = 0; i < size(sym); ++i) {
     Color color = abs[i].color();
-    HBITMAP bm = colorIcon(color, iUiScale);
-    ImageList_Add(hColorIcons, bm, nullptr);
-    DeleteObject(bm);
+    iColorIcons.push_back(color);
     String s = sym[i].string();
     WString ws(s);
     item.iImage = i;
@@ -858,8 +992,7 @@ void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
     iComboContents[EUiStroke].push_back(s);
     iComboContents[EUiFill].push_back(s);
   }
-  SendMessage(hSelector[EUiStroke], CBEM_SETIMAGELIST, 0, (LPARAM) hColorIcons);
-  SendMessage(hSelector[EUiFill], CBEM_SETIMAGELIST, 0, (LPARAM) hColorIcons);
+  createColorIcons();
 }
 
 void AppUi::addCombo(int sel, String s)
@@ -874,9 +1007,8 @@ void AppUi::setComboCurrent(int sel, int idx)
 
 void AppUi::setButtonColor(int sel, Color color)
 {
-  // ipeDebug("Changing color for %d");
   SendMessage(hButton[sel], BM_SETIMAGE, IMAGE_BITMAP,
-	      (LPARAM) colorIcon(color, iUiScale));
+	      (LPARAM) colorIcon(color, uiscale(COLORBUTTON_SIZE)));
 }
 
 void AppUi::setPathView(const AllAttributes &all, Cascade *sheet)
@@ -946,8 +1078,15 @@ void AppUi::setLayers(const Page *page, int view)
       lvI.lParam |=  1;
     if (page->isLocked(i))
       lvI.lParam |=  2;
-    if (!page->hasSnapping(i))
+    switch (page->snapping(i)) {
+    case Page::SnapMode::Never:
       lvI.lParam |=  4;
+      break;
+    case Page::SnapMode::Always:
+      lvI.lParam |=  8;
+    default:
+      break;
+    }
     WString text(page->layer(i));
     char buf[16];
     sprintf(buf, " (%d)", objCounts[i]);
@@ -1204,8 +1343,10 @@ static const char * const aboutText =
   "(http://www.facebook.com/drawing.editor.Ipe7).\n\n"
   "If you are an Ipe fan and want to show others, have a look at the "
   "Ipe T-shirts (www.shirtee.com/en/store/ipe).\n\n"
-  "Platinum sponsors\n\n"
+  "Platinum and gold sponsors\n\n"
   " * Hee-Kap Ahn\n"
+  " * Günter Rote\n"
+  " * SCALGO\n"
   " * Martin Ziegler\n\n"
   "If you enjoy Ipe, you can become a member of the exclusive community of "
   "Ipe patrons (http://patreon.com/otfried). "
@@ -1221,7 +1362,8 @@ void AppUi::aboutIpe()
 	  IPELIB_VERSION % 100,
 	  COPYRIGHT_YEAR);
 
-  MessageBoxA(hwnd, &buf[0], "About Ipe",
+  WString wbuf(&buf[0]);
+  MessageBoxW(hwnd, wbuf.data(), L"About Ipe",
 	      MB_OK | MB_ICONINFORMATION | MB_APPLMODAL);
 }
 
@@ -1380,7 +1522,7 @@ static int clipboardBitmap(lua_State *L)
     }
     DWORD *r = bits;
     for (int x = 0; x < w; ++x)
-      *p++ = *r++;
+      *p++ = *r++ | 0xff000000; // add opacity
   }
 
   Bitmap ibm(w, h, Bitmap::ENative, pixels);
@@ -1520,6 +1662,8 @@ LRESULT CALLBACK AppUi::wndProc(HWND hwnd, UINT message, WPARAM wParam,
 	  q->clrTextBk = RGB(255, 220, 220);
 	if (flags & 0x04)   // no snapping
 	  q->clrText = RGB(0, 0, 160);
+	if (flags & 0x08)   // always snapping
+	  q->clrText = RGB(0, 160, 0);
 	return CDRF_NEWFONT; }
       default:
 	break;
@@ -1560,6 +1704,10 @@ LRESULT CALLBACK AppUi::wndProc(HWND hwnd, UINT message, WPARAM wParam,
       break;
     }
     break;
+  case WM_DPICHANGED:
+    if (ui && ui->iCanvas)
+      ui->handleDpiChange(hwnd, wParam, lParam);
+    break;
   case WM_CLOSE:
     if (ui) {
       ui->closeRequested();
