@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -79,8 +79,16 @@ void LayerBox::set(const Page *page, int view)
       item->setBackgroundColor(Qt::yellow);
     if (page->isLocked(i))
       item->setBackground(QColor(255, 220, 220));
-    if (!page->hasSnapping(i))
-      item->setForeground(Qt::gray);
+    switch (page->snapping(i)) {
+    case Page::SnapMode::Never:
+      item->setForeground(QColor(0, 0, 160));
+      break;
+    case Page::SnapMode::Always:
+      item->setForeground(QColor(0, 160, 0));
+      break;
+    default:
+      break;
+    }
   }
   iInSet = false;
 }
@@ -108,9 +116,9 @@ void LayerBox::mouseReleaseEvent(QMouseEvent *ev)
 // --------------------------------------------------------------------
 
 #if QT_VERSION < 0x050000
-PathView::PathView(QWidget* parent, Qt::WFlags f) : QWidget(parent, f)
+PathView::PathView(int uiScale, QWidget* parent, Qt::WFlags f) : QWidget(parent, f), iUiScale(uiScale)
 #else
-PathView::PathView(QWidget* parent, Qt::WindowFlags f) : QWidget(parent, f)
+PathView::PathView(int uiScale, QWidget* parent, Qt::WindowFlags f) : QWidget(parent, f), iUiScale(uiScale)
 #endif
 {
   iCascade = nullptr;
@@ -145,6 +153,9 @@ void PathView::paintEvent(QPaintEvent *ev)
     Vector u2 = (1.0/zoom) * Vector(0.80 * w, 0.5 * h);
     Vector u3 = (1.0/zoom) * Vector(0.88 * w, 0.2 * h);
     Vector u4 = (1.0/zoom) * Vector(0.96 * w, 0.5 * h);
+    Vector mid = 0.5 * (v0 + v1);
+    Vector vf = iAll.iFArrowShape.isMidArrow() ? mid : v1;
+    Vector vr = iAll.iRArrowShape.isMidArrow() ? mid : v0;
 
     CairoPainter painter(iCascade, nullptr, cc, 3.0, false);
     painter.setPen(iAll.iPen);
@@ -157,10 +168,10 @@ void PathView::paintEvent(QPaintEvent *ev)
     painter.lineTo(v1);
     painter.drawPath(EStrokedOnly);
     if (iAll.iFArrow)
-      Path::drawArrow(painter, v1, Angle(0),
+      Path::drawArrow(painter, vf, Angle(0),
 		      iAll.iFArrowShape, iAll.iFArrowSize, 100.0);
     if (iAll.iRArrow)
-      Path::drawArrow(painter, v0, Angle(IpePi),
+      Path::drawArrow(painter, vr, Angle(IpePi),
 		      iAll.iRArrowShape, iAll.iRArrowSize, 100.0);
     painter.setDashStyle(Attribute::NORMAL());
     painter.setTiling(iAll.iTiling);
@@ -189,7 +200,7 @@ void PathView::paintEvent(QPaintEvent *ev)
 
 QSize PathView::sizeHint() const
 {
-  return QSize(120, size().width() / 8.0);
+  return QSize(120 * iUiScale / 100, 40 * iUiScale / 100);
 }
 
 void PathView::mousePressEvent(QMouseEvent *ev)
@@ -248,7 +259,7 @@ bool PathView::event(QEvent *ev)
 
 // --------------------------------------------------------------------
 
-PageSorter::PageSorter(Document *doc, int itemWidth, QWidget *parent)
+PageSorter::PageSorter(Document *doc, int pno, int itemWidth, QWidget *parent)
   : QListWidget(parent)
 {
   iDoc = doc;
@@ -262,30 +273,59 @@ PageSorter::PageSorter(Document *doc, int itemWidth, QWidget *parent)
   setMovement(QListView::Static);
 
   Thumbnail r(iDoc, itemWidth);
-  setGridSize(QSize(itemWidth, r.height() + 50));
+  setGridSize(QSize(itemWidth + 40, r.height() + 50));
   setIconSize(QSize(itemWidth, r.height()));
 
-  for (int i = 0; i < doc->countPages(); ++i) {
-    Page *p = doc->page(i);
-    Buffer b = r.render(p, p->countViews() - 1);
-    QImage bits((const uchar *) b.data(), itemWidth, r.height(),
-		QImage::Format_RGB32);
-    // need to copy bits since buffer b is temporary
-    QIcon icon(QPixmap::fromImage(bits.copy()));
+  if (pno >= 0) {
+    const Page *p = doc->page(pno);
+    for (int i = 0; i < p->countViews(); ++i) {
+      Buffer b = r.render(p, i);
+      QImage bits((const uchar *) b.data(), itemWidth, r.height(),
+		  QImage::Format_RGB32);
+      // need to copy bits since buffer b is temporary
+      QIcon icon(QPixmap::fromImage(bits.copy()));
 
-    QString s;
-    QString t = QString::fromUtf8(p->title().z());
-    if (t != "") {
-      s.sprintf("%d: ", i+1);
-      s += t;
-    } else {
-      s.sprintf("Page %d", i+1);
+      QString s;
+      QString t = QString::fromUtf8(p->viewName(i).z());
+      if (t != "") {
+	s.sprintf("%d: ", i+1);
+	s += t;
+      } else {
+	s.sprintf("View %d", i+1);
+      }
+      iMarks.push_back(p->markedView(i));
+      QListWidgetItem *item = new QListWidgetItem(icon, s);
+      item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
+      item->setToolTip(s);
+      item->setData(Qt::UserRole, QVariant(i)); // view number
+      item->setCheckState(iMarks.back() ? Qt::Checked : Qt::Unchecked);
+      addItem(item);
+    }
+  } else {
+    for (int i = 0; i < doc->countPages(); ++i) {
+      Page *p = doc->page(i);
+      Buffer b = r.render(p, p->countViews() - 1);
+      QImage bits((const uchar *) b.data(), itemWidth, r.height(),
+		  QImage::Format_RGB32);
+      // need to copy bits since buffer b is temporary
+      QIcon icon(QPixmap::fromImage(bits.copy()));
+
+      QString s;
+      QString t = QString::fromUtf8(p->title().z());
+      if (t != "") {
+	s.sprintf("%d: ", i+1);
+	s += t;
+      } else {
+	s.sprintf("Page %d", i+1);
+      }
+      iMarks.push_back(p->marked());
+      QListWidgetItem *item = new QListWidgetItem(icon, s);
+      item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
+      item->setToolTip(s);
+      item->setData(Qt::UserRole, QVariant(i)); // page number
+      item->setCheckState(iMarks.back() ? Qt::Checked : Qt::Unchecked);
+      addItem(item);
     }
-    QListWidgetItem *item = new QListWidgetItem(icon, s);
-    item->setFlags(Qt::ItemIsSelectable|Qt::ItemIsEnabled);
-    item->setToolTip(s);
-    item->setData(Qt::UserRole, QVariant(i)); // page number
-    addItem(item);
   }
 }
 
@@ -332,6 +372,26 @@ void PageSorter::insertPages()
   iCutList.clear();
 }
 
+void PageSorter::markPages()
+{
+  QList<QListWidgetItem *> items = selectedItems();
+  for (int i = 0; i < items.count(); ++i) {
+    int page = pageAt(row(items[i]));
+    iMarks[page] = true;
+    items[i]->setCheckState(Qt::Checked);
+  }
+}
+
+void PageSorter::unmarkPages()
+{
+  QList<QListWidgetItem *> items = selectedItems();
+  for (int i = 0; i < items.count(); ++i) {
+    int page = pageAt(row(items[i]));
+    iMarks[page] = false;
+    items[i]->setCheckState(Qt::Unchecked);
+  }
+}
+
 void PageSorter::contextMenuEvent(QContextMenuEvent *ev)
 {
   ev->accept();
@@ -346,9 +406,17 @@ void PageSorter::contextMenuEvent(QContextMenuEvent *ev)
   connect(action_cut, SIGNAL(triggered()), SLOT(cutPages()));
   QAction *action_insert = new QAction("&Insert", this);
   connect(action_insert, SIGNAL(triggered()), SLOT(insertPages()));
+  QAction *action_mark = new QAction("&Mark", this);
+  connect(action_mark, SIGNAL(triggered()), SLOT(markPages()));
+  QAction *action_unmark = new QAction("&Unmark", this);
+  connect(action_unmark, SIGNAL(triggered()), SLOT(unmarkPages()));
   if (selectedItems().count() > 0) {
     m->addAction(action_delete);
     m->addAction(action_cut);
+    if (selectedItems().count() > 1 || !iMarks[pageAt(row(selectedItems()[0]))])
+      m->addAction(action_mark);
+    if (selectedItems().count() > 1 || iMarks[pageAt(row(selectedItems()[0]))])
+      m->addAction(action_unmark);
   }
   if (iCutList.count() > 0)
     m->addAction(action_insert);
