@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -81,26 +81,27 @@ class PathView : public QWidget {
 
 public:
 #if QT_VERSION < 0x050000
-  PathView(QWidget* parent = nullptr, Qt::WFlags f = nullptr);
+  PathView(int uiScale, QWidget* parent = nullptr, Qt::WFlags f = nullptr);
 #else
-  PathView(QWidget* parent = nullptr, Qt::WindowFlags f = nullptr);
+  PathView(int uiScale, QWidget* parent = nullptr, Qt::WindowFlags f = nullptr);
 #endif
 
   void set(const AllAttributes &all, Cascade *sheet);
 
-  virtual QSize sizeHint() const;
+  virtual QSize sizeHint() const override;
 
 signals:
   void activated(String name);
   void showPathStylePopup(Vector v);
 
 protected:
-  virtual void paintEvent(QPaintEvent *ev);
-  virtual void mouseReleaseEvent(QMouseEvent *e);
-  virtual void mousePressEvent(QMouseEvent *e);
-  virtual bool event(QEvent *ev);
+  virtual void paintEvent(QPaintEvent *ev) override;
+  virtual void mouseReleaseEvent(QMouseEvent *e) override;
+  virtual void mousePressEvent(QMouseEvent *e) override;
+  virtual bool event(QEvent *ev) override;
 
 private:
+  int iUiScale;
   Cascade *iCascade;
   AllAttributes iAll;
 };
@@ -111,14 +112,17 @@ class PageSorter : public QListWidget {
   Q_OBJECT
 
 public:
-  PageSorter(Document *doc, int width, QWidget *parent = nullptr);
+  PageSorter(Document *doc, int pno, int width, QWidget *parent = nullptr);
 
   int pageAt(int r) const;
+  std::vector<bool> iMarks;
 
 public slots:
   void deletePages();
   void cutPages();
   void insertPages();
+  void markPages();
+  void unmarkPages();
 
 private:
   virtual void contextMenuEvent(QContextMenuEvent *event);
