@@ -4,7 +4,7 @@
 --[[
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -97,6 +97,7 @@ function MODEL:paction(a1)
       self:warning("Operation '" .. a .. "' is not yet implemented")
     end
   end
+  self:checkType3Font()
 end
 
 local function has_text(page)
@@ -128,11 +129,19 @@ function MODEL:selector(prop, value)
     if not name then return end -- not found
     value = name
   end
+  if value == "<absolute>" then
+    local proptype = prop
+    if prop == "stroke" or prop == "fill" then proptype = "color" end
+    value = self.attributes[prop]
+    if type(value) == "string" then
+      value = self:findOldValue(prop, proptype)
+    end
+  end
   self.attributes[prop] = value
   self.ui:setAttributes(self.doc:sheets(), self.attributes)
   if self:page():hasSelection() then
     self:setAttribute(prop, value)
-    if prop == "textsize" or has_text(self:page()) then
+    if prop == "textsize" and has_text(self:page()) then
       self:autoRunLatex()
     end
   end
@@ -140,15 +149,21 @@ function MODEL:selector(prop, value)
 end
 
 function MODEL:set_absolute(prop, value)
+  self.attributes[prop] = value
+  self.ui:setAttributes(self.doc:sheets(), self.attributes)
   if self:page():hasSelection() then
     self:setAttribute(prop, value)
   end
+  -- self:print_attributes()
 end
 
 function MODEL:layerAction(a, layer, target)
   local name = a
   local arg = nil
-  if a:sub(-2) == "on" then
+  if a:sub(1,4) == "snap" then
+    name = "snap"
+    arg = a:sub(5)
+  elseif a:sub(-2) == "on" then
     name = a:sub(1,-3)
     arg = true
   elseif a:sub(-3) == "off" then
@@ -188,6 +203,12 @@ function MODEL:absoluteButton(button)
 			     old.r, old.g, old.b)
     if r then
       self:set_absolute(button, { r = r, g = g, b = b });
+    elseif config.platform == "apple" then
+      self:warning("Selecting absolute colors",
+      	           "I have displayed the MacOS color selection dialog.\n\n" ..
+		   "After you have found your color in this dialog, " ..
+		   "you'll need to press the absolute color button again to " ..
+		   "apply the new color to your selection in Ipe.")
     end
   elseif button == "pen" then
     local old = self:findOldValue(button, button)
@@ -256,6 +277,30 @@ function MODEL:action_mark_view(m)
   self:register(t)
 end
 
+function MODEL:mark_views_from_view(m, label)
+  local t = { label=label,
+	      pno=self.pno,
+	      vno=self.vno,
+	      original=self:page():clone(),
+	      undo=revertOriginal,
+	      final=m,
+	    }
+  t.redo = function (t, doc)
+	     for i = t.vno, doc[t.pno]:countViews() do
+	       doc[t.pno]:setMarkedView(i, t.final)
+	     end
+	   end
+  self:register(t)
+end
+
+function MODEL:action_mark_from_view()
+  self:mark_views_from_view(true, "mark some views")
+end
+
+function MODEL:action_unmark_from_view()
+  self:mark_views_from_view(false, "unmark some views")
+end
+
 function MODEL:action_mark_page(m)
   local t = { label="set page mark to " .. tostring(m),
 	      pno=self.pno,
@@ -298,6 +343,7 @@ function MODEL:showPathStylePopup(v)
 
   local opacities = sheet:allNames("opacity")
   m:add("opacity", "Opacity", opacities, nil, a.opacity)
+  m:add("strokeopacity", "Stroke opacity", opacities, nil, a.strokeopacity)
 
   local tilings = sheet:allNames("tiling")
   table.insert(tilings, 1, "normal")
@@ -338,10 +384,15 @@ function MODEL:showLayerBoxPopup(v, layer)
   else
     m:add("lockon", "Lock " .. layer)
   end
-  if p:hasSnapping(layer) then
-    m:add("snapoff", "Disable snapping for " .. layer)
-  else
-    m:add("snapon", "Enable snapping for " .. layer)
+  local sm = p:snapping(layer)
+  if sm ~= "never" then
+    m:add("snapnever", "Disable snapping for " .. layer)
+  end
+  if sm ~= "visible" then
+    m:add("snapvisible", "Snap to " .. layer .. " when visible")
+  end
+  if sm ~= "always" then
+    m:add("snapalways", "Always snap to " .. layer)
   end
   local layers = p:layers()
   if #layers > 1 then
@@ -451,7 +502,7 @@ function MODEL:layeraction_snap(layer, arg)
 	      pno=self.pno,
 	      vno=self.vno,
 	      layer=layer,
-	      original=p:hasSnapping(layer),
+	      original=p:snapping(layer),
 	      snapping=arg,
 	    }
   t.undo = function (t, doc)
@@ -639,7 +690,7 @@ end
 
 function MODEL:action_auto_latex()
   self.auto_latex = self.ui:actionState("auto_latex")
-  print("Auto is", self.auto_latex)
+  -- print("Auto is", self.auto_latex)
 end
 
 ----------------------------------------------------------------------
@@ -739,12 +790,23 @@ function MODEL:compute_rect(w, h, res)
   return rect
 end
 
-function MODEL:action_export_png()
+function MODEL:export(format)
   local dir
   if self.file_name then dir = self.file_name:match(prefs.dir_pattern) end
   if not dir then dir = prefs.save_as_directory end
+  local title, filter
+  if format == "png" then
+    title = "Export to PNG"
+    filter = filter_png
+  elseif format == "eps" then
+    title = "Export to EPS"
+    filter = filter_eps
+  else
+    title = "Export to SVG"
+    filter = filter_svg
+  end
   local s, f =
-    ipeui.fileDialog(self.ui:win(), "save", "Export to PNG", filter_png, dir)
+    ipeui.fileDialog(self.ui:win(), "save", title, filter, dir)
   if s then
     if config.toolkit ~= "cocoa" and ipe.fileExists(s) then
       local b = messageBox(self.ui:win(), "question",
@@ -757,37 +819,31 @@ function MODEL:action_export_png()
       end
     end
     self.ui:renderPage(self.doc, self.pno, self.vno,
-		       "png", s, self.ui:zoom(),
-		       false, false) -- transparent, nocrop
+       		       format, s, self.ui:zoom(),
+		       true, false) -- transparent, nocrop
   end
 end
 
 function MODEL:action_export_eps()
-  local dir
-  if self.file_name then dir = self.file_name:match(prefs.dir_pattern) end
-  if not dir then dir = prefs.save_as_directory end
-  local s, f =
-    ipeui.fileDialog(self.ui:win(), "save", "Export to EPS", filter_eps, dir)
-  if s then
-    if config.toolkit ~= "cocoa" and ipe.fileExists(s) then
-      local b = messageBox(self.ui:win(), "question",
-			   "File already exists!",
-			   "Do you wish to overwrite?\n\n" .. s,
-			   "okcancel")
-      if b ~= 1 then
-	self.ui:explain("Document not exported")
-	return
-      end
-    end
-    self.ui:renderPage(self.doc, self.pno, self.vno,
-		       "eps", s, 1.0, false, false)
-  end
+  self:export("eps")
+end
+
+function MODEL:action_export_png()
+  self:export("png")
+end
+
+function MODEL:action_export_svg()
+  self:export("svg")
 end
 
 function MODEL:action_insert_image()
+  if not self.insert_image_dir and self.file_name then
+    self.insert_image_dir = self.file_name:match(prefs.dir_pattern)
+  end
   local s, f = ipeui.fileDialog(self.ui:win(), "open", "Insert image",
-				filter_images)
+				filter_images, self.insert_image_dir)
   if s then
+    self.insert_image_dir = s:match(prefs.dir_pattern)
     local format = "png"
     if s:sub(-4):lower() == ".jpg" or s:sub(-5):lower() == ".jpeg"
     then format = "jpeg" end
@@ -808,6 +864,7 @@ end
 
 ----------------------------------------------------------------------
 
+-- this function is called without arguments on OSX from ipeAlwaysAction
 function action_manual()
   local url = config.docdir .. "/manual.html"
   if ipeui.startBrowser then
@@ -823,12 +880,14 @@ function MODEL:action_manual()
   action_manual()
 end
 
+-- this function is called without arguments on OSX from ipeAlwaysAction
 function action_show_libraries(win)
   local s = "Ipe relies on the following fine pieces of software:\n"
   s = s .. "\n * Pdftex, Xetex, or Luatex"
   s = s .. "\n * " .. _VERSION .. " (" .. math.floor(collectgarbage("count")) .. " kB used)"
   s = s .. "\n * The font rendering library " .. config.freetype_version
   s = s .. "\n * The rendering library Cairo " .. config.cairo_version
+  s = s .. "\n * The clothoid spline library libspiro " .. config.spiro_version
   if config.qt_version then
     s = s .. "\n * The GUI toolkit Qt " .. config.qt_version
   end
@@ -836,7 +895,8 @@ function action_show_libraries(win)
   messageBox(win, "information", "Ipe libraries", s)
 end
 
-function action_show_configuration(win)
+-- this function is called without arguments on OSX from ipeAlwaysAction
+function action_show_configuration(win, dpi)
   local s = ""
   s = s .. " * Lua code: " .. package.path
   s = s .. "\n * Style directories:\n  - " ..
@@ -850,18 +910,24 @@ function action_show_configuration(win)
   s = s .. "\n * Latex program path: " .. config.latexpath
   s = s .. "\n * Latex directory: " .. config.latexdir
   s = s .. "\n * Icons: " .. config.icons
+  s = s .. "\n * Screen geometry: " .. config.screen_geometry[1] .. " x " .. config.screen_geometry[2]
+  if config.platform == "win" then
+    s = s .. "\n * Monitor resolution: " .. dpi .. " dpi"
+  end
   s = s .. "\n * External editor: " .. (prefs.external_editor or "none")
   messageBox(win, "information", "Ipe configuration", s)
 end
 
 function MODEL:action_show_configuration()
-  action_show_configuration(self.ui:win())
+  local cs, dpi = self.ui:canvasSize()
+  action_show_configuration(self.ui:win(), dpi)
 end
 
 function MODEL:action_show_libraries()
   action_show_libraries(self.ui:win())
 end
 
+-- this function is called without arguments on OSX from ipeAlwaysAction
 function action_about_ipelets(win)
   local s = ""
   for i,v in ipairs(ipelets) do
@@ -880,6 +946,25 @@ function MODEL:action_about_ipelets()
   action_about_ipelets(self.ui:win())
 end
 
+function MODEL:action_preferences()
+  local d = ipeui.Dialog(win, "Ipe: Preferences")
+  local s = "Ipe preferences are changed by creating a Lua source file.\n\n"
+  s = s .. "You can find the available options in 'prefs.lua', 'shortcuts.lua', and 'mouse.lua'"
+  s = s .. " (all on the path '" .. package.path .. "').\n\n"
+  s = s .. "Normally, you would not modify those files directly.\n\nInstead, create a new file '"
+  if config.platform == "win" then
+    s = s .. config.ipeletDirs[#config.ipeletDirs] .. "\\" .. "customization.lua"
+  else
+    s = s .. "~/.ipe/ipelets/customization.lua"
+  end
+  s = s .. "', place your changes in this file, and restart Ipe.\n\n"
+  s = s .. "For further details, please read the manual."
+  d:add("text", "text", { read_only=true }, 1, 1)
+  d:set("text", s)
+  d:addButton("ok", "Ok", "accept")
+  d:execute(prefs.latexlog_size)
+end
+
 function MODEL:action_keyboard()
   local s = prefs.keyboard
   if s then
@@ -924,28 +1009,82 @@ function MODEL:action_developer_list_shortcuts()
   print(dupl, " duplicate shortcuts found")
 end
 
+function MODEL:action_cloud_latex()
+  local d = ipeui.Dialog(win, "Ipe: Latex in the cloud")
+  local urlFile = config.latexdir .. "url.txt"
+  if ipe.fileExists(urlFile) then
+    local f = ipe.openFile(urlFile, "r")
+    local url = f:read("a")
+    url = url:match("^(%S+)%s*$")
+    f:close()
+    d:add("label1", "label",
+	  { label="You are currently using the following online Latex-compilation service:" }, 1, 1, 1, 3)
+    d:add("url", "label", { label=url }, 2, 2)
+    d:add("label2", "label",
+	  { label="Please keep in mind that the contents of all your text objects "
+	    .. "is transmitted to this service at every Latex compilation." }, 3, 1, 1, 3)
+    d:add("label3", "label",
+	  { label = "You may not want to use online compilation for confidential information." }, 4, 1, 1, 3)
+    d:addButton("ok", "Disable online compilation", "accept")
+    d:addButton("cancel", "&Cancel", "reject")
+    if d:execute() then
+      os.remove(urlFile)
+    end
+  else
+    d:add("label1", "label",
+	  { label="Would you like to use the following online Latex-compilation service?" }, 1, 1, 1, 3)
+    d:add("url", "label", { label=prefs.latex_service_url }, 2, 2)
+    d:add("label2", "label",
+	  { label="Please keep in mind that the contents of all your text objects "
+	    .. "is transmitted to this service at every Latex compilation." }, 3, 1, 1, 3)
+    d:add("label3", "label",
+	  { label = "You may not want to use online compilation for confidential information." }, 4, 1, 1, 3)
+    d:addButton("ok", "Enable online compilation", "accept")
+    d:addButton("cancel", "&Cancel", "reject")
+    if d:execute() then
+      local f = ipe.openFile(urlFile, "w")
+      f:write(prefs.latex_service_url, "\n")
+      f:close()
+    end
+  end
+end
+
 ----------------------------------------------------------------------
 
-function MODEL:action_set_origin_snap()
+function MODEL:setOrigin(snapping)
   self.snap.origin = self.ui:simpleSnapPos()
-  self.snap.snapangle = true
   self.snap.with_axes = true
-  self.ui:setActionState("snapangle", true)
+  self.ui:setActionState("show_axes", self.snap.with_axes)
+  if snapping then
+    self.snap.snapangle = true
+    self.ui:setActionState("snapangle", true)
+  end
   self:setSnap()
   self.ui:setFifiVisible(true)
   self.ui:update()
 end
 
-function MODEL:action_hide_axes()
-  self.snap.with_axes = false
-  self.snap.snapangle = false
-  self.ui:setActionState("snapangle", false)
+function MODEL:action_set_origin()
+  self:setOrigin(false)
+end
+
+function MODEL:action_set_origin_snap()
+  self:setOrigin(true)
+end
+
+function MODEL:action_show_axes()
+  self.snap.with_axes = self.ui:actionState("show_axes");
+  if not self.snap.with_axes then
+    self.snap.snapangle = false
+    self.ui:setActionState("snapangle", false)
+  end
   self:setSnap()
   self.ui:update()
 end
 
 function MODEL:action_set_direction()
   self.snap.with_axes = true;
+  self.ui:setActionState("show_axes", self.snap.with_axes)
   self.snap.orientation = (self.ui:simpleSnapPos() - self.snap.origin):angle()
   self.ui:setSnap(self.snap)
   self.ui:update()
@@ -1027,6 +1166,7 @@ function MODEL:action_set_tangent_direction()
   local q = qs[1]
   if (pos1 - qs[2]):sqLen() < (pos1 - q):sqLen() then q = qs[2] end
   self.snap.with_axes = true;
+  self.ui:setActionState("show_axes", self.snap.with_axes)
   self.snap.orientation = (arc * q - self.snap.origin):angle()
   self.ui:setSnap(self.snap)
   self.ui:update()
@@ -1039,33 +1179,31 @@ function MODEL:action_reset_direction()
   self.ui:update()
 end
 
-function MODEL:setLine()
+function MODEL:setLine(snapping)
   local origin, dir = self:page():findEdge(self.vno, self.ui:unsnappedPos())
   if not origin then
     self.ui:explain("Mouse is not on an edge")
-    return false
   else
     self.snap.with_axes = true;
+    self.ui:setActionState("show_axes", self.snap.with_axes)
     self.snap.orientation = dir
     self.snap.origin = origin
     self.ui:update()
-    return true
+    if snapping then
+      self.snap.snapangle = true
+      self.ui:setFifiVisible(true)
+      self.ui:setActionState("snapangle", true)
+    end
+    self:setSnap()
   end
 end
 
 function MODEL:action_set_line()
-  if self:setLine() then
-    self:setSnap()
-  end
+  self:setLine(false)
 end
 
 function MODEL:action_set_line_snap()
-  if self:setLine() then
-    self.snap.snapangle = true
-    self.ui:setFifiVisible(true)
-    self.ui:setActionState("snapangle", true)
-    self:setSnap()
-  end
+  self:setLine(true)
 end
 
 ----------------------------------------------------------------------
@@ -1179,7 +1317,7 @@ function MODEL:wheel_zoom(xdelta, ydelta, kind)
   if kind ~= 2 and opts.pan then
     local p = self.ui:pan()
     local ppd = opts.pixel_per_degree / zoom
-    p = p + V(xdelta * ppd, opts.vertical_sign * ydelta * ppd)
+    p = p + V(opts.direction.x * xdelta * ppd, opts.direction.y * ydelta * ppd)
     self.ui:setPan(p)
   else
     local origin = self.ui:unsnappedPos()
@@ -1415,32 +1553,103 @@ function MODEL:action_delete_view()
   self.ui:explain(t.label)
 end
 
-function MODEL:action_edit_effects()
+function attribute_map(map)
+  local s = ""
+  for i,mapping in ipairs(map) do
+    s = s .. mapping.kind .. ":" .. mapping.map .. "=" .. mapping.to .. ";\n"
+  end
+  return s
+end
+
+function parse_attribute_map(s)
+  local kinds = { "pen", "symbolsize", "arrowsize", "opacity", "color", "dashstyle", "symbol" }
+  local map = {}
+  for kind, from, to in string.gmatch(s, "(%l+):([^=]+)=([^;]+);") do
+    if indexOf(kind, kinds) then  -- silently ignore unknown kinds
+      map[#map+1] = { kind=kind, map=from, to=to }
+    end
+  end
+  return map
+end
+
+function layer_matrices(tfm)
+  local s = ""
+  for l, m in pairs(tfm) do
+    s = s .. l .. "=" .. tostring(m) .. ";\n"
+  end
+  return s
+end
+
+function parse_layer_matrices(s)
+  local map = {}
+  for l, ms in string.gmatch(s, "([^%s=]+)=%s*%[([%d%s.e-]+)%]%s*;") do
+    local m = {}
+    for k in string.gmatch(ms, "%S+") do
+      m[#m+1] = tonumber(k)
+    end
+    if #m == 6 then
+      map[l] = ipe.Matrix(m)
+    end
+  end
+  return map
+end
+
+function MODEL:action_edit_view()
   local p = self:page()
-  local effect = p:effect(self.vno)
+  local effect0 = p:effect(self.vno)
+  local name0 = p:viewName(self.vno)
+  local map0 = p:viewMap(self.vno)
+  local tfm0 = p:layerMatrices(self.vno)
   local list = self.doc:sheets():allNames("effect")
   table.insert(list, 1, "normal")
-  local d = ipeui.Dialog(self.ui:win(), "Edit view effect")
-  local label = "Choose effect for view " .. self.vno
-  d:add("label", "label", { label=label }, 1, 1)
-  d:add("combo", "combo", list, 0, 1)
+  local d = ipeui.Dialog(self.ui:win(), "Edit view")
+  local label = "View " .. self.vno .. " of " .. p:countViews()
+  d:add("label", "label", { label=label }, 1, 1, 2, 1)
+  d:add("name-label", "label", { label="Name" }, 0, 1)
+  d:add("name", "input", {}, -1, 2)
+  d:add("map-label", "label", { label="Attribute map" }, 0, 1)
+  d:add("map", "text", {}, -1, 2)
+  d:add("tfm-label", "label", { label="Transformations" }, 0, 1)
+  d:add("tfm", "text", {}, -1, 2)
+  d:add("effect-label", "label", { label="Effect" }, 0, 1)
+  d:add("combo", "combo", list, -1, 2)
   d:addButton("ok", "&Ok", "accept")
   d:addButton("cancel", "&Cancel", "reject")
-  d:set("combo", indexOf(effect, list))
+  d:set("name", name0)
+  d:set("map", attribute_map(map0))
+  d:set("tfm", layer_matrices(tfm0))
+  d:set("combo", indexOf(effect0, list))
   if not d:execute() then return end
-  local final = list[d:get("combo")]
-  local t = { label="set effect of view " .. self.vno .. " to " .. final,
+  local name1 = d:get("name")
+  local effect1 = list[d:get("combo")]
+  local map1 = parse_attribute_map(d:get("map"))
+  local tfm1 = parse_layer_matrices(d:get("tfm"))
+  local t = { label="modify view " .. self.vno,
 	      pno = self.pno,
 	      vno = self.vno,
-	      original = effect,
-	      final = final,
+	      name0 = name0,
+	      effect0 = effect0,
+	      map0 = map0,
+	      tfm0 = tfm0,
+	      name1 = name1,
+	      effect1 = effect1,
+	      map1 = map1,
+	      tfm1 = tfm1,
 	    }
   t.undo = function (t, doc)
-	     p:setEffect(t.vno, t.original)
-	   end
+    local p = doc[t.pno]
+    p:setEffect(t.vno, t.effect0)
+    p:setViewName(t.vno, t.name0)
+    p:setViewMap(t.vno, t.map0)
+    p:setLayerMatrices(t.vno, t.tfm0)
+  end
   t.redo = function (t, doc)
-	     p:setEffect(t.vno, t.final)
-	   end
+    local p = doc[t.pno]
+    p:setEffect(t.vno, t.effect1)
+    p:setViewName(t.vno, t.name1)
+    p:setViewMap(t.vno, t.map1)
+    p:setLayerMatrices(t.vno, t.tfm1)
+  end
   self:register(t)
 end
 
@@ -1476,8 +1685,7 @@ end
 
 function MODEL:action_cut_page()
   if #self.doc == 1 then
-    self:warning("Cannot cut page",
-		 "This is the only page of the document.")
+    self:warning("Cannot cut page", "This is the only page of the document.")
     return
   end
   local t = { label="cut page " .. self.pno,
@@ -1676,6 +1884,8 @@ function MODEL:action_edit_notes()
   self:register(t)
 end
 
+----------------------------------------------------------------------
+
 -- rearrange pages of document
 -- new order is given by arr
 -- 0 entries in arr are taken from newpages
@@ -1704,14 +1914,25 @@ function arrange_pages(doc, arr, newpages)
   return unused
 end
 
+function mark_pages(doc, marks)
+  local originalMarks = {}
+  for i = 1,#doc do
+    local p = doc[i]
+    originalMarks[i] = p:marked()
+    p:setMarked(marks[i])
+  end
+  return originalMarks
+end
+
 function MODEL:action_page_sorter()
-  local arr = self.ui:pageSorter(self.doc)
+  local arr, marks = self.ui:pageSorter(self.doc)
   if not arr then return end -- canceled
   if #arr == 0 then
     self:warning("You cannot delete all pages of the document")
     return
   end
-  printTable(arr)
+  -- printTable(arr)
+  -- printTable(marks)
   -- compute reverse 'permutation'
   local rev = {}
   for i = 1,#self.doc do rev[i] = 0 end
@@ -1721,11 +1942,15 @@ function MODEL:action_page_sorter()
 	      vno = 1,
 	      arr = arr,
 	      rev = rev,
+	      marks = marks,
+	      originalMarks = marks
 	    }
   t.undo = function (t, doc)
 	     arrange_pages(doc, t.rev, t.deleted)
+	     mark_pages(doc, t.originalMarks)
 	   end
   t.redo = function (t, doc)
+	     t.originalMarks = mark_pages(doc, t.marks)
 	     t.deleted = arrange_pages(doc, t.arr, {})
 	   end
   self.pno = 1
@@ -1733,6 +1958,68 @@ function MODEL:action_page_sorter()
   self:register(t)
 end
 
+----------------------------------------------------------------------
+
+-- rearrange views of a page
+-- new order is given by arr
+function arrange_views(p, arr, marks)
+  -- collect view information from the page
+  local views = {}
+  local layers = p:layers()
+  for i = 1,p:countViews() do
+    local visible = {}
+    for _,l in ipairs(layers) do
+      visible[l] = p:visible(i, l)
+    end
+    local v = {
+      name=p:viewName(i),
+      marked=marks[i],
+      active=p:active(i),
+      effect=p:effect(i),
+      visible=visible,
+    }
+    views[#views+1] = v
+  end
+  p:clearViews()
+  -- insert views in new order
+  for i = 1,#arr do
+    local v = views[arr[i]]
+    p:insertView(i, v.active)
+    p:setEffect(i, v.effect)
+    p:setMarkedView(i, v.marked)
+    p:setViewName(i, v.name)
+    for _,l in ipairs(layers) do
+      p:setVisible(i, l, v.visible[l])
+    end
+  end
+end
+
+function MODEL:action_view_sorter()
+  local arr, marks = self.ui:pageSorter(self.doc, self.pno)
+  if not arr then return end -- canceled
+  if #arr == 0 then
+    self:warning("You cannot delete all views of the page")
+    return
+  end
+  -- printTable(arr)
+  -- printTable(marks)
+  local t = { label="rearrangement of views",
+	      pno=self.pno,
+	      vno=1,
+	      arr=arr,
+	      marks=marks,
+	      original=self:page():clone(),
+	      undo=revertOriginal,
+	    }
+  t.redo = function (t, doc)
+	     arrange_views(doc[t.pno], t.arr, t.marks)
+	   end
+  self.vno = 1
+  self:register(t)
+end
+
+----------------------------------------------------------------------
+
 function MODEL:action_toggle_notes()
   self.ui:showTool("notes", self.ui:actionState("toggle_notes"))
   self.ui:setNotes(self:page():notes())
@@ -2149,6 +2436,63 @@ function MODEL:action_paste()
   self:register(t)
 end
 
+function MODEL:action_paste_with_layer()
+  local data = self.ui:clipboard(false)
+  if not data then
+    self:warning("Nothing to paste")
+    return
+  end
+  if type(data) ~= "string" or data:sub(1,13) ~= "<ipeselection" then
+    self:warning("No Ipe selection to paste")
+    return
+  end
+  local elements, layers = ipe.Object(data)
+  if not elements then
+    self:warning("Could not parse Ipe selection on clipboard")
+    return
+  end
+  local p = self:page()
+  local active = p:active(self.vno)
+  local pLayers = p:layers()
+  local impossible = {}
+  for i,l in ipairs(layers) do
+    if l == "" or indexOf(l, pLayers) == nil or p:isLocked(l) then
+      if indexOf(l, impossible) == nil then
+	impossible[#impossible+1] = l
+      end
+      layers[i] = active  -- place in active layer instead
+    end
+  end
+  if #impossible > 0 then
+    local impossibleText = ""
+    for _,l in ipairs(impossible) do
+      impossibleText = impossibleText .. l .. ", "
+    end
+    self:warning("Impossible to paste to specified layer",
+		 "Some of the layers you are trying to paste to either do " ..
+		   "not exist on this page, or are locked.\n\n" ..
+		   "I have used the active layer instead.\n\n" ..
+		   "The following layer names could not be used:\n\n" .. impossibleText)
+  end
+  local t = { label="paste objects with layer",
+	      pno = self.pno,
+	      vno = self.vno,
+	      elements = elements,
+	      layers = layers,
+	      original = p:clone(),
+	      undo = revertOriginal,
+	    }
+  t.redo = function (t, doc)
+	     local p = doc[t.pno]
+	     for i,obj in ipairs(t.elements) do
+	       p:insert(nil, obj, 2, t.layers[i])
+	     end
+	     p:ensurePrimarySelection()
+	   end
+  p:deselectAll()
+  self:register(t)
+end
+
 ----------------------------------------------------------------------
 
 function MODEL:action_check_style()
@@ -2165,7 +2509,7 @@ function MODEL:action_update_style_sheets()
   local dir = nil
   if self.file_name then dir = string.gsub(self.file_name, "[^/\\]+$", "") end
   if dir == "" then dir = "." end
-  local sheets = self.doc:sheets()
+  local sheets = self.doc:sheets():clone()
   local qlog = ""
   for index=1,sheets:count() do
     local sheet = sheets:sheet(index)
@@ -2403,6 +2747,7 @@ function MODEL:action_document_properties()
   d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("column", 5, 1)
   d:setStretch("row", 7, 1)
+  addEditorField(d, "preamble")
   for n in pairs(p) do d:set(n, p[n]) end
   if not d:execute() then return end
 
@@ -2426,26 +2771,26 @@ end
 
 ----------------------------------------------------------------------
 
-function MODEL:saction_set_link_url()
+function MODEL:saction_set_link_action()
   local p = self:page()
   local prim = p:primarySelection()
   local obj = p[prim]
   assert(obj:type() == "group")
-  local url = self:getString("Enter link destination URL", nil, obj:text())
-  if not url then return end
-  if url:match("^%s*$)") then url = "" end
-  local t = { label="set link URL",
+  local link_action = self:getString("Enter link action", nil, obj:text())
+  if not link_action then return end
+  if link_action:match("^%s*$)") then link_action = "" end
+  local t = { label="set link action",
 	      pno=self.pno,
 	      vno=self.vno,
 	      primary=prim,
 	      original=obj:clone(),
-	      url=url,
+	      link_action=link_action,
 	    }
   t.undo = function (t, doc)
 	     doc[t.pno]:replace(t.primary, t.original)
 	   end
   t.redo = function (t, doc)
-	     doc[t.pno][t.primary]:setText(t.url)
+	     doc[t.pno][t.primary]:setText(t.link_action)
 	   end
   self:register(t)
 end
