@@ -4,7 +4,7 @@
 --[[
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -32,7 +32,7 @@ MODEL = {}
 MODEL.__index = MODEL
 
 MODEL.snapmodes = { "snapvtx", "snapctl", "snapbd", "snapint", "snapgrid",
-		    "snapangle", "snapauto"}
+		    "snapangle", "snapcustom", "snapauto"}
 
 function MODEL:new(fname)
   local model = {}
@@ -52,6 +52,7 @@ function MODEL:init(fname)
     snapint = prefs.snap.intersection,
     snapgrid = prefs.snap.grid,
     snapangle = prefs.snap.angle,
+    snapcustom = prefs.snap.custom,
     snapauto = prefs.snap.autoangle,
     grid_visible = prefs.initial.grid_visible,
     pretty_display = false,
@@ -68,6 +69,8 @@ function MODEL:init(fname)
   self.pristine = false
   self.first_show = true
 
+  self.type3_font = false
+
   self.auto_latex = prefs.auto_run_latex
   self.ui:setActionState("auto_latex", self.auto_latex)
 
@@ -118,6 +121,15 @@ function MODEL:sizeChanged()
   self.first_show = false
 end
 
+function MODEL:dpiChange(oldDpi, newDpi)
+  -- adjust zoom by ratio of the two dpi values
+  local nzoom = self.ui:zoom() * newDpi / oldDpi
+  if nzoom > prefs.max_zoom then nzoom = prefs.max_zoom end
+  if nzoom < prefs.min_zoom then nzoom = prefs.min_zoom end
+  self.ui:setZoom(nzoom)
+  self.ui:update()
+end
+
 ----------------------------------------------------------------------
 
 function MODEL:resetGridSize()
@@ -232,6 +244,7 @@ function MODEL:setSnap()
   s = ind(self.snap.snapint, s, "x")
   s = ind(self.snap.snapgrid, s, "+")
   s = ind(self.snap.snapangle, s, "r")
+  s = ind(self.snap.snapcustom, s, "#")
   s = ind(self.snap.snapauto, s, "a")
   s = s .. " " .. self.snap.gridsize
   s = s .. " " .. self.snap.anglesize
@@ -243,6 +256,16 @@ function MODEL:warning(text, details)
   messageBox(self.ui:win(), "warning", text, details)
 end
 
+function numberFormatForMax(m)
+  if m < 10 then
+    return "%d"
+  elseif m < 100 then
+    return "%02d"
+  else
+    return "%03d"
+  end
+end
+
 -- Set canvas to current page
 function MODEL:setPage()
   local p = self:page()
@@ -253,15 +276,23 @@ function MODEL:setPage()
   self:setCaption()
   self:setBookmarks()
   self.ui:setNotes(p:notes())
-  local vno
-  if p:countViews() > 1 then
-    vno = string.format("View %d/%d", self.vno, p:countViews())
+  -- to avoid changing the width of the properties gadget, base the
+  -- format on the max # of views and on the # of pages
+  if #self.doc > 1 or p:countViews() > 1 then
+    local vno
+    local pno
+    local pnofmt = numberFormatForMax(#self.doc)
+    local maxNoViews = 1
+    for _,q in self.doc:pages() do
+      if q:countViews() > maxNoViews then maxNoViews = q:countViews() end
+    end
+    local vnofmt = numberFormatForMax(maxNoViews)
+    pno = string.format(prefs.page_button_prefix .. pnofmt .. "/" .. pnofmt, self.pno, #self.doc)
+    vno = string.format(prefs.view_button_prefix .. vnofmt .. "/" .. vnofmt, self.vno, p:countViews())
+    self.ui:setNumbers(vno, p:markedView(self.vno), pno, p:marked())
+  else
+    self.ui:setNumbers(nil, p:markedView(self.vno), nil, p:marked())
   end
-  local pno
-  if #self.doc > 1 then
-    pno = string.format("Page %d/%d", self.pno, #self.doc)
-  end
-  self.ui:setNumbers(vno, p:markedView(self.vno), pno, p:marked())
 end
 
 function MODEL:getBookmarks()
@@ -296,6 +327,19 @@ function MODEL:findPageForBookmark(index)
   end
 end
 
+function MODEL:checkType3Font()
+  if not self.type3_font and self.ui:type3Font() then
+    self.type3_font = true
+    self:warning("Type3 font detected",
+		 "It appears your document uses a Type3 font.\n\n" ..
+		   "These are bitmapped fonts, typically created by Latex from a Metafont source.\n\n" ..
+		   "Ipe cannot display these fonts (you'll see a box instead), and since they are " ..
+		   "bitmaps, they will not look good when you scale your figure later.\n\n" ..
+		   "A modern Latex installation should not normally use Type3 fonts. You could " ..
+		   "try to install the 'cm-super' package to avoid using Type3 fonts.")
+  end
+end
+
 ----------------------------------------------------------------------
 
 -- Deselect all objects not in current view, or in a locked layer.
@@ -431,6 +475,8 @@ function MODEL:latexErrorBox(log)
 end
 
 function MODEL:runLatex()
+  self.ui:type3Font() -- reset flag in canvas
+  self.type3_font = false
   local success, errmsg, result, log = self.doc:runLatex(self.file_name)
   if success then
     self.ui:setResources(self.doc)
@@ -587,7 +633,7 @@ function MODEL:saveDocument(fname)
     return
   end
 
-  if fm == "xml" and prefs.auto_export_to_pdf then
+  if fm == "xml" and #prefs.auto_export > 0 then
     self:auto_export(fname)
   end
 
@@ -600,13 +646,20 @@ end
 
 function MODEL:auto_export(fname)
   if not self:runLatex() then
-    self.ui:explain("Latex error - could not export to PDF!")
+    self.ui:explain("Latex error - could not auto-export!")
   else
-    local ename = fname:sub(1,-4) .. "pdf"
-    if not self.doc:save(ename, "pdf", { export=true } ) then
-      self:warning("Auto-exporting failed",
-		   "I could not export in PDF format to file '"
-		     .. ename .. "'.")
+    for _, format  in ipairs(prefs.auto_export) do
+      local ename = fname:sub(1,-4) .. format
+      if format == "pdf" then
+	if not self.doc:save(ename, "pdf", { export=true } ) then
+	  self:warning("Auto-exporting failed",
+		       "I could not export in PDF format to file '" .. ename .. "'.")
+	end
+      else
+	self.ui:renderPage(self.doc, 1, 1,
+			   format, ename, prefs.auto_export_resolution / 72.0,
+			   true, false) -- transparent, nocrop
+      end
     end
   end
 end
@@ -681,6 +734,15 @@ function MODEL:register(t)
 end
 
 function MODEL:creation(label, obj)
+  local p = self:page()
+  local active = p:active(self.vno)
+  if not p:visible(self.vno, active) then
+    self:warning("Active layer is invisible",
+		 "You have just created an object in layer '" ..
+		   active .. "'.\n\n" ..
+		   "This layer is currently not visible, so don't be surprised " ..
+		   "that you can't see your new object!")
+  end
   local t = { label=label, pno=self.pno, vno=self.vno,
 	      layer=self:page():active(self.vno), object=obj }
   t.undo = function (t, doc) doc[t.pno]:remove(#doc[t.pno]) end
