@@ -3,7 +3,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -74,8 +74,8 @@ static void setup_globals(lua_State *L)
   lua_setglobal(L, "argv");
 
   NSRect e = [[NSScreen mainScreen] frame];
-  int cx = int(e.size.height);
-  int cy = int(e.size.width);
+  int cx = int(e.size.width);
+  int cy = int(e.size.height);
   lua_createtable(L, 0, 2);
   lua_pushinteger(L, cx);
   lua_rawseti(L, -2, 1);
@@ -120,8 +120,10 @@ static const char * const about_text =
   "(http://www.facebook.com/drawing.editor.Ipe7).\n\n"
   "If you are an Ipe fan and want to show others, have a look at the "
   "Ipe T-shirts (www.shirtee.com/en/store/ipe).\n\n"
-  "Platinum sponsors\n\n"
+  "Platinum and gold sponsors\n\n"
   " * Hee-Kap Ahn\n"
+  " * Günter Rote\n"
+  " * SCALGO\n"
   " * Martin Ziegler\n\n"
   "If you enjoy Ipe, you can become a member of the exclusive community of "
   "Ipe patrons (http://patreon.com/otfried). "
