@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -228,9 +228,11 @@ static int appui_finishTool(lua_State *L)
 
 static int appui_canvasSize(lua_State *L)
 {
-  CanvasBase *canvas = (CanvasBase *) check_canvas(L, 1);
+  AppUiBase **ui = check_appui(L, 1);
+  CanvasBase *canvas = (*ui)->canvas();
   push_vector(L, Vector(canvas->canvasWidth(), canvas->canvasHeight()));
-  return 1;
+  lua_pushinteger(L, (*ui)->dpi());
+  return 2;
 }
 
 static void check_rgb(lua_State *L, int i, int &r, int &g, int &b)
@@ -284,6 +286,13 @@ static int appui_setPretty(lua_State *L)
   return 0;
 }
 
+static int appui_type3Font(lua_State *L)
+{
+  CanvasBase *canvas = check_canvas(L, 1);
+  lua_pushboolean(L, canvas->type3Font());
+  return 1;
+}
+
 // --------------------------------------------------------------------
 
 static int appui_pantool(lua_State *L)
@@ -662,11 +671,17 @@ static int appui_pageSorter(lua_State *L)
 {
   AppUiBase **ui = check_appui(L, 1);
   Document **doc = check_document(L, 2);
+  int page = -1;
+  if (lua_isnumber(L, 3)) {
+    page = lua_tointegerx(L, 3, nullptr);
+    luaL_argcheck(L, 1 <= page && page <= (*doc)->countPages(),
+		  3, "invalid page number");
+  }
 
   int width, height, thumbWidth;
   get_page_sorter_size(L, width, height, thumbWidth);
 
-  return (*ui)->pageSorter(L, *doc, width, height, thumbWidth);
+  return (*ui)->pageSorter(L, *doc, page - 1, width, height, thumbWidth);
 }
 
 static const char *const render_formats[] = {"svg", "png", "eps", "pdf", nullptr };
@@ -717,6 +732,7 @@ static const struct luaL_Reg appui_methods[] = {
   { "setCursor", appui_setCursor },
   { "setNumbering", appui_setNumbering },
   { "setPretty", appui_setPretty },
+  { "type3Font", appui_type3Font },
   // --------------------------------------------------------------------
   { "panTool", appui_pantool },
   { "selectTool", appui_selecttool },
