@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -92,10 +92,10 @@ static bool getFunctionType2(const PdfObj *obj, std::vector<double> &fun)
 static void drawShading(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r)
 {
   // ipeDebug("drawShading: %s", d->dictRepr().z());
-  double type;
-  if (!d->getNumber("ShadingType", type, nullptr) || type < 2.0 || type > 3.0)
+  int type = d->getInteger("ShadingType");
+  if (type < 2 || type > 3)
     return;
-  bool axial = (type == 3.0);
+  bool axial = (type == 3);
   std::vector<double> coords;
   if (!d->getNumberArray("Coords", nullptr, coords) ||
       coords.size() != (axial ? 6 : 4))
@@ -114,16 +114,14 @@ static void drawShading(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r)
   const PdfDict *f = r->getDict(d, "Function");
   if (!f)
     return;
-  double ftype;
   std::vector<double> funs;
   std::vector<double> domain;
   std::vector<double> bounds;
-  if (!f->getNumber("FunctionType", ftype, nullptr))
-    return;
-  if (ftype == 2.0) {
+  int ftype = f->getInteger("FunctionType");
+  if (ftype == 2) {
     if (!getFunctionType2(f, funs))
       return;
-  } else if (ftype == 3.0) {
+  } else if (ftype == 3) {
     if (!f->getNumberArray("Domain", nullptr, domain) || domain.size() != 2) {
       domain.clear();
       domain.push_back(0.0);
@@ -171,19 +169,83 @@ static void drawShading(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r)
   cairo_pattern_destroy(p);
 }
 
+// /DecodeParms << /Colors 3 /Predictor 10 /Columns 466 >>
+// https://en.wikipedia.org/wiki/Portable_Network_Graphics#Filtering
+// http://www.libpng.org/pub/png/spec/1.2/PNG-Filters.html
+
+static bool applyPngPrediction(Buffer &data, int width, int height, int components)
+{
+  int stride = width * components + 1;
+  if (data.size() != height * stride)  // doesn't seem to be the right prediction
+    return false;
+  for (int row = 0; row < height; ++row) {
+    uint8_t *p = (uint8_t *) data.data() + row * stride;
+    const uint8_t *secondPixel = p + 1 + components;
+    const uint8_t *fin = p + stride;
+    int predictor = *p++;
+    while (p < secondPixel) {
+      switch (predictor) {
+      case 2: // up
+      case 4: // Paeth filter
+	*p += p[-stride];
+	break;
+      case 3: // average
+	*p += p[-stride] >> 1;
+	break;
+      default:
+	break;  // do nothing
+      }
+      ++p;
+    }
+    int left, up, upLeft, pre, pa, pb, pc;
+    while (p < fin) {
+      switch (predictor) {
+      case 1: // left
+	*p += p[-components];
+	break;
+      case 2: // up
+	*p += p[-stride];
+	break;
+      case 3: // average
+	*p += (p[-components] + p[-stride]) >> 1;
+	break;
+      case 4: // Paeth filter
+	left = p[-components];
+	up = p[-stride];
+	upLeft = p[-components-stride];
+	pre = left + up - upLeft;
+	if ((pa = pre - left) < 0)
+	  pa = -pa;
+	if ((pb = pre - up) < 0)
+	  pb = -pb;
+	if ((pc = pre - upLeft) < 0)
+	  pc = -pc;
+	if (pa <= pb && pa <= pc)
+	  *p += left;
+	else if (pb <= pc)
+	  *p += up;
+	else
+	  *p += upLeft;
+	break;
+      case 0:
+      default:
+	break;  // do nothing
+      }
+      ++p;
+    }
+  }
+  return true;
+}
 
 static void drawImage(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r, double opacity)
 {
   // ipeDebug("Image: %s", d->dictRepr().z());
-  double width, height, bpc;
-  const PdfObj *cs = d->get("ColorSpace", nullptr);
-  if (!d->getNumber("Width", width, nullptr) ||
-      !d->getNumber("Height", height, nullptr) ||
-      !d->getNumber("BitsPerComponent", bpc, nullptr) ||
-      !cs || !cs->name() ||
-      (cs->name()->value() != "DeviceRGB" &&
-       cs->name()->value() != "DeviceGray") ||
-      bpc != 8) {
+  int width = d->getInteger("Width");
+  int height = d->getInteger("Height");
+  int bpc = d->getInteger("BitsPerComponent");
+  const PdfObj *cs = d->get("ColorSpace");
+  if (width < 0 || height < 0 || bpc != 8 || !cs || !cs->name() ||
+      (cs->name()->value() != "DeviceRGB" && cs->name()->value() != "DeviceGray")) {
     ipeDebug("Unsupported image: %s", d->dictRepr().z());
     return;
   }
@@ -191,16 +253,36 @@ static void drawImage(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r, d
   if (cs->name()->value() == "DeviceGray")
     components = 1;
   bool jpg = false;
-  const PdfObj *filter = d->get("Filter", nullptr);
+  bool pngPrediction = false;
+  const PdfObj *filter = d->get("Filter");
   if (filter && filter->name()) {
     String fn = filter->name()->value();
     if (fn == "DCTDecode")
       jpg = true;
-    else if (fn != "FlateDecode") {
+    else if (fn == "FlateDecode") {
+      const PdfObj *decodeParms = d->get("DecodeParms");
+      if (decodeParms && decodeParms->dict() &&
+	  decodeParms->dict()->getInteger("Predictor") >= 10)
+	pngPrediction = true;
+    } else {
       ipeDebug("Unsupported filter in image: %s", d->dictRepr().z());
       return;
     }
   }
+  Buffer alphaChannel;
+  uint8_t *alpha = nullptr;
+  const PdfDict *smask = r->getDict(d, "SMask");
+  if (smask) {
+    const PdfObj *mcs = smask->get("ColorSpace");
+    if (!mcs || !mcs->name() || mcs->name()->value() != "DeviceGray"
+	|| smask->getInteger("BitsPerComponent") != 8) {
+      ipeDebug("Unsupported /SMask: %s", smask->dictRepr().z());
+    } else {
+      alphaChannel = smask->inflate();
+      if (alphaChannel.size() == width * height)
+	alpha = (uint8_t *) alphaChannel.data();
+    }
+  }
   Buffer stream = d->inflate();
   Buffer pixels(4 * width * height);
   if (jpg) {
@@ -208,22 +290,32 @@ static void drawImage(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r, d
       return;
   } else {
     const char *p = stream.data();
+
+    pngPrediction = pngPrediction &&
+      applyPngPrediction(stream, width, height, components);
+
     uint32_t *q = (uint32_t *) pixels.data();
     if (components == 3) {
       for (int y = 0; y < height; ++y) {
+	if (pngPrediction)
+	  ++p; // skip predictor byte
 	for (int x = 0; x < width; ++x) {
 	  uint8_t r = uint8_t(*p++);
 	  uint8_t g = uint8_t(*p++);
 	  uint8_t b = uint8_t(*p++);
-	  uint32_t pixel = 0xff000000 | (r << 16) | (g << 8) | b;
+	  uint8_t a = alpha ? *alpha++ : 0xff;
+	  uint32_t pixel = (a << 24) | (r << 16) | (g << 8) | b;
 	  *q++ = pixel;
 	}
       }
     } else {
       for (int y = 0; y < height; ++y) {
+	if (pngPrediction)
+	  ++p; // skip predictor byte
 	for (int x = 0; x < width; ++x) {
 	  uint8_t r = uint8_t(*p++);
-	  uint32_t pixel = 0xff000000 | (r << 16) | (r << 8) | r;
+	  uint8_t a = alpha ? *alpha++ : 0xff;
+	  uint32_t pixel = (a << 24) | (r << 16) | (r << 8) | r;
 	  *q++ = pixel;
 	}
       }
@@ -257,7 +349,7 @@ static void drawImage(cairo_t *cr, const PdfDict *d, const PdfResourceBase *r, d
   to avoid drawing text without Latex. */
 CairoPainter::CairoPainter(const Cascade *sheet, Fonts *fonts,
 			   cairo_t *cc, double zoom, bool pretty)
-  : Painter(sheet), iFonts(fonts), iCairo(cc), iZoom(zoom), iPretty(pretty)
+  : Painter(sheet), iFonts(fonts), iCairo(cc), iZoom(zoom), iPretty(pretty), iType3Font(false)
 {
   iDimmed = false;
 }
@@ -736,6 +828,8 @@ void CairoPainter::execute(const PdfDict *xform, const PdfDict *resources, bool
 	// content markers, ignore
       } else if (op == "ri") {
 	// set rendering intent, ignore
+      } else if (op == "cs") {
+	// set color space, ignore
       } else {
 	String a;
 	for (const auto & arg : iArgs)
@@ -932,7 +1026,7 @@ void CairoPainter::opDo()
   const PdfDict *xf = findResource("XObject", name);
   if (!xf)
     return;
-  const PdfObj *subtypeObj = xf->get("Subtype", nullptr);
+  const PdfObj *subtypeObj = xf->get("Subtype");
   if (!subtypeObj || !subtypeObj->name())
     return;
   String subtype = subtypeObj->name()->value();
@@ -1047,23 +1141,21 @@ void CairoPainter::createPattern()
 {
   auto & ps = iPdfState.back();
   const PdfDict *pat = findResource("Pattern", ps.iFillPattern);
-  double patternType, paintType, xstep, ystep;
-  if (pat && pat->getNumber("PatternType", patternType, nullptr)
-      && pat->getNumber("PaintType", paintType, nullptr)
-      && pat->getNumber("XStep", xstep, nullptr)
-      && pat->getNumber("YStep", ystep, nullptr)) {
-    ipeDebug("Fill pattern /%s PatternType %g PaintType %g xstep %g ystep %g", ps.iFillPattern.z(),
-	     patternType, paintType, xstep, ystep);
-    // handle tiling patterns only
-    if (patternType != 1.0)
+  // handle tiling patterns only
+  if (pat && pat->getInteger("PatternType") == 1) {
+    int paintType = pat->getInteger("PaintType");
+    double xstep, ystep;
+    if (!pat->getNumber("XStep", xstep) || !pat->getNumber("YStep", ystep))
       return;
+    ipeDebug("Tiling pattern /%s PaintType %d xstep %g ystep %g", ps.iFillPattern.z(),
+	     paintType, xstep, ystep);
 
     // to get good quality patterns, the pattern surface cannot be too small
     // except that for Ipe patterns this isn't necessary.
     double xscale = 1.0;
     double yscale = 1.0;
     // Heuristic: if a matrix exists, let's assume small cell is okay
-    if (pat->get("Matrix", nullptr) == nullptr) {
+    if (pat->get("Matrix") == nullptr) {
       while (xscale * xstep < 100)
 	xscale *= 2.0;
       while (yscale * ystep < 100)
@@ -1203,8 +1295,12 @@ void CairoPainter::opTf()
   String name = iArgs[0]->name()->value();
   iPdfState.back().iFontSize = iArgs[1]->number()->value();
   const PdfDict *fd = findResource("Font", name);
-  if (fd)
-    iPdfState.back().iFont = iFonts->getFace(fd);
+  if (fd) {
+    Face *f = iFonts->getFace(fd);
+    iPdfState.back().iFont = f;
+    if (f->type() == FontType::Type3)
+      iType3Font = true;
+  }
 }
 
 void CairoPainter::opTd(bool setLeading)
