@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -72,14 +72,20 @@ Buffer Thumbnail::render(const Page *page, int view)
 					iWidth, iHeight, iWidth * 4);
   cairo_t *cc = cairo_create(surface);
   cairo_scale(cc, iZoom, -iZoom);
-  Vector offset = iLayout->iOrigin - iLayout->paper().topLeft();
+  Vector offset = - iLayout->paper().topLeft();
   cairo_translate(cc, offset.x, offset.y);
 
   CairoPainter painter(iDoc->cascade(), iFonts.get(), cc, iZoom, true);
+  painter.setAttributeMap(&page->viewMap(view));
+  std::vector<Matrix> layerMatrices = page->layerMatrices(view);
   painter.pushMatrix();
   for (int i = 0; i < page->count(); ++i) {
-    if (page->objectVisible(view, i))
+    if (page->objectVisible(view, i)) {
+      painter.pushMatrix();
+      painter.transform(layerMatrices[page->layerOf(i)]);
       page->object(i)->draw(painter);
+      painter.popMatrix();
+    }
   }
   painter.popMatrix();
   cairo_surface_flush(surface);
@@ -97,13 +103,17 @@ static cairo_status_t stream_writer(void *closure, const unsigned char *data, un
   return CAIRO_STATUS_SUCCESS;
 }
 
+// zoom and transparent are ignored for formats other than EPNG
 bool Thumbnail::saveRender(TargetFormat fm, const char *dst,
 			   const Page *page, int view, double zoom,
 			   bool transparent, bool nocrop)
 {
+  if (fm != EPNG)
+    zoom = 1.0;
+
   Rect bbox;
   int wid, ht;
-  if (nocrop) {
+  if (nocrop || fm == EPS) {
     bbox = iLayout->paper();
     wid = int(bbox.width() * zoom);
     ht = int(bbox.height() * zoom);
@@ -147,11 +157,17 @@ bool Thumbnail::saveRender(TargetFormat fm, const char *dst,
 
   cairo_t *cc = cairo_create(surface);
   cairo_scale(cc, zoom, -zoom);
-  cairo_translate(cc, -bbox.topLeft().x, -bbox.topLeft().y);
+  Vector offset = bbox.topLeft();
+  if (fm != EPNG) {
+    // make sure integer coordinates remain integer
+    offset.x = floor(offset.x);
+    offset.y = ceil(offset.y);
+  }
+  cairo_translate(cc, -offset.x, -offset.y);
 
   CairoPainter painter(iDoc->cascade(), iFonts.get(), cc, zoom, true);
-  // painter.Transform(IpeLinear(zoom, 0, 0, -zoom));
-  // painter.Translate(-bbox.TopLeft());
+  painter.setAttributeMap(&page->viewMap(view));
+  std::vector<Matrix> layerMatrices = page->layerMatrices(view);
   painter.pushMatrix();
 
   if (nocrop) {
@@ -159,11 +175,19 @@ bool Thumbnail::saveRender(TargetFormat fm, const char *dst,
       iDoc->cascade()->findSymbol(Attribute::BACKGROUND());
     if (background && page->findLayer("BACKGROUND") < 0)
       painter.drawSymbol(Attribute::BACKGROUND());
+
+    const Text *title = page->titleText();
+    if (title)
+      title->draw(painter);
   }
 
   for (int i = 0; i < page->count(); ++i) {
-    if (page->objectVisible(view, i))
+    if (page->objectVisible(view, i)) {
+      painter.pushMatrix();
+      painter.transform(layerMatrices[page->layerOf(i)]);
       page->object(i)->draw(painter);
+      painter.popMatrix();
+    }
   }
 
   painter.popMatrix();
