@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -531,11 +531,9 @@ Canvas::Canvas(HWND parent, HINSTANCE hInstance) : iHimetric(0., 0.)
 UINT Canvas::getDpiForWindow(HWND hwnd)
 {
   if (pGetDpiForWindow) {
-    UINT nDpi = pGetDpiForWindow(hwnd) * 100 / 96;
-    ipeDebug("Updating DPI to %d", nDpi);
-    return nDpi;
+    return pGetDpiForWindow(hwnd);
   } else
-    return 100;
+    return 96;
 }
 
 void Canvas::init(HINSTANCE hInstance)
