@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -68,7 +68,7 @@ static void buildString(std::vector<short> &t, const char *s)
 
 struct SData {
   HIMAGELIST hImageList;
-  std::vector<String> labels;
+  std::vector<std::wstring> labels;
   int startIndex;
 };
 
@@ -78,7 +78,7 @@ static void populateView(HWND h, SData *d)
   lvI.mask = LVIF_TEXT | LVIF_IMAGE;
   lvI.iSubItem = 0;
   for (int i = 0; i < size(d->labels); ++i) {
-    lvI.pszText = d->labels[i].w().data();
+    lvI.pszText = d->labels[i].data();
     lvI.iItem  = i;
     lvI.iImage = i;
     (void) ListView_InsertItem(h, &lvI);
@@ -120,8 +120,9 @@ int showPageSelectDialog(int width, int height, const char * title,
 {
   SData sData;
   sData.hImageList = il;
-  sData.labels = std::move(labels);
   sData.startIndex = startIndex;
+  for (int i = 0; i < size(labels); ++i)
+    sData.labels.push_back(labels[i].w());
 
   // Convert to dialog units:
   LONG base = GetDialogBaseUnits();
@@ -180,7 +181,7 @@ int showPageSelectDialog(int width, int height, const char * title,
     If \a page is non-negative, all views of this page are shown, and
     the selected view number is returned.
 
-    \a itemWidth is the width of the page thumbnails (the height is
+    \a pageWidth is the width of the page thumbnails (the height is
     computed automatically).
 */
 int CanvasBase::selectPageOrView(Document *doc, int page, int startIndex,
