@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -68,7 +68,8 @@ namespace ipe {
     "opacity", "strokeopacity", "tiling", "gradient",
     "horizontalalignment", "verticalalignment",
     "linejoin", "linecap", "fillrule",
-    "pinned", "transformations", "transformabletext",
+    "pinned", "transformations",
+    "transformabletext", "splinetype",
     "minipage", "width", "decoration", nullptr };
 }
 
@@ -311,6 +312,7 @@ static const char * const enumeration_name[] =
     "none", "horizontal", "vertical", "fixed",    // pinned
     "translations", "rigid", "affine",            // transformations
     "stroked", "strokedfilled", "filled",         // pathmode
+    "bspline", "cardinal", "spiro",               // splinetype
   };
 
 //! Return string representing the attribute.
@@ -363,6 +365,12 @@ Color::Color(String str)
     st >> iBlue;
 }
 
+//! Is it a symbolic arrow name of the form "arrow/mid-*"?
+bool Attribute::isMidArrow() const
+{
+  return isSymbolic() && string().hasPrefix("arrow/mid-");
+}
+
 //! Make a color attribute.
 /*! If the string starts with a letter, make a symbolic attribute.
   Otherwise, it's either a single gray value (0.0 to 1.0), or the
@@ -458,6 +466,35 @@ Attribute Attribute::normal(Kind kind)
 
 // --------------------------------------------------------------------
 
+//! Map the given symbolic attribute sym.  Returns sym if there is no mapping.
+Attribute AttributeMap::map(Kind kind, Attribute sym) const
+{
+  for (const auto & m : iMap) {
+    if (m.kind == kind and m.from == sym)
+      return m.to;
+  }
+  return sym;
+}
+
+//! Save map in XML format
+void AttributeMap::saveAsXml(Stream &stream) const
+{
+  for (int k = 0; k < size(iMap); ++k) {
+    const AttributeMapping &mapping = iMap[k];
+    stream << "<map kind=\"" << kind_names[mapping.kind]
+	   << "\" from=\"" << mapping.from.string()
+	   << "\" to=\"" << mapping.to.string() << "\" />\n";
+  }
+}
+
+//! Add a mapping
+void AttributeMap::add(const AttributeMapping &map)
+{
+  iMap.push_back(map);
+}
+
+// --------------------------------------------------------------------
+
 /*! \class ipe::AllAttributes
   \ingroup attr
   \brief Collection of all object attributes.
