@@ -3,7 +3,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -45,13 +45,8 @@ using namespace ipe;
 
 #ifdef __APPLE__
 
-bool dctDecode(Buffer dctData, Buffer pixelData)
+bool cgImageDecode(CGImageRef bitmap, Buffer pixelData)
 {
-  CGDataProviderRef source =
-    CGDataProviderCreateWithData(nullptr, dctData.data(), dctData.size(), nullptr);
-  CGImageRef bitmap =
-    CGImageCreateWithJPEGDataProvider(source, nullptr, false, kCGRenderingIntentDefault);
-
   if (CGImageGetBitsPerComponent(bitmap) != 8)
     return false;
 
@@ -65,7 +60,7 @@ bool dctDecode(Buffer dctData, Buffer pixelData)
 
   CGBitmapInfo info = CGImageGetBitmapInfo(bitmap);
   // Do we need to check for alpha channel, float pixel values, and byte order?
-  ipeDebug("dctDecode: %d x %d x %d, stride %d, info %x", w, h, bytes, stride, info);
+  ipeDebug("cgImageDecode: %d x %d x %d, stride %d, info %x", w, h, bytes, stride, info);
 
   CFDataRef pixels = CGDataProviderCopyData(CGImageGetDataProvider(bitmap));
   const uint8_t *inRow = CFDataGetBytePtr(pixels);
@@ -89,9 +84,19 @@ bool dctDecode(Buffer dctData, Buffer pixelData)
     }
   }
   CFRelease(pixels);
+  return true;
+}
+
+bool dctDecode(Buffer dctData, Buffer pixelData)
+{
+  CGDataProviderRef source =
+    CGDataProviderCreateWithData(nullptr, dctData.data(), dctData.size(), nullptr);
+  CGImageRef bitmap =
+    CGImageCreateWithJPEGDataProvider(source, nullptr, false, kCGRenderingIntentDefault);
+  int result = cgImageDecode(bitmap, pixelData);
   CGImageRelease(bitmap);
   CGDataProviderRelease(source);
-  return true;
+  return result;
 }
 
 #else
