@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -116,17 +116,6 @@ FileFormat Document::fileFormat(DataSource &source)
     return FileFormat::Xml;
   if (s1.substr(0, 4) == "%PDF")
     return FileFormat::Pdf;  // let's assume it contains an Ipe stream
-  if (s1.substr(0, 4) == "%!PS") {
-    if (s2.substr(0, 11) != "%%Creator: ")
-      return FileFormat::Unknown;
-    if (s2.substr(11, 6) == "Ipelib" || s2.substr(11, 4) == "xpdf")
-      return FileFormat::Eps;
-    if (s2.substr(11, 3) == "Ipe")
-      return FileFormat::Ipe5;
-    return FileFormat::Unknown;
-  }
-  if (s1.substr(0, 5) == "%\\Ipe" || s1.substr(0, 6) == "%\\MIPE")
-    return FileFormat::Ipe5;
   return FileFormat::Unknown;
 }
 
@@ -141,8 +130,6 @@ FileFormat Document::formatFromFilename(String fn)
     return FileFormat::Xml;
   else if (s == ".pdf")
     return FileFormat::Pdf;
-  else if (s == ".eps")
-    return FileFormat::Eps;
   else
     return FileFormat::Unknown;
 }
@@ -171,114 +158,6 @@ Buffer PdfStreamParser::pdfStream(int objNum)
   return obj->dict()->stream();
 }
 
-// --------------------------------------------------------------------
-
-class PsSource : public DataSource {
-public:
-  PsSource(DataSource &source) : iSource(source) { /* nothing */ }
-  bool skipToXml();
-  String readLine();
-  Buffer image(int index) const;
-  int getNext() const;
-  inline bool deflated() const { return iDeflated; }
-
-  virtual int getChar();
-private:
-  DataSource &iSource;
-  std::vector<Buffer> iImages;
-  bool iEos;
-  bool iDeflated;
-};
-
-int PsSource::getChar()
-{
-  int ch = iSource.getChar();
-  if (ch == '\n')
-    iSource.getChar(); // remove '%'
-  return ch;
-}
-
-String PsSource::readLine()
-{
-  String s;
-  int ch = iSource.getChar();
-  while (ch != EOF && ch != '\n') {
-    s += char(ch);
-    ch = iSource.getChar();
-  }
-  iEos = (ch == EOF);
-  return s;
-}
-
-Buffer PsSource::image(int index) const
-{
-  if (1 <= index && index <= int(iImages.size()))
-    return iImages[index - 1];
-  else
-    return Buffer();
-}
-
-bool PsSource::skipToXml()
-{
-  iDeflated = false;
-
-  String s1 = readLine();
-  String s2 = readLine();
-
-  if (s1.substr(0, 11) != "%!PS-Adobe-" ||
-      s2.substr(0, 17) != "%%Creator: Ipelib")
-    return false;
-
-  do {
-    s1 = readLine();
-    if (s1.substr(0, 17) == "%%BeginIpeImage: ") {
-      Lex lex(s1.substr(17));
-      int num, len;
-      lex >> num >> len;
-      if (num != int(iImages.size() + 1))
-	return false;
-      (void) readLine();  // skip 'image'
-      Buffer buf(len);
-      A85Source a85(iSource);
-      char *p = buf.data();
-      char *p1 = p + buf.size();
-      while (p < p1) {
-	int ch = a85.getChar();
-	if (ch == EOF)
-	  return false;
-	*p++ = char(ch);
-      }
-      iImages.push_back(buf);
-    }
-  } while (!iEos && s1.substr(0, 13) != "%%BeginIpeXml");
-
-  iDeflated = (s1.substr(13, 14) == ": /FlateDecode");
-  if (iEos)
-    return false;
-  (void) iSource.getChar(); // skip '%' before <ipe>
-  return true;
-}
-
-class PsStreamParser : public ImlParser {
-public:
-  explicit PsStreamParser(DataSource &source, PsSource &psSource);
-  virtual Buffer pdfStream(int objNum);
-private:
-  PsSource &iPsSource;
-};
-
-PsStreamParser::PsStreamParser(DataSource &source, PsSource &psSource)
-  : ImlParser(source), iPsSource(psSource)
-{
-  // nothing
-}
-
-Buffer PsStreamParser::pdfStream(int objNum)
-{
-  return iPsSource.image(objNum);
-}
-
-
 // --------------------------------------------------------------------
 
 Document *doParse(Document *self, ImlParser &parser, int &reason)
@@ -302,25 +181,6 @@ Document *doParseXml(DataSource &source, int &reason)
   return doParse(self, parser, reason);
 }
 
-Document *doParsePs(DataSource &source, int &reason)
-{
-  PsSource psSource(source);
-  reason = Document::EFileOpenError; // could not find Xml stream
-  if (!psSource.skipToXml())
-    return nullptr;
-
-  Document *self = new Document;
-  if (psSource.deflated()) {
-    A85Source a85(psSource);
-    InflateSource source(a85);
-    PsStreamParser parser(source, psSource);
-    return doParse(self, parser, reason);
-  } else {
-    PsStreamParser parser(psSource, psSource);
-    return doParse(self, parser, reason);
-  }
-}
-
 Document *doParsePdf(DataSource &source, int &reason)
 {
   PdfFile loader;
@@ -338,7 +198,7 @@ Document *doParsePdf(DataSource &source, int &reason)
   // was the object really created by Ipe?
   if (!obj || !obj->dict())
     return nullptr;
-  const PdfObj *type = obj->dict()->get("Type", nullptr);
+  const PdfObj *type = obj->dict()->get("Type");
   if (!type || !type->name() || type->name()->value() != "Ipe")
     return nullptr;
 
@@ -371,10 +231,7 @@ Document *Document::load(DataSource &source, FileFormat format,
   if (format == FileFormat::Pdf)
     return doParsePdf(source, reason);
 
-  if (format == FileFormat::Eps)
-    return doParsePs(source, reason);
-
-  reason = (format == FileFormat::Ipe5) ? EVersionTooOld : ENotAnIpeFile;
+  reason = ENotAnIpeFile;
   return nullptr;
 }
 
@@ -674,21 +531,6 @@ void Document::setResources(PdfResources *resources)
   iResources = resources;
 }
 
-#if 0
-//! Load a style sheet and add at top of cascade.
-bool Document::addStyleSheet(DataSource &source)
-{
-  ImlParser parser(source);
-  StyleSheet *sheet = parser.parseStyleSheet();
-  if (sheet) {
-    sheet->setCascade(iCascade);
-    setStyleSheet(sheet);
-    return true;
-  } else
-    return false;
-}
-#endif
-
 //! Return total number of views in all pages.
 int Document::countTotalViews() const
 {
@@ -856,9 +698,6 @@ int Document::runLatex(String docname)
   case ErrWritingSource:
     fprintf(stderr, "Error writing Latex source.\n");
     return 1;
-  case ErrOldPdfLatex:
-    fprintf(stderr, "Your installed version of Pdflatex is too old.\n");
-    return 1;
   case ErrRunLatex:
     fprintf(stderr, "There was an error trying to run Pdflatex.\n");
     return 1;
