@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -37,6 +37,15 @@
 
 */
 
+#ifdef IPE_GSL
+#include <gsl/gsl_vector.h>
+#include <gsl/gsl_min.h>
+#include <gsl/gsl_multimin.h>
+#include <gsl/gsl_poly.h>
+#endif
+
+// #define GSL_OUTPUT 1
+
 const double BEZIER_INTERSECT_PRECISION = 1.0;
 
 #include "ipegeo.h"
@@ -50,6 +59,18 @@ inline double sq(double x)
 
 // --------------------------------------------------------------------
 
+struct _bezctx {
+  Vector pos;
+  std::vector<Bezier> &bez;
+};
+
+
+#include "spiroentrypoints.h"
+#include "bezctx_intf.h"
+#include "spiro.h"
+
+// --------------------------------------------------------------------
+
 /*! \class ipe::Angle
   \ingroup geo
   \brief A double that's an angle.
@@ -777,6 +798,89 @@ void Bezier::closedSpline(int n,  const Vector *v, std::vector<Bezier> &result)
   }
 }
 
+static Bezier cardinal(Vector prev, Vector p, Vector q, Vector next, double tension)
+{
+  Vector t1 = q - prev; // tangent in p
+  Vector t2 = next - p; // tangent in q
+  Vector cp1 = p + (tension / 3.0) * t1;
+  Vector cp2 = q - (tension / 3.0) * t2;
+  return Bezier(p, cp1, cp2, q);
+}
+
+//! Convert a cardinal spline to a series of Bezier splines.
+/*! Bezier splines are appended to \a result.
+*/
+void Bezier::cardinalSpline(int n,  const Vector *v, double tension, std::vector<Bezier> &result)
+{
+  if (n == 2) {
+    result.push_back(Bezier(v[0], v[0], v[1], v[1]));
+  } else {
+    result.push_back(cardinal(v[0], v[0], v[1], v[2], tension));
+    for (int i = 1; i < n - 2; ++i)
+      result.push_back(cardinal(v[i-1], v[i], v[i+1], v[i+2], tension));
+    result.push_back(cardinal(v[n-3], v[n-2], v[n-1], v[n-1], tension));
+  }
+}
+
+// ------------------------------------------------------------------------------------------
+// These functions are called by spiro to generate Bezier output
+
+void bezctx_moveto(bezctx *bc, double x, double y, int is_open) {
+  // ipeDebug("bezctx_moveto(%g,%g)_%d\n",x,y,is_open);
+  bc->pos = Vector(x, y);
+}
+
+void bezctx_lineto(bezctx *bc, double x, double y) {
+  // ipeDebug("bezctx_lineto(%g,%g)\n",x,y);
+  Vector pos{x, y};
+  bc->bez.push_back(Bezier(bc->pos, bc->pos, pos, pos));
+  bc->pos = pos;
+}
+
+// most likely this will not be called with our settings
+void bezctx_quadto(bezctx *bc, double x1, double y1, double x2, double y2) {
+  // ipeDebug("bezctx_quadto(%g,%g, %g,%g)\n",x1,y1,x2,y2);
+  Vector cp1{x1, y1}, cp2{x2, y2};
+  bc->bez.push_back(Bezier::quadBezier(bc->pos, cp1, cp2));
+  bc->pos = cp2;
+}
+
+void bezctx_curveto(bezctx *bc, double x1, double y1, double x2, double y2,
+		    double x3, double y3) {
+  // ipeDebug("bezctx_curveto(%g,%g, %g,%g, %g,%g)\n",x1,y1,x2,y2,x3,y3);
+  Vector cp1{x1, y1}, cp2{x2, y2}, cp3{x3, y3};
+  bc->bez.push_back(Bezier(bc->pos, cp1, cp2, cp3));
+  bc->pos = cp3;
+}
+
+// apparently this is called to indicate when the curve reaches each knot
+void bezctx_mark_knot(bezctx *bc, int knot_idx) {
+  // ipeDebug("bezctx_mark_knot()_%d\n",knot_idx);
+}
+
+// ------------------------------------------------------------------------------------------
+
+void Bezier::spiroSpline(int n,  const Vector *v, std::vector<Bezier> &result)
+{
+  std::vector<spiro_cp> spiroCp;
+  for (int i = 0; i < n; ++i) {
+    spiroCp.push_back({v[i].x, v[i].y, SPIRO_G2});
+  }
+  if (n > 2 && v[n-2] == v[n-1])
+    spiroCp.pop_back();
+  if (spiroCp.size() < 3) {
+    result.push_back(Bezier(v[0], v[0], v[1], v[1]));
+    return;
+  }
+
+  bezctx context{v[0], result};
+  if (SpiroCPsToBezier2(spiroCp.data(), spiroCp.size(),
+			SPIRO_CUBIC_TO_BEZIER, false, &context) != 1 ) {
+    // spiro failed, use another spline instead
+    Bezier::spline(n, v, result);
+  }
+}
+
 //! Return distance to Bezier spline.
 /*! But may just return \a bound if actual distance is larger.  The
   Bezier spline is approximated to a precision of 1.0, and the
@@ -893,6 +997,155 @@ bool Bezier::snap(const Vector &v, double &t, Vector &pos, double &bound) const
 
 // --------------------------------------------------------------------
 
+#ifdef IPE_GSL
+
+// Use Gnu Scientific Library to compute more precise intersection points
+// https://www.gnu.org/software/gsl/doc/html/multimin.html
+
+// Distance between points on two Bezier splines
+static double bezierDistance(const gsl_vector *v, void *params)
+{
+  Bezier *p = (Bezier *) params;
+  double t1 = gsl_vector_get(v, 0);
+  double t2 = gsl_vector_get(v, 1);
+  return (p[0].point(t1) - p[1].point(t2)).sqLen();
+}
+
+/* This uses a minimizer that doesn't need gradients.
+   Our function is polynomial, so one could work out the derivatives -
+   but this already works very well... */
+static Vector nearbyBezierIntersection(const Vector &q, const Bezier &a, const Bezier &b)
+{
+  Bezier par[2] = { a, b };
+
+  gsl_multimin_function minFunc;
+  minFunc.n = 2;
+  minFunc.f = bezierDistance;
+  minFunc.params = par;
+
+  // Starting point x
+  gsl_vector *x = gsl_vector_alloc(2);
+  gsl_vector_set(x, 0, (q-a.iV[0]).len() / (a.iV[3] - a.iV[0]).len());
+  gsl_vector_set(x, 1, (q-b.iV[0]).len() / (b.iV[3] - b.iV[0]).len());
+
+  // Initial step size
+  gsl_vector *ss = gsl_vector_alloc(2);
+  gsl_vector_set_all(ss, 0.1);
+
+  const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
+  gsl_multimin_fminimizer *s = gsl_multimin_fminimizer_alloc(T, 2);
+  gsl_multimin_fminimizer_set(s, &minFunc, x, ss);
+
+  size_t iter = 0;
+  int status;
+  do {
+    iter++;
+    status = gsl_multimin_fminimizer_iterate(s);
+    if (status) break;
+    double size = gsl_multimin_fminimizer_size(s);
+    status = gsl_multimin_test_size(size, 1e-4);
+#ifdef GSL_OUTPUT
+    if (status == GSL_SUCCESS) ipeDebug("Minimum obtained");
+    ipeDebug("minimize: %3ld t1=%.9f t2=%.9f sqDist=%.3e size=%.3e", iter,
+	     gsl_vector_get(s->x, 0), gsl_vector_get(s->x, 1), s->fval, size);
+#endif
+  } while (status == GSL_CONTINUE && iter < 100);
+  Vector result = (status == GSL_SUCCESS) ? a.point(gsl_vector_get(s->x, 0)) : q;
+  gsl_vector_free(x);
+  gsl_vector_free(ss);
+  gsl_multimin_fminimizer_free(s);
+  return result;
+}
+
+struct BezierLineParams {
+  Bezier b;
+  Line l;
+};
+
+// distance of a point on the Bezier spline from the line
+static double bezierLineDistance(double x, void * params)
+{
+  BezierLineParams *p = (BezierLineParams *) params;
+  return sq(p->l.distance(p->b.point(x)));
+}
+
+static Vector nearbyBezierLineIntersection(const Vector &q, const Bezier &a, const Line &l)
+{
+  BezierLineParams params { a, l };
+
+  gsl_function minFunc;
+  minFunc.function = &bezierLineDistance;
+  minFunc.params = &params;
+
+  const gsl_min_fminimizer_type *T = gsl_min_fminimizer_brent;
+  gsl_min_fminimizer *s = gsl_min_fminimizer_alloc(T);
+  gsl_min_fminimizer_set(s, &minFunc, (q-a.iV[0]).len() / (a.iV[3] - a.iV[0]).len(), 0.0, 1.0);
+
+  size_t iter = 0;
+  int status;
+  do {
+      iter++;
+      status = gsl_min_fminimizer_iterate(s);
+      if (status) break;
+      double x0 = gsl_min_fminimizer_x_lower(s);
+      double x1 = gsl_min_fminimizer_x_upper(s);
+      status = gsl_min_test_interval(x0, x1, 1e-4, 0.0);
+#ifdef GSL_OUTPUT
+      if (status == GSL_SUCCESS) ipeDebug("Minimum obtained");
+      ipeDebug("minimize: %3ld [%.7f, %.7f] %.7f", iter, x0, x1, gsl_min_fminimizer_x_minimum(s));
+#endif
+  } while (status == GSL_CONTINUE && iter < 100);
+  Vector result = (status == GSL_SUCCESS) ? a.point(gsl_min_fminimizer_x_minimum(s)) : q;
+  gsl_min_fminimizer_free(s);
+  return result;
+}
+
+struct BezierEllipseParams {
+  Bezier b;
+  Matrix m;
+};
+
+// distance of a point on the Bezier spline from an ellipse (in ellipse space)
+static double bezierEllipseDistance(double x, void * params)
+{
+  BezierEllipseParams *p = (BezierEllipseParams *) params;
+  return sq((p->m * p->b.point(x)).len() - 1.0);
+}
+
+static Vector nearbyBezierEllipseIntersection(const Vector &q, const Bezier &a, const Matrix &m)
+{
+  BezierEllipseParams params { a, m.inverse() };
+
+  gsl_function minFunc;
+  minFunc.function = &bezierEllipseDistance;
+  minFunc.params = &params;
+
+  const gsl_min_fminimizer_type *T = gsl_min_fminimizer_brent;
+  gsl_min_fminimizer *s = gsl_min_fminimizer_alloc(T);
+  gsl_min_fminimizer_set(s, &minFunc, (q-a.iV[0]).len() / (a.iV[3] - a.iV[0]).len(), 0.0, 1.0);
+
+  size_t iter = 0;
+  int status;
+  do {
+      iter++;
+      status = gsl_min_fminimizer_iterate(s);
+      if (status) break;
+      double x0 = gsl_min_fminimizer_x_lower(s);
+      double x1 = gsl_min_fminimizer_x_upper(s);
+      status = gsl_min_test_interval(x0, x1, 1e-4, 0.0);
+#ifdef GSL_OUTPUT
+      if (status == GSL_SUCCESS) ipeDebug("Minimum obtained");
+      ipeDebug("minimize: %3ld [%.7f, %.7f] %.7f", iter, x0, x1, gsl_min_fminimizer_x_minimum(s));
+#endif
+  } while (status == GSL_CONTINUE && iter < 100);
+  Vector result = (status == GSL_SUCCESS) ? a.point(gsl_min_fminimizer_x_minimum(s)) : q;
+  gsl_min_fminimizer_free(s);
+  return result;
+}
+#endif
+
+// --------------------------------------------------------------------
+
 /* Determines intersection point(s) of two cubic Bezier-Splines.
  * The found intersection points are stored in the vector intersections.
  */
@@ -924,7 +1177,11 @@ static void intersectBeziers(std::vector<Vector> &intersections,
     Segment bs = Segment(b.iV[0], b.iV[3]);
     Vector p;
     if (as.intersects(bs, p))
+#ifdef IPE_GSL
+      intersections.push_back(nearbyBezierIntersection(p, a, b));
+#else
       intersections.push_back(p);
+#endif
   } else {
     Bezier leftA, rightA, leftB, rightB;
     a.subdivide(leftA, rightA);
@@ -948,8 +1205,13 @@ void Bezier::intersect(const Line &l, std::vector<Vector> &result) const
 
   if (straight(BEZIER_INTERSECT_PRECISION)) {
     Vector p;
-    if (Segment(iV[0], iV[3]).intersects(l, p))
+    if (Segment(iV[0], iV[3]).intersects(l, p)) {
+#ifdef IPE_GSL
+      result.push_back(nearbyBezierLineIntersection(p, *this, l));
+#else
       result.push_back(p);
+#endif
+    }
   } else {
     Bezier leftA, rightA;
     subdivide(leftA, rightA);
@@ -1118,6 +1380,108 @@ void Arc::intersect(const Segment &s, std::vector<Vector> &result) const
   }
 }
 
+#ifdef IPE_GSL
+/* Computes intersection points between a unit SEMI-circle and the
+ * ellipse defined by the inverse of m.
+ *
+ * We write x = 2t/(1+t^2) and y = (1-t^2)/(1+t^2) for -1 <= t <= 1
+ * This parameterizes the points (x, y) on the upper unit semicircle.
+ * We map with respect to m to obtain (x1, y1), and set x1^2 + y1^2 - 1 = 0.
+ *
+ * If m = [ a, b, c, d, e, f ], the point m * (x, y) is ax + cy + e, bx + dy + f,
+ * so we set (ax + cy +e)^2 + (bx + dy + f)^2 - 1 = 0, where x, y are as above.
+ * aaxx + ccyy + ee + 2axcy + 2axe + 2cye + bbxx + ddyy + ff + 2bxdy + 2bxf + 2dyf - 1 = 0
+ * or (aa + bb)xx + (cc + dd)yy + (2ac + 2bd)xy + (2ae + 2bf)x + (2ce + 2df)y + (ee + ff - 1) = 0
+ *
+ * Substituting x and y using t and multiplying with (1+tt)^2, this gives
+ * (aa + bb)4tt + (cc + dd)(1-tt)(1-tt) + (2ac + 2bd)2t(1-tt) +
+ * (2ae + 2bf)2t(1+tt) + (2ce + 2df)(1-tt)(1+tt) + (ee + ff - 1)(1+tt)(1+tt) = 0
+ *
+ * Collecting terms of t:
+ *  t^4: cc + dd - (2ce + 2df) + (ee + ff - 1)
+ *  t^3: -2(2ac + 2bd) + 2(2ae + 2bf)
+ *  t^2: 4(aa + bb) - 2(cc + dd) + 2(ee + ff - 1)
+ *  t  : 2(2ac + 2bd) + 2(2ae + 2bf)
+ *  1  : (cc + dd) + (2ce + 2df) + (ee + ff - 1)
+ *
+ * We return points on the unit circle.
+ */
+static void ellipseIntersections(const Matrix &m, double ysign, std::vector<Vector> &pts)
+{
+  double a = m.a[0];
+  double b = m.a[1];
+  double c = ysign * m.a[2];
+  double d = ysign * m.a[3];
+  double e = m.a[4];
+  double f = m.a[5];
+
+  double t4 = c*c + d*d - (2*c*e + 2*d*f) + (e*e + f*f - 1);
+  double t3 = -2*(2*a*c + 2*b*d) + 2*(2*a*e + 2*b*f);
+  double t2 = 4*(a*a + b*b) - 2*(c*c + d*d) + 2*(e*e + f*f - 1);
+  double t1 = 2*(2*a*c + 2*b*d) + 2*(2*a*e + 2*b*f);
+  double t0 = (c*c + d*d) + (2*c*e + 2*d*f) + (e*e + f*f - 1);
+
+  // coefficients of P(x) =  t4 * t^4 + ... t1 * t + t0
+  double coeff[5] = { t0, t1, t2, t3, t4 };
+
+  int deg = 4;
+  while (deg > 0 && coeff[deg] == 0.0)
+    --deg;
+  if (!deg)  // either identical zero or never zero
+    return;
+
+  gsl_poly_complex_workspace *w = gsl_poly_complex_workspace_alloc(deg+1);
+
+  double z[8];
+  gsl_poly_complex_solve(coeff, deg+1, w, z);
+
+  gsl_poly_complex_workspace_free(w);
+
+  for (int i = 0; i < deg; i++) {
+#ifdef GSL_OUTPUT
+    ipeDebug("z%d = %+.18f %+.18f", i, z[2*i], z[2*i+1]);
+#endif
+    if (fabs(z[2*i + 1]) < 1e-12) {
+      // we take this as a real root
+      double t = z[2*i];
+      if (-1.0 - 1e-6 <= t && t <= 1.0 + 1e-6) {
+	double tt = t*t;
+	pts.push_back(Vector(2*t/(1+tt),ysign * (1-tt)/(1+tt)));
+      }
+    }
+  }
+}
+
+//! Compute intersection points of Arc with Arc.
+void Arc::intersect(const Arc &a, std::vector<Vector> &result) const
+{
+  if (!bbox().intersects(a.bbox()))
+    return;
+
+  std::vector<Vector> pts;
+  Matrix m = iM.inverse() * a.iM;
+  ellipseIntersections(m, +1.0, pts); // upper semi-circle
+  ellipseIntersections(m, -1.0, pts); // lower semi-circle
+
+  for (int i = 0; i < size(pts); ++i) {
+    Vector &p = pts[i];
+    if (p.angle().liesBetween(a.iAlpha, a.iBeta)) {
+      // if (1,0) or (-1,0) is an intersection point, it will be
+      // returned by both calls, filter out
+      int j = 0;
+      while (j < i && (pts[j] - p).sqLen() > 1e-12)
+	++j;
+      if (j == i) {
+	Vector q = m * p;
+	if (q.angle().liesBetween(iAlpha, iBeta))
+	  result.push_back(a.iM * p);
+      }
+    }
+  }
+}
+
+#else
+
 //! Compute intersection points of Arc with Arc.
 void Arc::intersect(const Arc &a, std::vector<Vector> &result) const
 {
@@ -1128,7 +1492,6 @@ void Arc::intersect(const Arc &a, std::vector<Vector> &result) const
    * of the partial Arcs overlap. If they do, they are subdivided
    * again and so on, until a special precision is achieved.
    */
-
   const double PRECISION = 0.05; // 0.05 is about ~2.8647 degrees
 
   if (!bbox().intersects(a.bbox()))
@@ -1148,6 +1511,7 @@ void Arc::intersect(const Arc &a, std::vector<Vector> &result) const
     ar.intersect(br, result);
   }
 }
+#endif
 
 //! Compute intersection points of Arc with Bezier.
 void Arc::intersect(const Bezier &b, std::vector<Vector> &result) const
@@ -1169,20 +1533,37 @@ void Arc::intersect(const Bezier &b, std::vector<Vector> &result) const
   if (!bbox().intersects(bboxB))
     return;
 
-  if (b.straight(PRECISION)) {
+  if (b.straight(BEZIER_INTERSECT_PRECISION)) {
+#ifdef IPE_GSL
+    std::vector<Vector> pts;
+    intersect(Segment(b.iV[0], b.iV[3]), pts);
+    for (const auto &q : pts) {
+      result.push_back(nearbyBezierEllipseIntersection(q, b, iM));
+    }
+#else
     intersect(Segment(b.iV[0], b.iV[3]), result);
+#endif
   } else {
     // is it really useful to divide the arc?
     // the hope is to achieve emptiness of intersection more quickly
-    Arc al, ar;
-    subdivide(al, ar);
-    Bezier bl, br;
-    b.subdivide(bl, br);
+    if (straight(PRECISION)) {
+      // do not subdivide the arc any further
+      Bezier bl, br;
+      b.subdivide(bl, br);
 
-    al.intersect(bl, result);
-    al.intersect(br, result);
-    ar.intersect(bl, result);
-    ar.intersect(br, result);
+      intersect(bl, result);
+      intersect(br, result);
+    } else {
+      Arc al, ar;
+      subdivide(al, ar);
+      Bezier bl, br;
+      b.subdivide(bl, br);
+
+      al.intersect(bl, result);
+      al.intersect(br, result);
+      ar.intersect(bl, result);
+      ar.intersect(br, result);
+    }
   }
 }
 
