@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -177,8 +177,18 @@ String PdfString::decode() const noexcept
     return iValue;
   String result;
   Lex lex(iValue);
-  while (!lex.eos())
-    result += char(lex.getHexByte());
+  if (iValue.hasPrefix("FEFF")) {
+    lex.getHexByte(); // skip unicode mark
+    lex.getHexByte();
+    while (!lex.eos()) {
+      int ch = lex.getHexByte() << 8;
+      ch |= lex.getHexByte();
+      result.appendUtf8(ch);
+    }
+  } else {
+    while (!lex.eos())
+      result += char(lex.getHexByte());
+  }
   return result;
 }
 
@@ -358,6 +368,33 @@ const PdfObj *PdfDict::get(String key, const PdfFile *file) const noexcept
   return nullptr; // not in dictionary
 }
 
+//! Look up key and return if it is a dictionary.
+const PdfDict *PdfDict::getDict(String key, const PdfFile *file) const noexcept
+{
+  const PdfObj *obj = get(key, file);
+  if (obj && obj->dict())
+    return obj->dict();
+  return nullptr;
+}
+
+//! Look up key and return if it is an array.
+const PdfArray *PdfDict::getArray(String key, const PdfFile *file) const noexcept
+{
+  const PdfObj *obj = get(key, file);
+  if (obj && obj->array())
+    return obj->array();
+  return nullptr;
+}
+
+//! Look up key and return its value if it is a /Name, otherwise return empty string.
+String PdfDict::getName(String key, const PdfFile *file) const noexcept
+{
+  const PdfObj *obj = get(key, file);
+  if (obj && obj->name())
+    return obj->name()->value();
+  return String();
+}
+
 //! Retrieve a single number and stor in \a val.
 bool PdfDict::getNumber(String key, double &val, const PdfFile *file)
   const noexcept
@@ -369,6 +406,16 @@ bool PdfDict::getNumber(String key, double &val, const PdfFile *file)
   return true;
 }
 
+//! Retrieve a single integer.
+/*! Returns -1 on failure. */
+int PdfDict::getInteger(String key, const PdfFile *file) const noexcept
+{
+  double val;
+  if (!getNumber(key, val, file))
+    return -1;
+  return int(val);
+}
+
 //! Retrieve an array of numbers and store in \a vals.
 bool PdfDict::getNumberArray(String key, const PdfFile *file,
 			     std::vector<double> &vals) const noexcept
@@ -436,7 +483,6 @@ Buffer PdfDict::inflate() const noexcept
 PdfParser::PdfParser(DataSource &source)
   : iSource(source)
 {
-  iPos = 0;
   getChar();  // init iCh
   getToken(); // init iTok
 }
@@ -610,7 +656,7 @@ PdfArray *PdfParser::makeArray()
   }
 }
 
-PdfDict *PdfParser::makeDict()
+PdfDict *PdfParser::makeDict(bool lateStream)
 {
   std::unique_ptr<PdfDict> dict(new PdfDict);
   for (;;) {
@@ -625,12 +671,18 @@ PdfDict *PdfParser::makeDict()
       // time to read the stream
       while (!eos() && iCh != '\n')
 	getChar();
+      int streamPos = iSource.position();
       getChar(); // skip '\n'
       // now at beginning of stream
-      const PdfObj *len = dict->get("Length", nullptr);
+      const PdfObj *len = dict->get("Length");
       if (len && len->ref()) {
-	ipeDebug("/Length entry of dictionary is a reference.");
-	return nullptr;
+	if (lateStream) {
+	  dict->setLateStream(streamPos);
+	  return dict.release();
+	} else {
+	  ipeDebug("/Length entry of dictionary is a reference");
+	  return nullptr;
+	}
       }
       if (!len || !len->number())
 	return nullptr;
@@ -679,7 +731,7 @@ PdfDict *PdfParser::makeDict()
 }
 
 //! Read one object from input stream.
-PdfObj *PdfParser::getObject()
+PdfObj *PdfParser::getObject(bool lateStream)
 {
   PdfToken tok = iTok;
   getToken();
@@ -702,7 +754,7 @@ PdfObj *PdfParser::getObject()
   case PdfToken::EArrayBg:
     return makeArray();
   case PdfToken::EDictBg:
-    return makeDict();
+    return makeDict(lateStream);
     // anything else is an error
   case PdfToken::EErr:
   default:
@@ -711,7 +763,7 @@ PdfObj *PdfParser::getObject()
 }
 
 //! Parse an object definition (current token is object number).
-PdfObj *PdfParser::getObjectDef()
+PdfObj *PdfParser::getObjectDef(bool lateStream)
 {
   getToken();
   if (iTok.iType != PdfToken::ENumber || iTok.iString != "0")
@@ -720,9 +772,11 @@ PdfObj *PdfParser::getObjectDef()
   if (iTok.iType != PdfToken::EOp || iTok.iString != "obj")
     return nullptr;
   getToken();
-  PdfObj *obj = getObject();
+  PdfObj *obj = getObject(lateStream);
   if (!obj)
     return nullptr;
+  if (obj->dict() && obj->dict()->lateStream() > 0)
+    return obj;
   if (iTok.iType != PdfToken::EOp || iTok.iString != "endobj")
     return nullptr;
   getToken();
@@ -737,12 +791,31 @@ void PdfParser::skipXRef()
   int k = toInt(iTok.iString);
   getToken();
   while (k--) {
-    getToken(); // obj num
+    getToken(); // position
     getToken(); // gen num
     getToken(); // n or f
   }
 }
 
+//! Read xref table (current token is 'xref')
+std::vector<int> PdfParser::readXRef()
+{
+  getToken(); // first object number
+  getToken(); // number of objects
+  int k = toInt(iTok.iString);
+  std::vector<int> objects(k, 0); // fill with zeros
+  getToken(); // position
+  for (int obj = 0; obj < k; ++obj) {
+    int pos = toInt(iTok.iString);
+    getToken(); // gen num
+    getToken(); // n or f
+    if (iTok.iString == "n")
+      objects[obj] = pos;
+    getToken(); // next token
+  }
+  return objects;
+}
+
 //! Parse trailer dictionary (current token is 'trailer')
 PdfDict *PdfParser::getTrailer()
 {
@@ -750,7 +823,63 @@ PdfDict *PdfParser::getTrailer()
   if (iTok.iType != PdfToken::EDictBg)
     return nullptr;
   getToken();
-  return makeDict();
+  return makeDict(false);
+}
+
+// --------------------------------------------------------------------
+
+static bool addStreamToDict(DataSource &source, PdfDict *d, const PdfFile *file)
+{
+  int pos = d->lateStream();
+  if (pos == 0)
+    return true;
+  source.setPosition(pos);
+
+  int bytes = d->getInteger("Length", file);
+  if (bytes < 0)
+    return false;
+
+  Buffer buf(bytes);
+  char *p = buf.data();
+  while (bytes--) {
+    *p++ = source.getChar();
+  }
+  d->setStream(buf);
+  d->setLateStream(0);
+  PdfParser parser(source);
+  PdfToken t = parser.token();
+  return (t.iType == PdfToken::EOp && t.iString == "endstream");
+}
+
+static int readBytes(DataSource &source, int n)
+{
+  int value = 0;
+  while (n--) {
+    int ch = uint8_t(source.getChar());
+    value = value * 256 + ch;
+  }
+  return value;
+}
+
+// return size or -1
+static int checkXRefObj(const PdfDict *d)
+{
+  const PdfObj *type = d->get("Type");
+  if (!type || !type->name() || type->name()->value() != "XRef")
+    return -1;
+  int size = d->getInteger("Size");
+  if (size < 0)
+    return -1;
+  const PdfObj *indexObj = d->get("Index");
+  if (indexObj) {
+    const PdfArray *index = indexObj->array();
+    if (!index || index->count() != 2) return false;
+    const PdfObj *val = index->obj(0, nullptr);
+    if (!val->number() || val->number()->value() != 0.0) return -1;
+    val = index->obj(1, nullptr);
+    if (!val->number() || val->number()->value() != size) return -1;
+  }
+  return size;
 }
 
 // --------------------------------------------------------------------
@@ -760,42 +889,137 @@ PdfDict *PdfParser::getTrailer()
  * \brief All information obtained by parsing a PDF file.
  */
 
-bool PdfFile::parseObjectStream(const PdfDict *d)
-{
-  const PdfObj *objn = d->get("N", this);
-  const PdfObj *objfirst = d->get("First", this);
-  int n = objn->number() ? objn->number()->value() : -1;
-  int first = objfirst->number() ? objfirst->number()->value() : -1;
-  if (n < 0 || first < 0)
-    return false;
-  Buffer stream = d->inflate();
-  BufferSource source(stream);
-  PdfParser parser(source);
-  std::vector<int> dir;
-  for (int i = 0; i < 2 * n; ++i) {
-    PdfToken t = parser.token();
-    if (t.iType != PdfToken::ENumber)
-      return false;
-    dir.push_back(toInt(t.iString));
-    parser.getToken();
-  }
-  for (int i = 0; i < n; ++i) {
-    int num = dir[2*i];
-    source.setPosition(first + dir[2*i+1]);
-    parser.getChar();
-    parser.getToken();
-    PdfObj *obj = parser.getObject();
-    if (!obj)
-      return false;
-    // ipeDebug("Object: %s", obj->repr().z());
-    iObjects[num] = std::unique_ptr<const PdfObj>(obj);
-  }
-  return true;
-}
-
 //! Parse entire PDF stream, and store objects.
 bool PdfFile::parse(DataSource &source)
 {
+  int length = source.length();
+  if (length < 0)
+    // could not seek to end
+    return parseSequentially(source);
+
+  if (length < 400)
+    return false;
+
+  source.setPosition(length - 40);
+  String s;
+  int ch;
+  while ((ch = source.getChar()) != EOF)
+    s += ch;
+  int i = s.find("startxref");
+  if (i < 0)
+    return parseSequentially(source);
+
+  Lex lex(s.substr(i+9));
+  int xrefPos = lex.getInt();
+  source.setPosition(xrefPos);
+
+  PdfParser parser(source);
+  PdfToken t = parser.token();
+
+  if (t.iType == PdfToken::ENumber)
+    // seems to be the /XRef object
+    return parseFromXRefObj(parser, source);
+
+  if (t.iType != PdfToken::EOp || t.iString != "xref")
+    return parseSequentially(source);
+
+  std::vector<int> xref = parser.readXRef();
+  t = parser.token();
+
+  if (t.iType != PdfToken::EOp || t.iString != "trailer")
+    return false;
+
+  iTrailer = std::unique_ptr<const PdfDict>(parser.getTrailer());
+  if (!iTrailer)
+    return false;
+
+  std::vector<int> delayed;
+  // read objects
+  for (int num = 0; num < size(xref); ++num) {
+    if (xref[num] > 0) {
+      source.setPosition(xref[num]);
+      PdfParser objParser(source);
+      std::unique_ptr<const PdfObj> obj(objParser.getObjectDef(true));
+      if (!obj) {
+	ipeDebug("Failed to get object %d", num);
+	return false;
+      }
+      if (obj->dict() && obj->dict()->lateStream() > 0)
+	delayed.push_back(num);
+      // ipeDebug("Object: %s", obj->repr().z());
+      iObjects[num] = std::move(obj);
+    }
+  }
+  return readDelayedStreams(delayed, source);
+}
+
+// read all the streams we had to postpone
+bool PdfFile::readDelayedStreams(std::vector<int> &delayed, DataSource &source)
+{
+  for (int num : delayed) {
+    const PdfDict *d = iObjects[num]->dict();
+    if (!addStreamToDict(source, (PdfDict *) d, this)) {
+      ipeDebug("Failed to read stream for object %d", num);
+      return false;
+    }
+  }
+  return readPageTree();
+}
+
+// ------------------------------------------------------------------------------------------
+
+bool PdfFile::parseFromXRefObj(PdfParser &parser, DataSource &source)
+{
+  std::unique_ptr<const PdfObj> obj(parser.getObjectDef(false));
+  if (!obj)
+    return false;
+  iTrailer = std::unique_ptr<const PdfDict>(obj.release()->dict());
+  int size = checkXRefObj(iTrailer.get());
+  std::vector<double> w;
+
+  if (size < 0 || !iTrailer->getNumberArray("W", nullptr, w) || w.size() != 3)
+    return parseSequentially(source);
+
+  std::vector<int> delayed;
+
+  Buffer stream = iTrailer->inflate();
+  BufferSource xb(stream);
+  for (int num = 0; num < size; ++num) {
+    int objType = readBytes(xb, int(w[0]));
+    int pos = readBytes(xb, int(w[1]));
+    readBytes(xb, int(w[2]));  // not used
+    if (objType == 1) {
+      source.setPosition(pos);
+      PdfParser objParser(source);
+      std::unique_ptr<const PdfObj> obj(objParser.getObjectDef(true));
+      if (!obj) {
+	ipeDebug("Failed to get object %d from XRef object", num);
+	return false;
+      }
+      const PdfDict *d = obj->dict();
+      const PdfObj *type = d ? d->get("Type", this) : nullptr;
+      if (type && type->name() && type->name()->value() == "ObjStm") {
+	if (!parseObjectStream(d))
+	  return false;
+      } else {
+	// ipeDebug("Object: %s", obj->repr().z());
+	if (obj->dict() && obj->dict()->lateStream() > 0)
+	  delayed.push_back(num);
+	iObjects[num] = std::move(obj);
+      }
+    }
+  }
+  return readDelayedStreams(delayed, source);
+}
+
+// ------------------------------------------------------------------------------------------
+
+// Read PDF file from front to back, collecting objects
+// Used when startxref access and /XRef object access fails
+bool PdfFile::parseSequentially(DataSource &source)
+{
+  ipeDebug("Falling back on sequential PDF parser");
+  source.setPosition(0);
   PdfParser parser(source);
 
   for (;;) {
@@ -804,9 +1028,9 @@ bool PdfFile::parse(DataSource &source)
     if (t.iType == PdfToken::ENumber) {
       // <num> 0 obj starts an object
       int num = toInt(t.iString);
-      std::unique_ptr<const PdfObj> obj(parser.getObjectDef());
+      std::unique_ptr<const PdfObj> obj(parser.getObjectDef(false));
       if (!obj) {
-	ipeDebug("Failed to get object %d", num);
+	ipeDebug("Failed to get object %d in sequential reader", num);
 	return false;
       }
       const PdfDict *d = obj->dict();
@@ -845,6 +1069,43 @@ bool PdfFile::parse(DataSource &source)
   }
 }
 
+// ------------------------------------------------------------------------------------------
+
+bool PdfFile::parseObjectStream(const PdfDict *d)
+{
+  const PdfObj *objn = d->get("N", this);
+  const PdfObj *objfirst = d->get("First", this);
+  int n = objn->number() ? objn->number()->value() : -1;
+  int first = objfirst->number() ? objfirst->number()->value() : -1;
+  if (n < 0 || first < 0)
+    return false;
+  Buffer stream = d->inflate();
+  BufferSource source(stream);
+  PdfParser parser(source);
+  std::vector<int> dir;
+  for (int i = 0; i < 2 * n; ++i) {
+    PdfToken t = parser.token();
+    if (t.iType != PdfToken::ENumber)
+      return false;
+    dir.push_back(toInt(t.iString));
+    parser.getToken();
+  }
+  for (int i = 0; i < n; ++i) {
+    int num = dir[2*i];
+    source.setPosition(first + dir[2*i+1]);
+    parser.getChar();
+    parser.getToken();
+    PdfObj *obj = parser.getObject();
+    if (!obj)
+      return false;
+    // ipeDebug("Object: %s", obj->repr().z());
+    iObjects[num] = std::unique_ptr<const PdfObj>(obj);
+  }
+  return true;
+}
+
+// ------------------------------------------------------------------------------------------
+
 //! Return object with number \a num.
 const PdfObj *PdfFile::object(int num) const noexcept
 {
@@ -881,21 +1142,24 @@ bool PdfFile::readPageTree(const PdfObj *ptn)
     ptn = catalog()->get("Pages", this);
   if (!ptn || !ptn->dict())
     return false;
-  const PdfObj *kids = ptn->dict()->get("Kids", this);
-  if (!kids || !kids->array())
+  const PdfArray *kids = ptn->dict()->getArray("Kids", this);
+  if (!kids)
     return false;
-  for (int i = 0; i < kids->array()->count(); ++i) {
-    const PdfObj *page = kids->array()->obj(i, this);
+  for (int i = 0; i < kids->count(); ++i) {
+    const PdfObj *pageRef = kids->obj(i, nullptr);
+    if (!pageRef || !pageRef->ref())
+      return false;
+    int pageObjNum = pageRef->ref()->value();
+    const PdfObj *page = object(pageObjNum);
     if (!page || !page->dict())
       return false;
-    const PdfObj *type = page->dict()->get("Type", this);
-    if (!type || !type->name())
-      return false;
-    if (type->name()->value() == "Pages")
+    String type = page->dict()->getName("Type", this);
+    if (type == "Pages")
       readPageTree(page);
-    else if (type->name()->value() == "Page")
+    else if (type == "Page") {
       iPages.push_back(page->dict());
-    else
+      iPageObjectNumbers.push_back(pageObjNum);
+    } else
       return false;
   }
   return true;
@@ -909,6 +1173,17 @@ const PdfDict *PdfFile::page(int pno) const noexcept
   return iPages[pno];
 }
 
+//! Return page number given the PDF object number.
+/*! Returns -1 if the object number is not a page. */
+int PdfFile::findPageFromPageObjectNumber(int objNum) const
+{
+  for (int i = 0; i < size(iPageObjectNumbers); ++i) {
+    if (iPageObjectNumbers[i] == objNum)
+      return i;
+  }
+  return -1;
+}
+
 //! Return mediabox of a page.
 Rect PdfFile::mediaBox(const PdfDict *pg) const
 {
