@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -474,7 +474,7 @@ void PdfWriter::embedResources()
     for (auto &num : seq) {
       const PdfObj *obj = iResources->object(num);
       int embedNum = startObject();
-      if (obj->dict() && obj->dict()->get("IpeId", nullptr))
+      if (iResources->isIpeXForm(num) && obj->dict())
 	embedIpeXForm(obj->dict());
       else
 	obj->write(iStream, &iResourceNumber, inflate);
@@ -744,6 +744,8 @@ void PdfWriter::paintView(Stream &stream, int pno, int view)
 {
   const Page *page = iDoc->page(pno);
   PdfPainter painter(iDoc->cascade(), stream);
+  painter.setAttributeMap(&page->viewMap(view));
+  std::vector<Matrix> layerMatrices = page->layerMatrices(view);
 
   const Symbol *background =
     iDoc->cascade()->findSymbol(Attribute::BACKGROUND());
@@ -761,8 +763,12 @@ void PdfWriter::paintView(Stream &stream, int pno, int view)
     title->draw(painter);
 
   for (int i = 0; i < page->count(); ++i) {
-    if (page->objectVisible(view, i))
+    if (page->objectVisible(view, i)) {
+      painter.pushMatrix();
+      painter.transform(layerMatrices[page->layerOf(i)]);
       page->object(i)->draw(painter);
+      painter.popMatrix();
+    }
   }
 }
 
