@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@
 #include "ipeattributes.h"
 
 #ifdef WIN32
+#define NTDDI_VERSION 0x06000000
 #include <windows.h>
 #include <shlobj.h>
 #include <direct.h>
@@ -56,6 +57,7 @@
 #include <unistd.h>
 #include <clocale>
 #include <cstring>
+#include <cerrno>
 
 using namespace ipe;
 
@@ -89,12 +91,12 @@ static String appImageBase;
 #ifdef WIN32
 static ULONG_PTR gdiplusToken = 0;
 _locale_t ipeLocale;
+/*
 typedef _locale_t (*LPCreateLocale)(int category, const char *locale);
 static LPCreateLocale p_create_locale = nullptr;
 typedef void (*LPFreeLocale)(_locale_t locale);
 static LPFreeLocale p_free_locale = nullptr;
-typedef double (*LPStrtodL)(const char *strSource, char **endptr, _locale_t locale);
-static LPStrtodL p_strtod_l = nullptr;
+*/
 #else
 locale_t ipeLocale;
 #endif
@@ -158,8 +160,10 @@ static void shutdownIpelib()
 {
 #ifdef WIN32
   Gdiplus::GdiplusShutdown(gdiplusToken);
+  /*
   if (p_create_locale != nullptr)
-    p_free_locale(ipeLocale);
+  */
+  _free_locale(ipeLocale);
 #else
   freelocale(ipeLocale);
 #endif
@@ -196,14 +200,15 @@ void Platform::initLib(int version)
     appImageBase = ".";
 #endif
 #ifdef WIN32
+  /*
   HMODULE hDll = LoadLibraryA("msvcrt.dll");
   if (hDll) {
     p_create_locale = (LPCreateLocale) GetProcAddress(hDll, "_create_locale");
     p_free_locale = (LPFreeLocale) GetProcAddress(hDll, "_free_locale");
-    p_strtod_l = (LPStrtodL) GetProcAddress(hDll, "_strtod_l");
   }
   if (p_create_locale != nullptr)
-    ipeLocale = p_create_locale(LC_NUMERIC, "C");
+  */
+  ipeLocale = _create_locale(LC_NUMERIC, "C");
   Gdiplus::GdiplusStartupInput gdiplusStartupInput;
   Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, nullptr);
 #else
@@ -357,10 +362,10 @@ String Platform::latexDirectory()
     if (latexDir.left(4) == "ipe:")
       latexDir = ipeDrive() + latexDir.substr(4);
   } else {
-    wchar_t szPath[MAX_PATH];
-    if (SUCCEEDED(SHGetFolderPathW(nullptr, CSIDL_LOCAL_APPDATA,
-				   nullptr, 0, szPath))) {
-      latexDir = String(szPath) + "\\ipe";
+    wchar_t *path;
+    if (SUCCEEDED(SHGetKnownFolderPath(FOLDERID_LocalAppData, 0, nullptr, &path))) {
+      latexDir = String(path) + "\\ipe";
+      CoTaskMemFree(path);
     } else {
       p = _wgetenv(L"LOCALAPPDATA");
       if (p)
@@ -589,7 +594,9 @@ int Platform::runLatex(String dir, LatexType engine, String docname) noexcept
   }
   if (online) {
 #ifdef __APPLE__
-    s += ipeDir("../MacOS", "ipecurl ");
+    s += "\"";
+    s += ipeDir("../MacOS", "ipecurl");
+    s += "\" ";
 #else
     s += "ipecurl ";
 #endif
@@ -653,10 +660,10 @@ String::String(const wchar_t *wbuf)
 double Platform::toDouble(String s)
 {
 #ifdef WIN32
-  if (p_create_locale != nullptr)
-    return _strtod_l(s.z(), nullptr, ipeLocale);
-  else
-    return strtod(s.z(), nullptr);
+  // if (p_create_locale != nullptr)
+  return _strtod_l(s.z(), nullptr, ipeLocale);
+  // else
+  // return strtod(s.z(), nullptr);
 #else
   return strtod_l(s.z(), nullptr, ipeLocale);
 #endif
@@ -671,10 +678,10 @@ int Platform::toNumber(String s, int &iValue, double &dValue)
   if (*fin == '\0')
     return 1; // integer
 #ifdef WIN32
-  if (p_create_locale != nullptr)
-    dValue = _strtod_l(s.z(), &fin, ipeLocale);
-  else
-    dValue = strtod(s.z(), &fin);
+  // if (p_create_locale != nullptr)
+  dValue = _strtod_l(s.z(), &fin, ipeLocale);
+  // else
+  // dValue = strtod(s.z(), &fin);
 #else
   dValue = strtod_l(s.z(), &fin, ipeLocale);
 #endif
