@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -54,9 +54,9 @@ inline void snapBezier(const Vector &mouse, const Bezier &bez,
   \ingroup geo
   \brief A segment on an SubPath.
 
-  A segment is either an elliptic arc, a straight segment, or a
-  B-spline curve, depending on its type().  This is a lightweight
-  object, created on the fly by Curve::segment().  There is no public
+  A segment is either an elliptic arc, a straight segment, or a spline
+  curve, depending on its type().  This is a lightweight object,
+  created on the fly by Curve::segment().  There is no public
   constructor, so the only way to create such an object is through
   that method.
 
@@ -72,6 +72,13 @@ inline void snapBezier(const Vector &mouse, const Bezier &bez,
     spline with 3 control points is defined to be a quadratic Bezier
     curve with those control points.
 
+  - \c ECardinalSpline: a cardinal spline curve with n control points.
+    The curve passes through the control points in the given order.
+    The tension is currently fixed at 0.5.
+
+  - \c ESpiroSpline: a clothoid spline curve with n control points,
+    computed using Raph Levien's spiro code (through libspiro).
+
   - \c EOldSpline: an incorrectly defined B-spline, used by Ipe for
     many years.  Supported for compatibility.
 
@@ -89,11 +96,15 @@ inline void snapBezier(const Vector &mouse, const Bezier &bez,
 
 //! Create a segment.
 /*! Matrix \a m defaults to null, for all segments but arcs. */
-CurveSegment::CurveSegment(Type type, int num, const Vector *cp,
-			   const Matrix *m)
-  : iType(type), iCP(cp), iNumCP(num), iM(m)
+CurveSegment::CurveSegment(const Curve *curve, int index)
+  : iCurve(curve), index(index)
 {
-  // nothing
+  if (type() == ESpiroSpline) {
+    iNumCP = iCurve->iSeg[index].iLastCP - iCurve->iSeg[index].iBezier;
+  } else {
+    int cpbg = (index > 0) ? iCurve->iSeg[index-1].iLastCP : 0;
+    iNumCP = iCurve->iSeg[index].iLastCP - cpbg + 1;
+  }
 }
 
 //! Return segment as Arc.
@@ -101,16 +112,43 @@ CurveSegment::CurveSegment(Type type, int num, const Vector *cp,
 Arc CurveSegment::arc() const
 {
   assert(type() == EArc);
-  return Arc(*iM, cp(0), cp(1));
+  return Arc(matrix(), cp(0), cp(1));
 }
 
-//! Convert B-spline to a sequence of Bezier splines.
+//! Tension (if Type() == ECardinalSpline)
+/*! Panics if segment is not a cardinal spline. */
+float CurveSegment::tension() const
+{
+  assert(type() == ECardinalSpline);
+  return iCurve->iSeg[index].iTension;
+}
+
+//! Convert spline curve to a sequence of Bezier splines.
 void CurveSegment::beziers(std::vector<Bezier> &bez) const
 {
-  if (iType == EOldSpline)
-    Bezier::oldSpline(countCP(), iCP, bez);
-  else
-    Bezier::spline(countCP(), iCP, bez);
+  switch (type()) {
+  case EOldSpline:
+    Bezier::oldSpline(countCP(), cps(), bez);
+    break;
+  case ESpline:
+    Bezier::spline(countCP(), cps(), bez);
+    break;
+  case ECardinalSpline:
+    Bezier::cardinalSpline(countCP(), cps(), tension(), bez);
+    break;
+  case ESpiroSpline: {
+    // use pre-computed Bezier curve
+    int cpbg = (index > 0) ? iCurve->iSeg[index-1].iLastCP : 0;
+    const Vector *cp = iCurve->iCP.data() + cpbg;
+    const Vector *fin = iCurve->iCP.data() + iCurve->iSeg[index].iBezier - 1;
+    while (cp < fin) {
+      bez.push_back(Bezier(cp[0], cp[1], cp[2], cp[3]));
+      cp += 3;
+    }
+    break; }
+  default:
+    break;
+  }
 }
 
 //! Draw the segment.
@@ -123,7 +161,9 @@ void CurveSegment::draw(Painter &painter) const
     painter.lineTo(cp(1));
     break;
   case EOldSpline:
-  case ESpline: {
+  case ESpline:
+  case ECardinalSpline:
+  case ESpiroSpline: {
     std::vector<Bezier> bez;
     beziers(bez);
     for (const auto & b : bez)
@@ -157,6 +197,8 @@ void CurveSegment::addToBBox(Rect &box, const Matrix &m, bool cpf) const
     break;
   case ESpline:
   case EOldSpline:
+  case ECardinalSpline:
+  case ESpiroSpline:
     if (cpf) {
       for (int i = 0; i < countCP(); ++i)
 	box.addPoint(m * cp(i));
@@ -180,7 +222,9 @@ double CurveSegment::distance(const Vector &v, const Matrix &m,
   case EArc:
     return (m * arc()).distance(v, bound);
   case EOldSpline:
-  case ESpline: {
+  case ESpline:
+  case ECardinalSpline:
+  case ESpiroSpline: {
     std::vector<Bezier> bez;
     beziers(bez);
     double d = bound;
@@ -218,6 +262,8 @@ void CurveSegment::snapVtx(const Vector &mouse, const Matrix &m,
       snapVertex(mouse, m * cp(1), pos, bound);
     break;
   case ESpline:
+  case ECardinalSpline:
+  case ESpiroSpline:
   case EOldSpline:
     // real end point is cp(countCP() - 1)
     // snap to all control points
@@ -247,8 +293,10 @@ void CurveSegment::snapBnd(const Vector &mouse, const Matrix &m,
       pos = pos1;
     }
     break; }
+  case ESpline:
   case EOldSpline:
-  case ESpline: {
+  case ECardinalSpline:
+  case ESpiroSpline: {
     std::vector<Bezier> bez;
     beziers(bez);
     for (const auto & b : bez)
@@ -299,10 +347,11 @@ void Curve::appendArc(const Matrix &m, const Vector &v0, const Vector &v1)
   iSeg.push_back(seg);
 }
 
-//! Append B-spline curve.
+//! Append spline curve.
 void Curve::appendSpline(const std::vector<Vector> &v, CurveSegment::Type type)
 {
   assert(type == CurveSegment::ESpline ||
+	 type == CurveSegment::ECardinalSpline ||
 	 type == CurveSegment::EOldSpline);
   if (iSeg.empty())
     iCP.push_back(v[0]);
@@ -312,14 +361,74 @@ void Curve::appendSpline(const std::vector<Vector> &v, CurveSegment::Type type)
   Seg seg;
   seg.iType = type;
   seg.iLastCP = iCP.size() - 1;
-  seg.iMatrix = iM.size() - 1;
+  seg.iMatrix = 0;
+  iSeg.push_back(seg);
+}
+
+//! Append a cardinal spline curve.
+void Curve::appendCardinalSpline(const std::vector<Vector> &v, float tension)
+{
+  appendSpline(v, CurveSegment::ECardinalSpline);
+  iSeg.back().iTension = tension;
+}
+
+//! Append a spiro spline curve.
+void Curve::appendSpiroSpline(const std::vector<Vector> &v)
+{
+  if (iSeg.empty())
+    iCP.push_back(v[0]);
+  assert(v[0] == iCP.back());
+  // compute Bezier representation
+  std::vector<Bezier> bez;
+  Bezier::spiroSpline(v.size(), v.data(), bez);
+  // save Bezier cps
+  for (const auto & b : bez) {
+    iCP.push_back(b.iV[1]);
+    iCP.push_back(b.iV[2]);
+    iCP.push_back(b.iV[3]);
+  }
+  int bezIndex = iCP.size() - 1;
+  // now save spiro control points, including first one
+  for (int i = 0; i < size(v); ++i)
+    iCP.push_back(v[i]);
+  Seg seg;
+  seg.iType = CurveSegment::ESpiroSpline;
+  seg.iLastCP = iCP.size() - 1;
+  seg.iBezier = bezIndex;
+  iSeg.push_back(seg);
+}
+
+//! Append a spiro spline curve with precomputed Bezier control points.
+void Curve::appendSpiroSplinePrecomputed(const std::vector<Vector> &v, int sep)
+{
+  if (iSeg.empty())
+    iCP.push_back(v[0]);
+  assert(v[0] == iCP.back());
+  // add Bezier representation
+  for (int i = 1; i < sep; ++i)
+    iCP.push_back(v[i]);
+  // need to insert final target and origin point
+  iCP.push_back(v.back());
+  int bezIndex = iCP.size() - 1;
+  iCP.push_back(v.front());
+  // now save actual spiro control points, including first one
+  for (int i = sep; i < size(v); ++i)
+    iCP.push_back(v[i]);
+  Seg seg;
+  seg.iType = CurveSegment::ESpiroSpline;
+  seg.iLastCP = iCP.size() - 1;
+  seg.iBezier = bezIndex;
   iSeg.push_back(seg);
 }
 
 //! Set whether subpath is closed or not.
+/*! Must be called after all segments have been added to the path. */
 void Curve::setClosed(bool closed)
 {
+  assert(!iSeg.empty() && !iClosed);
   iClosed = closed;
+  if (iClosed)
+    appendSegment(iCP.back(), iCP.front());
 }
 
 SubPath::Type Curve::type() const
@@ -333,19 +442,14 @@ const Curve *Curve::asCurve() const
 }
 
 //! Return segment.
-/*! If \a i is negative, elements from the end are returned.  The
-  closing segment of a closed path is not accessible this way (use
-  closingSegment() instead)!
+/*! If \a i is negative, elements from the end are returned.  If
+  i == countSegments(), the closing segment of a closed path is returned.
 */
 CurveSegment Curve::segment(int i) const
 {
   if (i < 0)
     i += iSeg.size();
-  const Seg &seg = iSeg[i];
-  const Matrix *m = &iM[seg.iMatrix];
-  int cpbg = (i > 0) ? iSeg[i-1].iLastCP : 0;
-  const Vector *cp = &iCP[cpbg];
-  return CurveSegment(seg.iType, seg.iLastCP - cpbg + 1, cp, m);
+  return CurveSegment(this, i);
 }
 
 void Curve::save(Stream &stream) const
@@ -354,8 +458,8 @@ void Curve::save(Stream &stream) const
   stream << iCP[0] << " m\n";
   int vtx = 1; // next control point
   int mat = 0;
-  for (std::vector<Seg>::const_iterator it = iSeg.begin();
-       it != iSeg.end(); ++it) {
+  std::vector<Seg>::const_iterator fin = iClosed ? iSeg.end() - 1 : iSeg.end();
+  for (std::vector<Seg>::const_iterator it = iSeg.begin(); it != fin; ++it) {
     switch (it->iType) {
     case CurveSegment::ESegment:
       assert(vtx == it->iLastCP);
@@ -375,6 +479,20 @@ void Curve::save(Stream &stream) const
 	stream << iCP[vtx++] << "\n";
       stream << iCP[vtx++] << " c\n";
       break;
+    case CurveSegment::ECardinalSpline:
+      while (vtx < it->iLastCP)
+	stream << iCP[vtx++] << "\n";
+      stream << iCP[vtx++] << " " << it->iTension << " C\n";
+      break;
+    case CurveSegment::ESpiroSpline: {
+      while (vtx < it->iBezier - 1)
+	stream << iCP[vtx++] << "\n";
+      stream << iCP[it->iBezier - 1] << " *\n";
+      vtx = it->iBezier + 2; // skip repeated last and first cp
+      while (vtx < it->iLastCP)
+	stream << iCP[vtx++] << "\n";
+      stream << iCP[vtx++] << " L\n"; // in honor of Raph Levien
+      break; }
     }
   }
   if (closed())
@@ -400,17 +518,11 @@ double Curve::distance(const Vector &v, const Matrix &m,
 		       double bound) const
 {
   double d = bound;
-  for (int i = 0; i < countSegments(); ++i) {
+  for (int i = 0; i < countSegmentsClosing(); ++i) {
     double d1 = segment(i).distance(v, m, d);
     if (d1 < d)
       d = d1;
   }
-  if (closed()) {
-    Vector u[2];
-    double d1 = closingSegment(u).distance(v, m , d);
-    if (d1 < d)
-      d = d1;
-  }
   return d;
 }
 
@@ -430,26 +542,16 @@ void Curve::snapBnd(const Vector &mouse, const Matrix &m,
 		    Vector &pos, double &bound) const
 {
   snapVertex(mouse, m * segment(0).cp(0), pos, bound);
-  for (int i = 0; i < countSegments(); ++i)
+  for (int i = 0; i < countSegmentsClosing(); ++i)
     segment(i).snapBnd(mouse, m, pos, bound);
-  if (closed()) {
-    Vector u[2];
-    closingSegment(u).snapBnd(mouse, m, pos, bound);
-  }
 }
 
 //! Returns the closing segment of a closed path.
-/*! Since the closing segment isn't actually stored inside this
-  object, you have to provide a length-2 vector for the control points.
-
-  This method panics if the Curve is not closed.
-*/
-CurveSegment Curve::closingSegment(Vector u[2]) const
+/*! This method panics if the Curve is not closed. */
+CurveSegment Curve::closingSegment() const
 {
   assert(iClosed);
-  u[0] = iCP.back();
-  u[1] = iCP.front();
-  return CurveSegment(CurveSegment::ESegment, 2, u, nullptr);
+  return CurveSegment(this, iSeg.size() - 1);
 }
 
 // --------------------------------------------------------------------
@@ -895,6 +997,7 @@ bool Shape::load(String data)
   String type;
   Curve *sp = nullptr;
   Vector org;
+  int mid = -1;
   std::vector<double> args;
   do {
     if (stream.token() == "h") { // closing path
@@ -903,6 +1006,7 @@ bool Shape::load(String data)
       stream.nextToken(); // eat token
       sp->setClosed(true);
       sp = nullptr;
+      mid = -1;
     } else if (stream.token() == "m") {
       if (args.size() != 2)
 	return false;
@@ -911,8 +1015,8 @@ bool Shape::load(String data)
       sp = new Curve;
       appendSubPath(sp);
       org = getVector(args);
+      mid = -1;
     } else if (stream.token() == "l") {
-      // assert(sp && args.size() > 0 && (args.size() % 2 == 0));
       if (!sp || args.size() != 2)
 	return false;
       stream.nextToken(); // eat token
@@ -921,6 +1025,7 @@ bool Shape::load(String data)
 	sp->appendSegment(org, v);
 	org = v;
       }
+      mid = -1;
     } else if (stream.token() == "a") {
       if (!sp || args.size() != 8)
 	return false;
@@ -931,26 +1036,44 @@ bool Shape::load(String data)
       Vector v1 = getVector(args);
       sp->appendArc(m, org, v1);
       org = v1;
+      mid = -1;
     } else if (stream.token() == "s" || stream.token() == "q"
-	       || stream.token() == "c") {
-      if (!sp || args.size() < 2 || (args.size() % 2 != 0))
+	       || stream.token() == "c" || stream.token() == "C"
+	       || stream.token() == "L") {
+      size_t parity = (stream.token() == "C") ? 1 : 0;
+      if (!sp || args.size() < 2 || (args.size() % 2 != parity))
 	return false;
-      bool oldStyle = (stream.token() == "s");
+      String typeToken = stream.token();
       stream.nextToken();
       std::vector<Vector> v;
       v.push_back(org);
-      while (!args.empty())
+      while (args.size() >= 2)
 	v.push_back(getVector(args));
-      if (oldStyle)
+      if (typeToken == "s")
 	sp->appendOldSpline(v);
-      else
+      else if (typeToken == "C") {
+	sp->appendCardinalSpline(v, float(args.back()));
+	args.pop_back();  // remove tension
+      } else if (typeToken == "L") {
+	if (mid >= 0) {
+	  if ((mid % 2) != 0)  // wrong parity
+	    return false;
+	  mid = (mid / 2);     // count Vectors
+	  if (mid < 2 || (mid % 3) != 2)  // * marker in wrong position
+	    return false;
+	  sp->appendSpiroSplinePrecomputed(v, mid + 1);
+	} else
+	  sp->appendSpiroSpline(v);
+      } else
 	sp->appendSpline(v);
       org = v.back();
+      mid = -1;
     } else if (stream.token() == "e") {
       if (args.size() != 6)
 	return false;
       stream.nextToken();
       sp = nullptr;
+      mid = -1;
       Ellipse *e = new Ellipse(getMatrix(args));
       appendSubPath(e);
     } else if (stream.token() == "u") {
@@ -958,11 +1081,16 @@ bool Shape::load(String data)
 	return false;
       stream.nextToken();
       sp = nullptr;
+      mid = -1;
       std::vector<Vector> v;
       while (!args.empty())
 	v.push_back(getVector(args));
       ClosedSpline *e = new ClosedSpline(v);
       appendSubPath(e);
+    } else if (stream.token() == "*") {
+      // remember position in args
+      mid = args.size();
+      stream.nextToken();
     } else { // must be a number
       double num;
       stream >> num;
@@ -970,9 +1098,11 @@ bool Shape::load(String data)
     }
     stream.skipWhitespace();
   } while (!stream.eos());
+  // we allow the last subpath to be empty (a single trailing "m" operator)
+  int sbn = countSubPaths();
+  if (sbn > 0 && subPath(sbn-1)->asCurve() && subPath(sbn-1)->asCurve()->countSegments() == 0)
+    iImp->iSubPaths.pop_back();
   // sanity checks
-  if (countSubPaths() == 0)
-    return false;
   for (int i = 0; i < countSubPaths(); ++i) {
     if (subPath(i)->asCurve() && subPath(i)->asCurve()->countSegments() == 0)
       return false;
