@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -69,6 +69,14 @@ const char *const ipelua::linecap_names[] =
 const char *const ipelua::fillrule_names[] =
   { "normal", "wind", "evenodd", nullptr };
 
+const char * const ipelua::segtype_names[] =
+  { "arc", "segment", "spline", "oldspline", "cardinal", "spiro", nullptr };
+
+static int segtype_cp[] = { 2, 2, 0, 0, 0, 0 };
+
+static const char *const splinetype_names[] =
+  { "bspline", "cardinal", "spiro" };
+
 // --------------------------------------------------------------------
 
 void ipelua::push_string(lua_State *L, String str)
@@ -216,6 +224,9 @@ Attribute ipelua::check_property(Property prop, lua_State *L, int i)
   case EPropPathMode:
     val = luaL_checkoption(L, i, nullptr, pathmode_names);
     return Attribute(TPathMode(val));
+  case EPropSplineType:
+    val = luaL_checkoption(L, i, nullptr, splinetype_names);
+    return Attribute(TSplineType(val));
   case EPropPen:
   case EPropSymbolSize:
   case EPropFArrowSize:
@@ -326,6 +337,9 @@ void ipelua::check_allattributes(lua_State *L, int i, AllAttributes &all)
   t = get_option(L, i, "transformations", transformation_names);
   if (t >= 0) all.iTransformations = TTransformations(t);
 
+  t = get_option(L, i, "splinetype", splinetype_names);
+  if (t >= 0) all.iSplineType = TSplineType(t);
+
   t = get_option(L, i, "pathmode", pathmode_names);
   if (t >= 0) all.iPathMode = TPathMode(t);
 }
@@ -397,7 +411,8 @@ int ipelua::xml_constructor(lua_State *L)
   ImlParser parser(source);
   String tag = parser.parseToTag();
   if (tag == "ipeselection") {
-    lua_newtable(L);
+    lua_newtable(L);  // the objects
+    lua_newtable(L);  // their layers
     int index = 1;
 
     XmlAttributes attr;
@@ -413,11 +428,14 @@ int ipelua::xml_constructor(lua_State *L)
 
     for (;;) {
       if (tag == "/ipeselection")
-	return 1;
-      Object *obj = parser.parseObject(tag);
+	return 2;
+      String layer;
+      Object *obj = parser.parseObject(tag, layer);
       if (!obj)
 	return 0;
       push_object(L, obj);
+      lua_rawseti(L, -3, index);
+      push_string(L, layer);
       lua_rawseti(L, -2, index);
       ++index;
       tag = parser.parseToTag();
@@ -599,11 +617,6 @@ static int object_addToBBox(lua_State *L)
 
 // --------------------------------------------------------------------
 
-static const char * const segtype_names[] =
-  { "arc", "segment", "spline", "oldspline", nullptr };
-
-static int segtype_cp[] = { 2, 2, 0, 0 };
-
 static const char * const subpath_names[] =
   { "curve", "ellipse", "closedspline", nullptr };
 
@@ -647,7 +660,7 @@ static SubPath *get_curve(lua_State *L, int index)
   lua_getfield(L, -1, "closed");
   if (!lua_isboolean(L, -1))
     luaL_error(L, "element %d has no 'closed' field", index);
-  c->setClosed(lua_toboolean(L, -1));
+  bool closed = lua_toboolean(L, -1);
   lua_pop(L, 1);  // closed
   for (int i = 0; ; ++i) {
     lua_rawgeti(L, -1, i+1);
@@ -655,6 +668,7 @@ static SubPath *get_curve(lua_State *L, int index)
       lua_pop(L, 1);
       if (c->countSegments() == 0)
 	luaL_error(L, "element %d has no segments", index);
+      c->setClosed(closed);
       return c.release();
     }
     if (!lua_istable(L, -1))
@@ -692,6 +706,17 @@ static SubPath *get_curve(lua_State *L, int index)
     case CurveSegment::EOldSpline:
       c->appendOldSpline(cp);
       break;
+    case CurveSegment::ECardinalSpline: {
+      lua_getfield(L, -1, "tension");
+      if (!lua_isnumber(L, -1))
+	luaL_error(L, "segment %d of element %d has no tension", i+1, index);
+      float tension = lua_tonumberx(L, -1, nullptr);
+      c->appendCardinalSpline(cp, tension);
+      lua_pop(L, 1); // tension
+      break; }
+    case CurveSegment::ESpiroSpline:
+      c->appendSpiroSpline(cp);
+      break;
     default:
       break;
     }
@@ -737,7 +762,10 @@ Shape ipelua::check_shape(lua_State *L, int index)
 
 static void push_segment(lua_State *L, const CurveSegment &seg)
 {
-  lua_createtable(L, seg.countCP(), (seg.type() == CurveSegment::EArc ? 2 : 1));
+  int fields = 1;
+  if (seg.type() == CurveSegment::EArc) ++fields;
+  if (seg.type() == CurveSegment::ECardinalSpline) ++fields;
+  lua_createtable(L, seg.countCP(), fields);
   lua_pushstring(L, segtype_names[seg.type()]);
   lua_setfield(L, -2, "type");
   for (int i = 0; i < seg.countCP(); ++i) {
@@ -748,6 +776,10 @@ static void push_segment(lua_State *L, const CurveSegment &seg)
     push_arc(L, seg.arc());
     lua_setfield(L, -2, "arc");
   }
+  if (seg.type() == CurveSegment::ECardinalSpline) {
+    lua_pushnumber(L, round(seg.tension() * 10000.0) / 10000.0);
+    lua_setfield(L, -2, "tension");
+  }
 }
 
 static void push_subpath(lua_State *L, const SubPath *sp)
