@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -94,7 +94,9 @@ private:
   void layoutChildren();
   void zoomMain(int delta);
   void timerElapsed();
-  void handlePdfViewMessage(int param);
+  void handlePdfViewMessage(int param, const Vector &pos);
+  virtual void showType3Warning(const char *s) override;
+  virtual void browseLaunch(bool launch, String dest) override;
 
   static LRESULT CALLBACK wndProc(HWND hwnd, UINT Message,
 				  WPARAM wParam, LPARAM lParam);
@@ -169,6 +171,7 @@ void AppUi::initUi()
   HMENU hSubMenu = CreatePopupMenu();
   AppendMenuA(hSubMenu, MF_STRING, EShowPresentation, "Show &Presentation");
   AppendMenuA(hSubMenu, MF_STRING, EFullScreen, "Full &Screen\tF11");
+  AppendMenuA(hSubMenu, MF_STRING, EBlackout, "Blackout\tB");
   AppendMenuA(hSubMenu, MF_STRING, EZoomIn, "Larger\tCtrl++");
   AppendMenuA(hSubMenu, MF_STRING, EZoomOut, "Smaller\tCtrl+-");
   AppendMenuA(hMenuBar, MF_STRING | MF_POPUP, UINT_PTR(hSubMenu), "&View");
@@ -199,9 +202,10 @@ void AppUi::initUi()
 
   HINSTANCE hInstance = (HINSTANCE) GetWindowLongPtr(hwnd, GWLP_HINSTANCE);
 
-  iCurrent = new PdfView(hwnd, hwnd);
-  iNext = new PdfView(hwnd, hwnd);
-  iScreen = new PdfView(nullptr, hwnd, hInstance);
+  iCurrent = new PdfView(hwnd, hwnd, 0x00);
+  iNext = new PdfView(hwnd, hwnd, 0x10);
+  iScreen = new PdfView(nullptr, hwnd, 0x20, hInstance);
+  iScreen->setBackground(Color(0, 0, 0));
 
   hNotes = CreateWindowExW(0, L"edit", nullptr,
 			   WS_CHILD | WS_VISIBLE | WS_BORDER | WS_VSCROLL |
@@ -279,12 +283,22 @@ void AppUi::layoutChildren()
   InvalidateRect(hClock, nullptr, FALSE);
 }
 
-void AppUi::handlePdfViewMessage(int param)
+void AppUi::handlePdfViewMessage(int param, const Vector &pos)
 {
   // ignore all messages before PDF file loaded
   if (!iPdf)
     return;
-  switch (param) {
+  // ipeDebug("handlePdfViewMessage %x", param);
+  // button press on current view
+  if ((param & 0x0f) != 0 && (param & 0xf0) == 0) {
+    const PdfDict *action = findLink(iCurrent->devToUser(pos));
+    if (action) {
+      interpretAction(action);
+      setView();
+      return;
+    }
+  }
+  switch (param & 0x0f) {
   case 0:
     fitBoxAll();
     break;
@@ -300,6 +314,8 @@ void AppUi::handlePdfViewMessage(int param)
 void AppUi::toggleFullScreen()
 {
   HWND hwnd = iScreen->windowId();
+  if (!IsWindowVisible(hwnd))
+    return;
   if (!iFullScreen) {
     HMONITOR hmon = MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST);
     MONITORINFO mi = { sizeof(mi) };
@@ -366,9 +382,9 @@ bool AppUi::load(const char* fn)
 
 void AppUi::setPdf()
 {
-  iScreen->setPdf(iPdf.get());
-  iCurrent->setPdf(iPdf.get());
-  iNext->setPdf(iPdf.get());
+  iScreen->setPdf(iPdf.get(), iFonts.get());
+  iCurrent->setPdf(iPdf.get(), iFonts.get());
+  iNext->setPdf(iPdf.get(), iFonts.get());
 }
 
 void AppUi::setView()
@@ -426,6 +442,10 @@ void AppUi::cmd(int cmd)
   case EFullScreen:
     toggleFullScreen();
     break;
+  case EBlackout:
+    iScreen->setBlackout(!iScreen->blackout());
+    iScreen->updatePdf();
+    break;
   case EZoomIn:
     zoomMain(+1);
     break;
@@ -491,22 +511,30 @@ void AppUi::cmd(int cmd)
   }
 }
 
+void AppUi::browseLaunch(bool launch, String dest)
+{
+  ipeDebug("Launch %d %s", launch, dest.z());
+  ShellExecuteA(nullptr, "open", dest.z(), nullptr, nullptr, SW_SHOWNORMAL);
+}
+
 // --------------------------------------------------------------------
 
 static const char * const aboutText =
   "IpePresenter %d.%d.%d\n\n"
-  "Copyright (c) 2019 Otfried Cheong\n\n"
+  "Copyright (c) 2020 Otfried Cheong\n\n"
   "A presentation tool for giving PDF presentations "
   "created in Ipe or using beamer.\n"
   "Originally invented by Dmitriy Morozov, "
-  "IpePresenter is now released as part of Ipe under the GNU Public License.\n"
-  "See http://ipe.otfried.org for details.\n\n"
+  "IpePresenter is now developed together with Ipe and released under the GNU Public License.\n"
+  "See http://ipepresenter.otfried.org for details.\n\n"
   "You can \"like\" IpePresenter and follow IpePresenter announcements on Facebook "
   "(http://www.facebook.com/drawing.editor.Ipe7).\n\n"
   "If you are an IpePresenter fan and want to show others, have a look at the "
   "Ipe T-shirts (www.shirtee.com/en/store/ipe).\n\n"
-  "Platinum sponsors\n\n"
+  "Platinum and gold sponsors\n\n"
   " * Hee-Kap Ahn\n"
+  " * Günter Rote\n"
+  " * SCALGO\n"
   " * Martin Ziegler\n\n"
   "If you enjoy IpePresenter, you can become a member of the exclusive community of "
   "Ipe patrons (http://patreon.com/otfried). "
@@ -521,7 +549,20 @@ void AppUi::aboutIpePresenter()
 	  (IPELIB_VERSION / 100) % 100,
 	  IPELIB_VERSION % 100);
 
-  MessageBoxA(hwnd, &buf[0], "About IpePresenter",
+  const char *s = buf.data();
+  int n = strlen(s);
+  int rw = MultiByteToWideChar(CP_UTF8, 0, s, n, nullptr, 0);
+  std::wstring wbuf;
+  wbuf.resize(rw + 1, wchar_t(0));
+  MultiByteToWideChar(CP_UTF8, 0, s, n, wbuf.data(), rw);
+
+  MessageBoxW(hwnd, wbuf.data(), L"About IpePresenter",
+	      MB_OK | MB_ICONINFORMATION | MB_APPLMODAL);
+}
+
+void AppUi::showType3Warning(const char *s)
+{
+  MessageBoxA(hwnd, s, "Type3 font detected",
 	      MB_OK | MB_ICONINFORMATION | MB_APPLMODAL);
 }
 
@@ -655,8 +696,11 @@ LRESULT CALLBACK AppUi::wndProc(HWND hwnd, UINT message, WPARAM wParam,
     delete ui;
     break;
   case PdfView::WM_PDFVIEW:
-    if (ui)
-      ui->handlePdfViewMessage(lParam);
+    if (ui) {
+      int x = GET_X_LPARAM(lParam);
+      int y = GET_Y_LPARAM(lParam);
+      ui->handlePdfViewMessage(wParam, Vector(x, y));
+    }
     break;
   default:
     break;
@@ -760,6 +804,8 @@ int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
     { FVIRTKEY, VK_HOME, AppUi::EFirstView },
     { FVIRTKEY, VK_END, AppUi::ELastView },
     { FVIRTKEY, VK_F11, AppUi::EFullScreen },
+    { FVIRTKEY, VK_F5, AppUi::EShowPresentation },
+    { FVIRTKEY, 'B', AppUi::EBlackout },
   };
   HACCEL hAccel = CreateAcceleratorTable(accel, sizeof(accel)/sizeof(ACCEL));
 
