@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -123,50 +123,6 @@ static NSLayoutRelation layoutRelation(char ch)
   }
 }
 
-static NSView *owner(id a)
-{
-  if ([a respondsToSelector:@selector(superview)])
-    return [a superview];
-#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101100
-  if ([a respondsToSelector:@selector(owningView)])
-    return [a owningView];  // is a layout guide
-#endif
-  return nil;
-}
-
-void activateConstraint(NSLayoutConstraint *c, BOOL active)
-{
-#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101000
-  if ([c respondsToSelector:@selector(setActive:)]) {
-#pragma clang diagnostic push
-#pragma clang diagnostic ignored "-Wdeprecated-declarations"
-    c.active = active;
-#pragma clang diagnostic pop
-    return;
-  }
-#endif
-  // need to figure out lowest common ancestor
-  id a = c.firstItem;
-  id b = c.secondItem;
-  NSView *lca = nil;
-  NSView *ao = owner(a);
-  NSView *bo = owner(b);
-  if (b == nil)
-    lca = a;
-  else if (ao == b)
-    lca = b;
-  else if (bo == a)
-    lca = a;
-  else {
-    assert(ao == bo);
-    lca = ao;
-  }
-  if (active)
-    [lca addConstraint:c];
-  else
-    [lca removeConstraint:c];
-}
-
 NSLayoutConstraint *layout(id a, id b, const char *rel, double gap,
 			   double multiplier, BOOL activate)
 {
@@ -182,8 +138,7 @@ NSLayoutConstraint *layout(id a, id b, const char *rel, double gap,
 	       attribute:b1
 	      multiplier:multiplier
 		constant:gap];
-  if (activate)
-    activateConstraint(c, YES);
+  c.active = activate;
   return c;
 }
 
@@ -550,7 +505,7 @@ void PDialog::layoutControls()
 bool PDialog::buildAndRun(int w, int h)
 {
   NSUInteger style = NSTitledWindowMask|NSResizableWindowMask;
-  if (!iIgnoreEscape)
+  if (iIgnoreEscapeField >= 0)
     style |= NSClosableWindowMask;
 
   iPanel = [[NSPanel alloc]
@@ -1318,9 +1273,16 @@ static int ipeui_currentDateTime(lua_State *L)
 
 static int ipeui_startBrowser(lua_State *L)
 {
-  const char *url = luaL_checkstring(L, 1);
-  int res = [[NSWorkspace sharedWorkspace]
-	      openURL:[NSURL URLWithString:C2N(url)]];
+  NSString *urls = C2N(luaL_checkstring(L, 1));
+  NSURL *url = nil;
+  if ([urls hasPrefix:@"file:///"]) {
+    url = [NSURL fileURLWithPath:[urls substringFromIndex:8] isDirectory:NO];
+  } else {
+    // TODO: percent-escape chars not allowed in URLS (such as spaces)
+    // (not urgent since we only use this for the manual on the file system
+    url = [NSURL URLWithString:urls];
+  }
+  int res = [[NSWorkspace sharedWorkspace] openURL:url];
   lua_pushboolean(L, res);
   return 1;
 }
