@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -180,8 +180,11 @@ PDialog::~PDialog()
 
 void PDialog::keyPressEvent(QKeyEvent *e)
 {
-  if (iIgnoreEscape && e->key() == Qt::Key_Escape)
-    return;
+  if (iIgnoreEscapeField >= 0 && e->key() == Qt::Key_Escape) {
+    retrieveValues();
+    if (iElements[iIgnoreEscapeField].text != iIgnoreEscapeText)
+      return;  // text has been modified, do not allow ESC
+  }
   QDialog::keyPressEvent(e);
 }
 
