@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -129,12 +129,14 @@ protected:
   virtual void enableItem(int idx, bool value);
 
   void computeDimensions(int &w, int &h);
+  void getDimensions(SElement &m, int &x, int &y, int &w, int &h);
   void buildDimensions(std::vector<short> &t, SElement &m, int id);
 
   virtual void acceptDialog(lua_State *L);
 
   BOOL initDialog();
   BOOL dlgCommand(WPARAM wParam, LPARAM lParam);
+  BOOL handleResize();
   static BOOL CALLBACK dialogProc(HWND hwndDlg, UINT message,
 				  WPARAM wParam, LPARAM lParam);
 private:
@@ -306,12 +308,12 @@ void PDialog::retrieveValues()
   }
 }
 
-void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
+void PDialog::getDimensions(SElement &m, int &x, int &y, int &w, int &h)
 {
-  int x = BORDER;
-  int y = BORDER;
-  int w = 0;
-  int h = 0;
+  x = BORDER;
+  y = BORDER;
+  w = 0;
+  h = 0;
   if (m.row < 0) {
     y = BORDER;
     for (int i = 0; i < iNoRows; ++i)
@@ -332,6 +334,12 @@ void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
     for (int i = m.row + 1; i < m.row + m.rowspan; ++i)
       h += iRowHeight[i] + PAD;
   }
+}
+
+void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
+{
+  int x, y, w, h;
+  getDimensions(m, x, y, w, h);
   t.push_back(x);
   t.push_back(y);
   t.push_back(w);
@@ -347,10 +355,14 @@ BOOL PDialog::dlgCommand(WPARAM wParam, LPARAM lParam)
     EndDialog(hDialog, TRUE);
     return TRUE;
   }
-  if (id == IDCANCEL && !iIgnoreEscape) {
+  if (id == IDCANCEL) {
     retrieveValues();
-    EndDialog(hDialog, FALSE);
-    return TRUE;
+    if (iIgnoreEscapeField < 0 ||
+	iElements[iIgnoreEscapeField].text == iIgnoreEscapeText) {
+      // allowed to close dialog
+      EndDialog(hDialog, FALSE);
+      return TRUE;
+    }
   }
   if (id < IDBASE || id >= IDBASE + int(iElements.size()))
     return FALSE;
@@ -414,6 +426,10 @@ BOOL CALLBACK PDialog::dialogProc(HWND hwnd, UINT message,
       return d->dlgCommand(wParam, lParam);
     else
       return FALSE;
+  case WM_SIZE:
+    if (d)
+      return d->handleResize();
+    return FALSE;
   case WM_DESTROY:
     // Remove the subclasses from text edits
     for (int i = 0; i < int(d->iElements.size()); ++i) {
@@ -488,6 +504,7 @@ void PDialog::buildElements(std::vector<short> &t)
 
 bool PDialog::buildAndRun(int w, int h)
 {
+  // converts dimensions to dialog units and possibly changes them
   computeDimensions(w, h);
 
   RECT rect;
@@ -498,12 +515,12 @@ bool PDialog::buildAndRun(int w, int h)
   std::vector<short> t;
 
   // Dialog flags
-  buildFlags(t, WS_POPUP | WS_BORDER | DS_SHELLFONT |
+  buildFlags(t, WS_POPUP | WS_BORDER | DS_SHELLFONT | WS_SIZEBOX |
 	     WS_SYSMENU | DS_MODALFRAME | WS_CAPTION);
   t.push_back(iElements.size());
-  t.push_back((pw - w)/2);  // offset of popup-window from parent window
+  t.push_back((pw - w)/2);  // offset of popup-window from parent window in dialog units
   t.push_back((ph - h)/2);
-  t.push_back(w);
+  t.push_back(w);  // should be dialog units
   t.push_back(h);
   // menu
   t.push_back(0);
@@ -525,6 +542,7 @@ bool PDialog::buildAndRun(int w, int h)
   return (res > 0);
 }
 
+// converts (w, h) from pixels to dialog units and possibly increases them to fit dialog
 void PDialog::computeDimensions(int &w, int &h)
 {
   int minWidth[iNoCols];
@@ -613,8 +631,9 @@ void PDialog::computeDimensions(int &w, int &h)
       iRowHeight[i] = minHeight[i] + extra;
     }
   }
+
   /*
-  fprintf(stderr, "iColWidth\n");
+  fprintf(stderr, "iNoCols: %d, iColWidth = ", iNoCols);
   for (int i = 0; i < iNoCols; ++i)
     fprintf(stderr, "%d ", iColWidth[i]);
   fprintf(stderr, "\n");
@@ -622,6 +641,27 @@ void PDialog::computeDimensions(int &w, int &h)
   */
 }
 
+BOOL PDialog::handleResize()
+{
+  RECT rc;
+  GetClientRect(hDialog, &rc);
+  int bw = rc.right - rc.left;
+  int bh = rc.bottom - rc.top;
+  computeDimensions(bw, bh);
+
+  for (int i = 0; i < int(iElements.size()); ++i) {
+    SElement &m = iElements[i];
+    int id = i + IDBASE;
+    int x, y, w, h;
+    getDimensions(m, x, y, w, h);
+    HWND hwnd = GetDlgItem(hDialog, id);
+    // calculation is in dialog units, so convert back to pixels
+    MoveWindow(hwnd, x * iBaseX / 4 , y * iBaseY / 8 ,
+	       w * iBaseX / 4, h * iBaseY / 8, TRUE);
+  }
+  return TRUE;
+}
+
 void PDialog::enableItem(int idx, bool value)
 {
   EnableWindow(GetDlgItem(hDialog, idx+IDBASE), value);
